import "../smemory/memory.clp".
import "aliasing_sum.clp".

using set.
using display.
using boundops.

% ESCAPING PREDICATES

% level of precision controlling the granularity of an escape search
type t_escape_precision ::=
    ep_field{FLD:int}  % preserve at most FLD fields, then abstract
  | ep_nocontext       % never go to an invariant, but drop context info
  | ep_maximum.        % don't abstract any information

% direction to escape in, either forward or backward with a
% specified level of precision
type t_escape_state ::=
    es_forward
  | es_backward
  | es_forward_precise{t_escape_precision}
  | es_backward_precise{t_escape_precision}.

predicate try_escape(TR:int,Y:t_escape_state,L:t_trace_loc).
predicate escape(TR:int,Y:t_escape_state,L:t_trace_loc,NL:t_trace_loc).
predicate escape_fail(TR:int,Y:t_escape_state,L:t_trace_loc).

predicate escape_set(in TR:int,in Y:t_escape_state,in L:t_trace_loc,
                     NLS:set[t_trace_loc]).

predicate print_escape(TR:int,Y:t_escape_state,L:t_trace_loc).

% printed out according to generated follow_escape
predicate str_escape(Y:t_escape_state,L:string,NL:string).
predicate str_escape_why(Y:t_escape_state,L:string,NL:string,PL:string).
predicate str_escape_fail(Y:t_escape_state,L:string).
%?- str_escape(_,_,NL).
?- str_escape_fail(_,_).
?- str_escape_why(_,_,NL,PL).

predicate ignore_escape_edge(in FWD:bool,in L:t_trace_loc,in NL:t_trace_loc).
ignore_escape_edge(true,L,NL) :- ignore_escape_edge(false,NL,L).
ignore_escape_edge(false,L,NL) :- ignore_escape_edge(true,NL,L).

ignore_escape_edge(_,L,_) :- trace_loc_trace(L,T),
    trace_sub(T,rfld{_,F,C},_), ~salias_rfld(C,F)->srfld(_,_,_,_).

% ESCAPE UTILITY

predicate unsanitary_trace(in T:t_trace).
+unsanitary_trace(uc_sum{_,_,_}).
+unsanitary_trace(uc_exp{_,_,_}).
+unsanitary_trace(index{_,_,_}).

predicate unsanitary_trace_loc(in L:t_trace_loc).
unsanitary_trace_loc(tl_comp{_,T}) :-
    trace_sub(T,ST,_), unsanitary_trace(ST).
unsanitary_trace_loc(tl_glob{T}) :-
    trace_sub(T,ST,_), unsanitary_trace(ST).
unsanitary_trace_loc(tl_func{_,T,_}) :-
    trace_sub(T,ST,_), unsanitary_trace(ST).

predicate single_trace_skolem(in T:t_trace,NT:t_trace).
single_trace_skolem(T,NT) :-
    trace_sub(T,drf{fld{BT,F,C}},RT), comp_field(C,_,_,F,y_ptr{y_comp{C}}),
    \/comp_field(C,_,_,XF,y_ptr{y_comp{C}}):list_all(drf{fld{empty,XF,C}},RL),
    trace_simplify(uk_recurse{BT,RL},NBT),
    trace_compose(NBT,RT,NT).
single_trace_skolem(T,NT) :-
    trace_sub(T,index{BT,Y,_},RT),
    trace_simplify(uk_index{BT,Y},NBT),
    trace_compose(NBT,RT,NT).
single_trace_skolem(T,NT) :-
    trace_sub(T,uc_sum{_,_,BT},RT),
    trace_compose(drf{BT},RT,NT).

predicate reduce_trace_skolem(in T:t_trace,NT:t_trace).
reduce_trace_skolem(T,NT) :-
    \/single_trace_skolem(T,XT):list_all(XT,[MT|_]),
    reduce_trace_skolem(MT,NT).
reduce_trace_skolem(T,T) :-
    ~single_trace_skolem(T,_).

% replace any uses of array indexes, recursive type accesses, or clobbered
% traces with the appropriate skolem or drf{_}
predicate reduce_trace_loc_skolem(in L:t_trace_loc,NL:t_trace_loc)
    succeeds [once].
reduce_trace_loc_skolem(tl_func{FN,T,CXT},tl_func{FN,NT,CXT}) :-
    reduce_trace_skolem(T,NT).
reduce_trace_loc_skolem(tl_glob{T},tl_glob{NT}) :-
    reduce_trace_skolem(T,NT).
reduce_trace_loc_skolem(tl_comp{C,RT},tl_comp{C,NRT}) :-
    reduce_trace_skolem(RT,NRT).

predicate reduce_trace_loc_field(in L:t_trace_loc,NL:t_trace_loc)
    succeeds [zero,once].
reduce_trace_loc_field(L,tl_comp{C,RT}) :-
    (L=tl_func{_,T,_}; L=tl_glob{T}; L=tl_comp{_,T}),
    trace_sub(T,fld{STR,_,C},XT), ~trace_sub(XT,fld{_,_,_},_),
    trace_sub(T,STR,RT).

% L should already have had reduce_trace_skolem applied.
% all edges in salias_escape.db have already been skolemized
predicate reduce_trace_loc(in L:t_trace_loc,NL:t_trace_loc) succeeds [once].
reduce_trace_loc(L,NL) :- reduce_trace_loc_field(L,NL).
reduce_trace_loc(L,tl_glob{T}) :- ~reduce_trace_loc_field(L,_),
    (L=tl_glob{T}; L=tl_func{_,T,_}, trace_root(T,glob{_})).
reduce_trace_loc(L,L) :- ~reduce_trace_loc_field(L,_),
    (L=tl_comp{_,_}; L=tl_func{_,T,_}, ~trace_root(T,glob{_})).

predicate one_precise_reduce_trace_loc(in V:t_escape_precision,
                                       in L:t_trace_loc,NL:t_trace_loc).
one_precise_reduce_trace_loc(ep_nocontext,L,NL) :-
    L  = tl_func{FN,T,cxt_call{PFN,PI,CXT}}, CXT\=cxt_any,
    NL = tl_func{FN,T,cxt_call{PFN,PI,cxt_any}}.
one_precise_reduce_trace_loc(ep_field{N},L,NL) :-
    trace_loc_trace(L,T),
    trace_sub(T,fld{XT,_,C},_), trace_sub(T,XT,RT),
    \/trace_sub(RT,fld{MT,_,_},_):count_all(MT,COUNT), COUNT=N,
    NL = tl_comp{C,RT}, L\=NL.
one_precise_reduce_trace_loc(ep_field{_},L,NL) :-
    trace_loc_trace(L,T),
    (trace_sub(T,uk_recurse{_,_},XT); trace_sub(T,rfld{_,_,_},XT)),
    trace_sub(XT,fld{BT,_,C},_), ~trace_sub(BT,fld{_,_,_},_),
    trace_sub(XT,BT,RT), NL = tl_comp{C,RT}, L\=NL.

predicate precise_reduce_trace_loc(in V:t_escape_precision,
                                   in L:t_trace_loc,NL:t_trace_loc)
    succeeds [once].
precise_reduce_trace_loc(V,L,L) :- ~one_precise_reduce_trace_loc(V,L,_).
precise_reduce_trace_loc(V,L,NL) :-
    \/one_precise_reduce_trace_loc(V,L,WL):list_all(WL,[ML|_]),
    precise_reduce_trace_loc(V,ML,NL).

% reduction predicate that might succeed multiple times, with multiple levels
% of precision for the same tl_func{FN,T,CXT}
predicate one_reduce_trace_loc(in FN:string,in T:t_trace,in CXT:t_call_context,
                               NL:t_trace_loc) succeeds [many].
one_reduce_trace_loc(FN,T,CXT,NL) :-
    reduce_trace_loc(tl_func{FN,T,CXT},NL).
one_reduce_trace_loc(_,T,_,tl_comp{C,RT}) :-
    trace_sub(T,fld{ST,_,C},_), trace_sub(T,ST,RT).
one_reduce_trace_loc(_,T,_,tl_glob{T}) :-
    trace_root(T,glob{_}).
one_reduce_trace_loc(FN,T,CXT,tl_func{FN,T,CXT}) :-
    trace_root(T,R), R\=glob{_}.

predicate relative_access_candidate(in L:t_trace_loc).
+relative_access_candidate(tl_comp{_,fld{empty,_,_}}).
+relative_access_candidate(tl_comp{_,uk_index{fld{empty,_,_},_}}).
+relative_access_candidate(tl_glob{_}).
+relative_access_candidate(tl_func{_,root{local{_}},_}).

predicate follow_relative_access_candidate(in L:t_trace_loc,
                                           NLSET:set[t_trace_loc]).
follow_relative_access_candidate(L,NLSET) :-
    relative_access_candidate(L), set_singleton(L,NLSET).
follow_relative_access_candidate(L,NLSET) :-
    ~relative_access_candidate(L), L=tl_comp{_,_},
    ~(escape_edge(false,L,XL), reduce_trace_loc(XL,MXL),
      ~relative_access_candidate(MXL)),
    \/(escape_edge(false,L,NL), reduce_trace_loc(NL,MNL)):set_all(MNL,NLSET).
follow_relative_access_candidate(L,NLSET) :-
    ~relative_access_candidate(L),
    L=tl_func{_,T,_}, trace_root(T,R), (R=local{_}; R=arg{_}),
    ~(escape_edge(false,L,XL), reduce_trace_loc(XL,MXL),
      ~relative_access_candidate(MXL), XL\=tl_func{_,uk_index{T,_},_}),
    \/(escape_edge(false,L,NL), reduce_trace_loc(NL,MNL)):set_all(MNL,NLSET).

% if the data at L is assigned forward to NL, this may lead to a future
% write to TAIL relative to L/NL with the contents of NLS.
% if no success, we don't know about anything about resulting future writes
predicate trace_loc_any_write(in L:t_trace_loc,in NL:t_trace_loc,
                              in TAIL:t_trace,NLS:set[t_trace_loc]).

% taking the address of an internal structure, either &x->f or &g
% where we are interested in the value of an internal field h of x->f or g.
% if field h has no non-relative writes (all writes are of the form
% x->f.h = ... or g.h = ..., but not y->h = ...),
% find those relative writes which are compatible with the initial location L
trace_loc_any_write(L,_,TAIL,NLS) :-
    (TAIL=drf{fld{empty,F,C}}; TAIL=drf{fld{uk_index{empty,_},F,C}}),
    ~salias_comp_relative(C,F)->wcomp_nonrelative(_),
    reduce_trace_loc(L,ML), follow_relative_access_candidate(ML,NLSET),
    \/(set_member(NLSET,NL),
       salias_comp_relative(C,F)->wcomp_relative(NL,XL)):set_all(XL,NLS).

% data passed as an argument to a function, where we know that argument
% will not be used for any future writes
trace_loc_any_write(_,tl_func{FN,T,_},drf{RTAIL},NLS) :-
    trace_compose(T,RTAIL,NT), trace_root(NT,arg{_}),
    salias_func_readonly(FN)->wfunc_readonly(NT), set_empty(NLS).

% data passed to a function we're ignoring
trace_loc_any_write(_,tl_func{FN,_,_},_,NLS) :-
    ignore_escape_function(FN), set_empty(NLS).

% ESCAPE PROPAGATION

type t_escape_boundset =
    boundset[t_pair[t_escape_state,t_trace_loc],t_trace_loc].

predicate escape_edge_follow(in Y:t_escape_state,in L:t_trace_loc,
                             NL:t_trace_loc).
predicate escape_edge_follow_count(in Y:t_escape_state,in L:t_trace_loc,
                                   NLC:int).

predicate escape_boundset(in TR:int,in Y:t_escape_state,in L:t_trace_loc,
                          X:t_escape_boundset).
escape_boundset(TR,Y,L,X) :- boundset_create(pair{Y,L},TR,X).

predicate escape_boundset_overflow(in Y:t_escape_state,in L:t_trace_loc,
                                   in X:t_escape_boundset).
%escape_boundset_overflow(Y,L,X) :-
%    escape_edge_follow_count(Y,L,NLC), ~boundset_remaining(X,NLC).

try_escape(TR,Y,L), +escape(TR,Y,L,L).
escape(TR,Y,L,WL), escape_boundset(TR,Y,L,X),
    ~escape_boundset_overflow(Y,WL,X),
    escape_edge_follow(Y,WL,NL), boundset_insert(X,NL), +escape(TR,Y,L,NL).
escape(TR,Y,L,WL), escape_boundset(TR,Y,L,X),
    ~escape_boundset_overflow(Y,WL,X),
    escape_edge_follow(Y,WL,NL), ~boundset_insert(X,NL), +escape_fail(TR,Y,L).
escape(TR,Y,L,WL), escape_boundset(TR,Y,L,X),
    escape_boundset_overflow(Y,WL,X), +escape_fail(TR,Y,L).

try_escape(_,_,L), unsanitary_trace_loc(L), trace_loc_string(L,LS),
    +warning("Bad t_trace_loc for escape propagation",LS).

print_escape(TR,Y,L), +try_escape(TR,Y,L).
print_escape(TR,Y,L), escape_boundset(TR,Y,L,X), escape(TR,Y,L,WL),
    trace_loc_string(L,LS), trace_loc_string(WL,WLS), +str_escape(Y,LS,WLS),
    ~escape_boundset_overflow(Y,WL,X),
    escape_edge_follow(Y,WL,NL), boundset_insert(X,NL),
    trace_loc_string(NL,NLS), +str_escape_why(Y,LS,NLS,WLS).
print_escape(TR,Y,L), escape_fail(TR,Y,L),
    trace_loc_string(L,LS), +str_escape_fail(Y,LS).

predicate escape_set_try(in TR:int,in Y:t_escape_state,in L:t_trace_loc,
                         in WL:t_trace_loc,RES:maybe[t_trace_loc]).
+escape_set_try(_,_,_,WL,yes{WL}).
escape_set_try(TR,Y,L,WL,no) :- escape_boundset(TR,Y,L,X),
    escape_boundset_overflow(Y,WL,X).
escape_set_try(TR,Y,L,WL,no) :- escape_boundset(TR,Y,L,X),
    ~escape_boundset_overflow(Y,WL,X), escape_edge_follow(Y,WL,NL),
    ~boundset_insert(X,NL).
escape_set_try(TR,Y,L,WL,RES) :- escape_boundset(TR,Y,L,X),
    ~escape_boundset_overflow(Y,WL,X), escape_edge_follow(Y,WL,NL),
    boundset_insert(X,NL), escape_set_try(TR,Y,L,NL,RES).

escape_set(TR,Y,L,NLS) :-
    ~escape_set_try(TR,Y,L,L,no),
    \/escape_set_try(TR,Y,L,L,yes{NL}):set_all(NL,NLS).

?escape_set(_,_,L,_), unsanitary_trace_loc(L), trace_loc_string(L,LS),
    +warning("Bad t_trace_loc for escape propagation",LS).

% ESCAPE PROPAGATION RULES

predicate escape_global(in L:t_trace_loc,NL:t_trace_loc).
escape_global(tl_func{_,T,_},tl_glob{T}) :- trace_root(T,glob{_}).
escape_global(L,L) :- ~(L=tl_func{_,T,_}, trace_root(T,glob{_})).

predicate escape_edge(in FWD:bool,in L:t_trace_loc,NL:t_trace_loc).
escape_edge(FWD,L,NL) :- L=tl_func{FN,T,CXT},
    salias_escape(tl_func{FN,T,cxt_any},FWD)->sescape(_,SL,TL),
    SL=tl_func{FN,T,SX}, TL=tl_func{NFN,NT,TX},
    (SX=cxt_any, TX=cxt_call{PFN,PI,cxt_any}, NCXT=cxt_call{PFN,PI,CXT};
     SX=cxt_call{PFN,PI,cxt_any}, TX=cxt_any, CXT=cxt_call{PFN,PI,NCXT};
     SX=cxt_call{_,_,cxt_any}, TX=cxt_any, CXT=cxt_any, NCXT=cxt_any;
     SX=cxt_any, TX=cxt_any, NCXT=CXT),
    escape_global(tl_func{NFN,NT,NCXT},NL),
    ~ignore_escape_edge(FWD,L,NL).
escape_edge(FWD,L,NL) :- L\=tl_func{_,_,_},
    salias_escape(L,FWD)->sescape(_,_,TL),
    escape_global(TL,NL),
    ~ignore_escape_edge(FWD,L,NL).

predicate escape_edge_count(in FWD:bool,in T:t_trace_loc,NLC:int).
escape_edge_count(FWD,tl_func{FN,T,_},NLC) :-
    \/salias_escape(tl_func{FN,T,cxt_any},FWD)->sescape(_,_,TL)
     :count_all(TL,NLC).
escape_edge_count(FWD,L,NLC) :- L\=tl_func{_,_,_},
    \/salias_escape(L,FWD)->sescape(_,_,TL):count_all(TL,NLC).

predicate follow_reverse(in RT:t_trace).
follow_reverse(RT) :- trace_sub(RT,uk_recurse{_,_},_).
follow_reverse(RT) :- ~trace_sub(RT,uk_recurse{_,_},_), ~trace_derefs(RT,0).

escape_edge_follow(es_forward,L,NL) :-
    trace_loc_sub(L,SL,TAIL),
    (escape_edge(true,SL,MSL);
     follow_reverse(TAIL), escape_edge(false,SL,MSL)),
    trace_loc_compose(MSL,TAIL,ML), reduce_trace_loc(ML,NL).

escape_edge_follow(es_backward,L,NL) :-
    trace_loc_sub(L,SL,TAIL),
    (escape_edge(false,SL,MSL);
     follow_reverse(TAIL), escape_edge(true,SL,MSL)),
    trace_loc_compose(MSL,TAIL,ML), reduce_trace_loc(ML,NL).

escape_edge_follow(es_forward_precise{V},L,NL) :-
    trace_loc_sub(L,SL,TAIL),
    (escape_edge(true,SL,MSL);
     follow_reverse(TAIL), escape_edge(false,SL,MSL)),
    trace_loc_compose(MSL,TAIL,BNL),
    precise_reduce_trace_loc(V,BNL,NL).

% try to discharge a forward edge by using trace_loc_any_write
% to see if the forward edge could lead to a future write to
% the interested location
predicate resolve_forward_edge(in V:t_escape_precision,
                               in SL:t_trace_loc,in XL:t_trace_loc,
                               in TAIL:t_trace,NL:t_trace_loc).
resolve_forward_edge(V,SL,XL,TAIL,NL) :-
    trace_loc_any_write(SL,XL,TAIL,MSET), set_member(MSET,ML),
    trace_loc_compose(ML,TAIL,FNL), escape_edge(false,FNL,NL).
resolve_forward_edge(V,SL,XL,TAIL,NL) :-
    ~trace_loc_any_write(SL,XL,TAIL,_),
    trace_loc_compose(XL,TAIL,NL).

escape_edge_follow(es_backward_precise{V},L,NL) :-
    trace_loc_sub(L,SL,TAIL),
    (escape_edge(false,SL,MSL), trace_loc_compose(MSL,TAIL,BNL);
     follow_reverse(TAIL), escape_edge(true,SL,XL),
         resolve_forward_edge(V,SL,XL,TAIL,BNL)),
    precise_reduce_trace_loc(V,BNL,NL).

escape_edge_follow_count(es_backward_precise{_},L,NLC) :-
    trace_loc_sub(L,SL,_), escape_edge_count(false,SL,NLC).
escape_edge_follow_count(es_backward_precise{_},L,NLC) :-
    trace_loc_sub(L,SL,TAIL), follow_reverse(TAIL),
    escape_edge_count(true,SL,NLC).

% OTHER ESCAPE PREDICATES

% PFT is a feasible function pointer trace which could be used to invoke FN
% at call site PI within PFN
predicate feasible_caller(in PFN:string,in PI:c_instr,in FN:string,
                          PFT:t_trace).
feasible_caller(PFN,PI,FN,PFT) :-
    ~escape_set(20,es_forward,tl_glob{root{glob{FN}}},_),
    spoly_call(PFN,PI)->wpoly_call_target(PFT).
feasible_caller(PFN,PI,FN,PFT) :-
    escape_set(20,es_forward,tl_glob{root{glob{FN}}},LSET),
    spoly_call(PFN,PI)->wpoly_call_target(PFT),
    reduce_trace_skolem(PFT,NPFT),
    reduce_trace_loc(tl_func{PFN,NPFT,cxt_any},NL),
    (NL=tl_func{PFN,NPFT,_}, set_member(LSET,tl_func{PFN,NPFT,_});
     NL\=tl_func{_,_,_}, set_member(LSET,NL)).

% BAKED IN

% rewriting parameter with heap data. need flow sensitive locations for this
ignore_escape_edge(false,SL,TL) :-
    SL=tl_func{"fs/open.c:__dentry_open",ST,_},
    (TL=tl_func{"fs/open.c:__dentry_open",_,_};
     TL=tl_comp{"file_operations",_}),
    trace_root(ST,arg{4}).

predicate ignore_escape_function(in FN:string).
ignore_escape_edge(_,tl_func{FN,_,_},_) :- ignore_escape_function(FN).

predicate ignore_escape_field(in C:string,in F:string).
ignore_escape_edge(_,tl_func{_,T,_},_) :- trace_sub(T,fld{_,F,C},_), ignore_escape_field(C,F).
ignore_escape_edge(_,tl_glob{T},_) :- trace_sub(T,fld{_,F,C},_), ignore_escape_field(C,F).

predicate ignore_write_function(in FN:string).
trace_loc_any_write(tl_func{FN,_,_},_,_,NLS) :-
    ignore_write_function(FN), set_empty(NLS).
trace_loc_any_write(_,tl_func{FN,_,_},_,NLS) :-
    ignore_write_function(FN), set_empty(NLS).

predicate ignore_write_field(in C:string,in F:string).
trace_loc_any_write(L,_,_,NLS) :- trace_loc_trace(L,T),
    trace_sub(T,fld{_,F,C},_), ignore_write_field(C,F), set_empty(NLS).
trace_loc_any_write(_,L,_,NLS) :- trace_loc_trace(L,T),
    trace_sub(T,fld{_,F,C},_), ignore_write_field(C,F), set_empty(NLS).
trace_loc_any_write(_,_,T,NLS) :-
    trace_sub(T,fld{_,F,C},_), ignore_write_field(C,F), set_empty(NLS).

+ignore_escape_function("printk").
+ignore_escape_function("vmalloc").
+ignore_escape_function("include/linux/slab.h:kmalloc").
+ignore_escape_function("include/linux/slab.h:kzalloc").
+ignore_escape_function("include/linux/slab.h:kcalloc").
+ignore_escape_function("kmem_cache_alloc").
+ignore_escape_function("kfree").

+ignore_escape_function("include/asm/io.h:ioremap").
+ignore_escape_function("include/asm/io.h:phys_to_virt").
+ignore_escape_function("include/asm/io.h:virt_to_phys").
+ignore_escape_function("copy_from_user").
+ignore_escape_function("ioport_map").

+ignore_escape_function("include/linux/err.h:IS_ERR").
+ignore_escape_function("include/linux/err.h:ERR_PTR").
+ignore_escape_function("fs/bad_inode.c:return_EIO").

% bogus assignments when pointers indicate error codes
+ignore_escape_edge(true,tl_func{"lookup_instantiate_filp",drf{root{arg{1}}},_},_).
+ignore_escape_edge(true,tl_glob{drf{root{glob{"mm/shmem.c:shm_mnt"}}}},tl_func{"shmem_file_setup",_,_}).

% ignoring all assignments involving memcpy/memset. TODO: ought to do better
memcpy_func(FN,_,_,_), +ignore_escape_function(FN).
memset_func(FN,_,_,_), +ignore_escape_function(FN).

% handle imprecise treatment of snd_sb16dsp_get_pcm_ops and snd_ad1848_get_pcm_ops
+ignore_write_function("sound/isa/cmi8330.c:snd_cmi8330_pcm").

% kernel_param's can be used to propagate physical memory addresses
% to drivers, which we aren't interested in. we only get here because
% of analysis imprecision (can't directly deref a physical address anyways)
+ignore_escape_field("kernel_param","arg").
+ignore_escape_field("kparam_array","elem").

% additional physical memory fields which propagation currently breaks on
+ignore_escape_field("resource","start").
+ignore_escape_field("resource","end").

% for polydata (see poly_data_baked.clp; TODO: refactor)
+ignore_callgraph_site("fs/read_write.c:do_readv_writev",_,_).
+ignore_callgraph_site("fs/nfsd/nfsctl.c:nfsctl_transaction_read",_,"fs/nfsd/nfsctl.c:nfsctl_transaction_write").

% functions and fields we aren't precise enough yet for trace_loc_any_write
+ignore_write_field("snd_opl4","proc_entry").
+ignore_write_field("device_attribute",_).
+ignore_write_field("class_device_attribute",_).
+ignore_write_function("snd_info_register").
+ignore_write_function("snd_device_new").
+ignore_write_function("scsi_host_alloc").
