import "escape.clp".
import "../base/sumcallers.clp".
import "memory.clp".

using display.

% RULES

% fill in write edge location information

predicate matching_loc(in L:t_trace_loc,WL:t_trace_loc).
+matching_loc(tl_func{FN,T,_},tl_func{FN,T,cxt_any}).
+matching_loc(tl_glob{T},tl_glob{T}).
matching_loc(tl_func{_,T,_},tl_glob{T}) :- trace_root(T,glob{_}).
matching_loc(tl_func{_,T,_},tl_comp{C,RT}) :-
    trace_sub(T,STR,RT), trace_sub(T,fld{STR,_,C},_).
matching_loc(tl_glob{T},tl_comp{C,RT}) :-
    trace_sub(T,STR,RT), trace_sub(T,fld{STR,_,C},_).

predicate add_write_edge(W:t_escape_where,SL:t_trace_loc,TL:t_trace_loc).
add_write_edge(W,SL,TL), matching_loc(SL,WL),
    +salias_escape(WL,true)->sescape(W,SL,TL).
add_write_edge(W,SL,TL), matching_loc(TL,WL),
    +salias_escape(WL,false)->sescape(W,TL,SL).

% add swrite info for any L where drf{L} is the target of an edge
add_write_edge(W,_,XL), trace_loc_sub(XL,L,drf{empty}),
    matching_loc(L,WL), +salias_used(WL)->swrite(W,L).

% all the reads in the current location. these include expressions other than
% those on the right side of an assignment (e.g. branches, arithmetic subexprs,
% function call exprs)
predicate add_read_info(W:t_escape_where,L:t_trace_loc).

add_read_info(W,XL), trace_loc_sub(XL,L,TAIL),
    trace_sub(TAIL,drf{empty},_), matching_loc(L,WL),
    +salias_used(WL)->sread(W,L).
add_read_info(W,XL), trace_loc_sub(XL,L,TAIL),
    trace_sub(TAIL,rfld{empty,F,C},_), matching_loc(L,WL),
    +salias_used(WL)->srfld(W,L,C,F),
    +salias_rfld(C,F)->srfld(W,L,C,F).

% all read and write accesses in the current location
predicate add_access_info(W:t_escape_where,L:t_trace_loc).
add_read_info(W,L), +add_access_info(W,L).
add_write_edge(W,SL,TL), +add_access_info(W,SL), +add_access_info(W,TL).

add_access_info(W,XL), trace_loc_sub(XL,L,TAIL),
    (trace_sub(TAIL,drf{empty},XTAIL);
     trace_sub(TAIL,drf{uk_index{empty,_}},XTAIL)),
    XTAIL\=empty, matching_loc(L,WL),
    +salias_used(WL)->sdrf(W,L).
add_access_info(W,XL), trace_loc_sub(XL,L,TAIL),
    trace_sub(TAIL,uk_index{empty,_},_), matching_loc(L,WL),
    +salias_used(WL)->sindex(W,L).
add_access_info(W,XL), trace_loc_sub(XL,L,TAIL),
    (trace_sub(TAIL,fld{empty,F,C},_);
     trace_sub(TAIL,fld{uk_index{empty,_},F,C},_)),
    matching_loc(L,WL),
    +salias_used(WL)->sfld(W,L,C,F),
    +salias_fld(C,F)->sfld(W,L,C,F).

% handle rfld assignments, where x = &y->f moves not just y->f to *x,
% but also *y to rfld{*x}
add_write_edge(W,tl_func{SFN,fld{STR,F,C},SCXT},tl_func{TFN,NT,TCXT}),
    trace_simplify(rfld{NT,F,C},NFT),
    +add_write_edge(W,tl_func{SFN,STR,SCXT},tl_func{TFN,NFT,TCXT}).
add_write_edge(W,tl_glob{fld{STR,F,C}},tl_glob{NT}),
    trace_simplify(rfld{NT,F,C},NFT),
    +add_write_edge(W,tl_glob{STR},tl_glob{NFT}).

% ignore write edges involving L
predicate write_edge_ignore(in L:t_trace_loc).

predicate try_write_edge(W:t_escape_where,SL:t_trace_loc,TL:t_trace_loc).
try_write_edge(W,SL,TL), ~write_edge_ignore(SL), ~write_edge_ignore(TL),
    reduce_trace_loc_skolem(SL,NSL), reduce_trace_loc_skolem(TL,NTL),
    +add_write_edge(W,NSL,NTL).

predicate try_read_info(W:t_escape_where,L:t_trace_loc).
try_read_info(W,L), ~write_edge_ignore(L),
    reduce_trace_loc_skolem(L,NL),
    +add_read_info(W,NL).

cil_glob(G), assign(p_abort,LT,s_trace{RT},_),
    +try_write_edge(te_glob{G}, tl_glob{RT},tl_glob{drf{LT}}).
cursum(FN,SUM), iset(P,_,_), assign(P,LT,s_trace{RT},_),
    +try_write_edge(te_func{FN,SUM,P},
                    tl_func{FN,RT,cxt_any},tl_func{FN,drf{LT},cxt_any}).
cursum(FN,SUM), icall(P,_,I), argument(P,I,AT,s_trace{T},_), anycall(I,CFN,_),
    +try_write_edge(te_func{FN,SUM,P},
                    tl_func{FN,T,cxt_any},
                    tl_func{CFN,drf{AT},cxt_call{FN,I,cxt_any}}).
cursum(FN,SUM), icall(P,_,I), callret(P,I,T,_), anycall(I,CFN,_),
    +try_write_edge(te_func{FN,SUM,P},
                    tl_func{CFN,drf{root{return}},cxt_call{FN,I,cxt_any}},
                    tl_func{FN,drf{T},cxt_any}).
cursum(FN,SUM), cil_exp_lval(E,_), exp_point(E,P), eval(P,E,s_trace{T},_),
    +try_read_info(te_func{FN,SUM,P},tl_func{FN,T,cxt_any}).

write_edge_ignore(tl_func{_,T,_}) :- trace_root(T,cstr{_}).
write_edge_ignore(tl_glob{T}) :- trace_root(T,cstr{_}).

% baked-in information

write_edge_ignore(tl_func{_,T,_}) :- trace_sub(T,fld{_,_,"array_cache"},_).
write_edge_ignore(tl_func{FN,_,_}) :- str_cat("mm/slab.c:",_,FN).
