% readonly and non-escaped parameter traces to each function

import "escape.clp".
import "memory.clp".

analyze analysis_name("init_readonly_func").
analyze session_name("cil_body").

% OUTPUT

predicate str_func_readonly(FN:string,T:string).
?- str_func_readonly(FN,T).

predicate str_func_localwrite(FN:string,T:string).
?- str_func_localwrite(FN,T).

predicate str_func_nonescape(FN:string,T:string).
?- str_func_nonescape(FN,T).

% RULES

predicate escape_threshold(TR:int).

predicate func_escape_set(in Y:t_escape_state,in L:t_trace_loc,SET:set[t_trace_loc]).
func_escape_set(Y,L,SET) :- escape_threshold(TR), escape_set(TR,Y,L,SET).
func_escape_set(Y,L,SET) :- ~escape_threshold(_), escape_set(50,Y,L,SET).

predicate transitive_field(in C:string,RT:t_trace).
transitive_field(C,fld{empty,F,C}) :- comp_field(C,_,_,F,Y), Y\=y_comp{_}.
transitive_field(C,RT) :- comp_field(C,_,_,F,y_comp{XC}),
    transitive_field(XC,XRT), trace_compose(fld{empty,F,C},XRT,RT).

predicate external_trace(in T:t_trace).
external_trace(T) :- trace_root(T,glob{_}).
external_trace(T) :- trace_sub(T,uk_recurse{_,_},_).
external_trace(T) :- ~trace_sub(T,uk_recurse{_,_},_),
    trace_derefs(T,N), int_ge(N,2).

predicate not_readonly(in L:t_trace_loc).
not_readonly(L) :- ~func_escape_set(es_forward_precise{ep_nocontext},L,_).
not_readonly(L) :- func_escape_set(es_forward_precise{ep_nocontext},L,NLSET),
    set_member(NLSET,NL), trace_loc_compose(NL,drf{empty},DNL),
    escape_edge(false,DNL,_).

predicate ignore_nonlocal_trace_loc(in L:t_trace_loc).
predicate nonlocal_trace_loc(in L:t_trace_loc).
nonlocal_trace_loc(L) :- ~escape_set(20,es_backward_precise{ep_maximum},L,_).
nonlocal_trace_loc(L) :- escape_set(20,es_backward_precise{ep_maximum},L,PLSET),
    set_member(PLSET,tl_func{_,T,_}), external_trace(T).

predicate not_localwrite(in L:t_trace_loc).
not_localwrite(L) :- ~not_readonly(L).
not_localwrite(L) :- not_readonly(L),
    ~func_escape_set(es_forward_precise{ep_nocontext},L,_).
not_localwrite(L) :- not_readonly(L),
    func_escape_set(es_forward_precise{ep_nocontext},L,NLSET),
    set_member(NLSET,NL), trace_loc_compose(NL,drf{empty},DNL),
    escape_edge(false,DNL,PDNL),
    ~ignore_nonlocal_trace_loc(PDNL), nonlocal_trace_loc(PDNL).

predicate ignore_escaped_trace_loc(in L:t_trace_loc).
predicate escaped_trace_loc(in L:t_trace_loc).
escaped_trace_loc(L) :- ~escape_set(20,es_forward_precise{ep_nocontext},L,_).
escaped_trace_loc(L) :- escape_set(20,es_forward_precise{ep_nocontext},L,NLSET),
    set_member(NLSET,tl_func{_,T,_}), external_trace(T).

predicate not_nonescape(in L:t_trace_loc).
not_nonescape(L) :- not_readonly(L).
not_nonescape(L) :- ~not_readonly(L),
    func_escape_set(es_forward_precise{ep_nocontext},L,NLSET),
    set_member(NLSET,NL), trace_loc_compose(NL,drf{empty},DNL),
    escape_edge(true,DNL,NDNL),
    ~ignore_escaped_trace_loc(NDNL), escaped_trace_loc(NDNL).

predicate test_trace(T:t_trace).
predicate filter_trace(in T:t_trace).
+filter_trace(_).

cil_curfn(FN), test_trace(T), filter_trace(T), T\=root{arg{_}},
    ~not_readonly(tl_func{FN,T,cxt_any}),
    +salias_func_readonly(FN)->wfunc_readonly(T),
    trace_string(T,TS), +str_func_readonly(FN,TS).
cil_curfn(FN), test_trace(T), filter_trace(T), T\=root{arg{_}}, trace_type(FN,T,y_ptr{_}),
    ~not_localwrite(tl_func{FN,T,cxt_any}),
    +salias_func_readonly(FN)->wfunc_localwrite(T),
    trace_string(T,TS), +str_func_localwrite(FN,TS).
cil_curfn(FN), test_trace(T), filter_trace(T), trace_type(FN,T,y_ptr{_}),
    ~not_nonescape(tl_func{FN,T,cxt_any}),
    +salias_func_readonly(FN)->wfunc_nonescape(T),
    trace_string(T,TS), +str_func_nonescape(FN,TS).

cil_curfn(FN), var_trace(_,arg{A}), trace_type(FN,root{arg{A}},y_ptr{_}),
    +test_trace(root{arg{A}}).
cil_curfn(FN), var_trace(_,arg{A}),
    trace_type(FN,root{arg{A}},y_ptr{y_comp{C}}), AT=drf{root{arg{A}}},
    func_escape_set(es_forward_precise{ep_nocontext},tl_func{FN,AT,cxt_any},ASET),
    transitive_field(C,RT), trace_compose(AT,RT,T), +test_trace(T),
    set_member(ASET,XL), salias_used(XL)->sindex(_,_),
    trace_compose(uk_index{AT,y_comp{C}},RT,IT), +test_trace(IT).
cil_curfn(FN), var_trace(_,arg{A}),
    trace_type(FN,root{arg{A}},y_ptr{y_ptr{_}}),
    +test_trace(drf{root{arg{A}}}).

% BAKED

% increase the escape threshold when we're examining parts of the VFS
% create path. this stuff involves lots of functions
cil_curfn(FN), var_trace(_,R), trace_type(FN,root{R},Y),
    (Y=y_comp{"nameidata"}; Y=y_ptr{y_comp{"nameidata"}}),
    +escape_threshold(300).

% return assignment of __dentry_open here is assigning either the same file pointer
% or a PTR_ERR error code
+ignore_nonlocal_trace_loc(tl_func{"lookup_instantiate_filp",drf{root{temp{_,"__dentry_open",_}}},_}).

% restoring earlier (but still fresh) value of file pointer
+ignore_nonlocal_trace_loc(tl_func{"link_path_walk",drf{fld{_,"file","open_intent"}},_}).
