% initialization information computed for each structure field

import "escape.clp".
import "memory.clp".

analyze analysis_name("init_relative_comp").
analyze session_name("cil_body").

% OUTPUT

predicate str_comp_nonrelative(C:string,F:string,WL:string).
?- str_comp_nonrelative(C,F,WL).

predicate str_comp_relative(C:string,F:string,L:string,WL:string).
?- str_comp_relative(C,F,L,WL).

% RULES

% more precise version of follow_relative_access_candidate,
% that can perform its own subsidiary escapes to find all the possible
% relative_access_candidate's that L could correspond to. this is needed
% in some complex initialization cases, e.g. tcp_proc_register for
% file_operations fields
predicate precise_relative_access_candidate(in L:t_trace_loc,
                                            NLSET:set[t_trace_loc]).
precise_relative_access_candidate(L,NLSET) :-
    follow_relative_access_candidate(L,NLSET).
precise_relative_access_candidate(L,NLSET) :-
    ~follow_relative_access_candidate(L,_),
    escape_set(30,es_backward_precise{ep_maximum},L,XLSET),
    \/(set_member(XLSET,XL), relative_access_candidate(XL)):set_all(XL,NLSET).

predicate check_local_write(T:t_trace,C:string,F:string).
iset(P,_,_), assign(P,fld{T,F,C},s_trace{_},_), +check_local_write(T,C,F).
icall(P,_,I), callret(P,I,fld{T,F,C},_), +check_local_write(T,C,F).

cil_curfn(FN), check_local_write(T,C,F),
    reduce_trace_loc_skolem(tl_func{FN,T,cxt_any},WL),
    reduce_trace_loc(WL,MWL),
    ~precise_relative_access_candidate(MWL,_),
    +salias_comp_relative(C,F)->wcomp_nonrelative(WL),
    trace_loc_string(WL,WLS),
    +str_comp_nonrelative(C,F,WLS).

cil_curfn(FN), check_local_write(T,C,F),
    reduce_trace_loc_skolem(tl_func{FN,T,cxt_any},WL),
    reduce_trace_loc(WL,MWL),
    precise_relative_access_candidate(MWL,LSET), set_member(LSET,L),
    +salias_comp_relative(C,F)->wcomp_relative(L,WL),
    trace_loc_string(L,LS), trace_loc_string(WL,WLS),
    +str_comp_relative(C,F,LS,WLS).
