% initialization information computed for each structure field

import "escape.clp".
import "memory.clp".

analyze analysis_name("init_usemod_comp").
analyze session_name("cil_body").

% OUTPUT

predicate str_comp_impure(C:string,F:string,FN:string,T:string).
?- str_comp_impure(C,F,FN,T).

predicate str_comp_init(C:string,F:string,FN:string,T:string).
?- str_comp_init(C,F,FN,T).

% RULES

predicate escape_threshold(TR:int).
predicate use_threshold(TR:int).
~use_threshold(_), +escape_threshold(200).
use_threshold(TR), +escape_threshold(TR).

predicate check_local_write(T:t_trace,C:string,F:string).
iset(P,_,_), assign(P,fld{T,F,C},_,_),
    reduce_trace_skolem(T,NT), +check_local_write(NT,C,F).
icall(P,_,I), callret(P,I,fld{T,F,C},_),
    reduce_trace_skolem(T,NT), +check_local_write(NT,C,F).

cil_curfn(FN), escape_threshold(TR), check_local_write(T,_,_),
    +try_escape(TR,es_backward_precise{ep_maximum},tl_func{FN,T,cxt_any}).

predicate nonescape_trace(in FN:string,in T:t_trace).
nonescape_trace(FN,drf{AT}) :-
    salias_func_readonly(FN)->wfunc_readonly(AT).
nonescape_trace(FN,drf{AT}) :-
    salias_func_readonly(FN)->wfunc_localwrite(AT).

predicate escape_nonlocal_loc(in L:t_trace_loc).
escape_nonlocal_loc(L) :-
    trace_loc_sub(L,SL,TAIL), follow_reverse(TAIL),
    escape_edge(true,SL,NSL), trace_loc_compose(NSL,TAIL,NL),
    NL=tl_func{FN,T,_}, ~nonescape_trace(FN,T).
escape_nonlocal_loc(L) :-
    trace_loc_trace(L,T), trace_root(T,glob{_}).

predicate impure_write(T:t_trace,C:string,F:string).
cil_curfn(FN), escape_threshold(TR), check_local_write(T,C,F),
    escape_fail(TR,es_backward_precise{ep_maximum},tl_func{FN,T,cxt_any}),
    +impure_write(T,C,F).
cil_curfn(FN), escape_threshold(TR), check_local_write(T,C,F),
    ~escape_fail(TR,es_backward_precise{ep_maximum},tl_func{FN,T,cxt_any}),
    escape(TR,es_backward_precise{ep_maximum},tl_func{FN,T,cxt_any},XL),
    escape_nonlocal_loc(XL), +impure_write(T,C,F).

cil_curfn(FN), impure_write(T,C,F),
    +salias_write_comp(C,F)->wcomp_impure(FN),
    trace_string(T,TS), +str_comp_impure(C,F,FN,TS).
cil_curfn(FN), escape_threshold(TR), check_local_write(T,C,F),
    ~impure_write(T,C,F),
    escape(TR,es_backward_precise{ep_maximum},tl_func{FN,T,cxt_any},XL),
    XL=tl_func{XFN,XT,cxt_any}, trace_root(XT,arg{_}),
    +salias_write_comp(C,F)->wcomp_init(XFN,XT),
    trace_string(XT,XTS), +str_comp_init(C,F,XFN,XTS).
