% eagerly computed per-function aliasing/escaping summary information

import "aliasing_sum.clp".
import "../smemory/memory.clp".
import "../smemory/aliasing_full.clp".
import "../smemory/scalar_sat.clp".

analyze analysis_name("init_usemod_func").
analyze session_name("cil_sum_body").

% OUTPUT

predicate str_write_inline(FN:string,SUM:sum,T:string,V:string).
?- str_write_inline(FN,SUM,T,V).

predicate str_allocate(FN:string,T:string,ESCAPE:bool).
?- str_allocate(FN,T,ESCAPE).

% RULES

cursum(FN,SUM), +salias_write_func(FN,SUM)->clear_old_preds().

% infer relatively pure function/sums and type fields

predicate add_func_impure().
predicate add_write(T:t_trace).

predicate check_write(T:t_trace).
(iset(P,_,_); isum(P,_,_)), assign(P,T,_,_), +check_write(T).
isum(P,_,I), isum_target(I,CFN,CSUM,direct),
    salias_write_func(CFN,CSUM)->wfunc_write(CT),
    inst_trace(I,P,CT,s_trace{T},_), +check_write(T),
    cursum(CFN,CSUM), CT\=T, +add_func_impure().
isum(_,_,I), isum_target(I,CFN,CSUM,indirect),
    salias_write_func(CFN,CSUM)->wfunc_write(_),
    +add_func_impure().
isum(_,_,I), isum_target(I,CFN,CSUM,_),
    salias_write_func(CFN,CSUM)->wfunc_impure(), +add_func_impure().

check_write(T), cursum(_,SUM), trace_alloc(SUM,T,ak_external),
    +add_write(T).
check_write(T), cursum(_,SUM), ~trace_alloc(SUM,T,ak_external),
    ~trace_alloc(SUM,T,ak_stack_local),
    ~(trace_fld_sub(T,XT,_), allocate_root(XT,_,_,_)),
    +add_func_impure().

predicate omit_write(T:t_trace).
omit_write(_), +add_func_impure().

add_write(T), trace_root(T,glob{_}), +omit_write(T).
add_write(T), trace_derefs(T,N), int_gt(N,2), +omit_write(T).
add_write(T), trace_sub(T,index{_,_,_},_), +omit_write(T).
add_write(T), trace_sub(T,rfld{_,_,_},_), +omit_write(T).
add_write(T), trace_sub(T,fld{ST,_,C},_), trace_sub(ST,fld{_,_,C},_), +omit_write(T).
add_write(T), trace_sub(T,fld{ST,_,C},_),
    cil_curfn(FN), ~trace_type(FN,ST,y_comp{C}), +omit_write(T).

cursum(FN,SUM), add_func_impure(), +salias_write_func(FN,SUM)->wfunc_impure().
cursum(FN,SUM), add_write(T), ~omit_write(T), ~trace_root(T,cstr{_}),
    +salias_write_func(FN,SUM)->wfunc_write(T).

% infer writes which should be inlined

% try to add inline information for writes on T
predicate try_write_inline(T:t_trace).
cursum(FN,SUM), try_write_inline(T),
    sum_bound(SUM,_,P), val(P,T,V,_),
    ~(sum_bound(SUM,_,XP), val(XP,T,XV,_), V\=XV),
    ~(scalar_used(V,su_scalar{s_trace{XT}}),
      ~trace_alloc(s_func,XT,ak_external)),
    +salias_write_func(FN,SUM)->wfunc_write_inline(T,V),
    trace_string(T,TS), scalar_string(V,VS), +str_write_inline(FN,SUM,TS,VS).

% try for exact return values
iset(P,_,_), assign(P,root{return},_,_), +try_write_inline(root{return}).

% try for assigns of arguments to particular heap fields:
% foo(a,b) { a->f = b; }
(iset(P,_,_); isum(P,_,_)), assign(P,T,s_trace{drf{root{arg{_}}}},_),
    trace_root(T,arg{_}), +try_write_inline(T).

% infer allocator functions

predicate allocated_data(in V:scalar,in G:g_guard).
+allocated_data(s_const{_},_).
allocated_data(s_trace{AT},_) :- allocate_root(AT,_,_,_).
allocated_data(s_trace{T},G) :- error_trace(T,XG), guard_implies(G,XG).

cursum(FN,s_func), exit(EXITP), guard(EXITP,XG),
    (iset(P,_,_); isum(P,_,_)), assign(P,T,s_trace{AT},_),
    trace_alloc(s_func,T,ak_external), allocate_root(AT,_,_,_),
    ~(val(EXITP,T,V,TG), #and(TG,XG,G), guard_sat(G), ~allocated_data(V,G)),
    (allocate_escape(AT,EXITP,T), ESCAPE=true;
     ~allocate_escape(AT,EXITP,T), ESCAPE=false),
    +salias_write_func(FN,s_func)->wfunc_allocate(T,ESCAPE),
    trace_string(T,TS), +str_allocate(FN,TS,ESCAPE).

% infer functions making indirect calls

cursum(FN,SUM), isum(_,_,I), isum_target(I,CFN,CSUM,_),
    salias_write_func(CFN,CSUM)->wfunc_indirect(),
    +salias_write_func(FN,SUM)->wfunc_indirect().

cursum(FN,SUM), icall(_,_,I), ~dircall(I,_),
    +salias_write_func(FN,SUM)->wfunc_indirect().
