import "aliasing_sum.clp".
import "zerofield.clp".

import "../smemory/memory.clp".
import "../smemory/scalar_sat.clp".
import "../smemory/aliasing_full.clp".
import "../base/sumcallers.clp".

import "escape.clp".

% types

% text output

predicate str_inline_assignment(FN:string,T:string,V:string).
?- str_inline_assignment(FN,T,V).

predicate add_str_failed(C:string,FT:string,T:string,AT:string).
predicate add_str_warning(C:string,FT:string,T:string,AT:string,FNPTR:string).
predicate add_str_received(FNPTR:string,AT:string,
                           DFN:string,DSUM:sum,DP:pp,DT:string).

% writes affecting poly

predicate gassign(P:pp,T:t_trace,V:t_trace,G:g_guard).

% for now the only assignments we inline are those we have annotated as such
predicate inline_assignment(in T:t_trace,in V:t_trace).
%inline_assignment(T,_) :- cursum(FN,s_func),
%    salias_write_func(FN,s_func)->wfunc_write_inline(T,_),
%    ~(function_caller(FN,PFN,PI), sum_funptr(PFN)->sindirect(PI,_)).

predicate add_inline_assignment(T:t_trace,V:t_trace).
cil_curfn(FN), add_inline_assignment(T,V),
    trace_string(T,TS), trace_string(V,VS),
    +str_inline_assignment(FN,TS,VS),
    +spoly_func(FN)->wpoly_write_func(T,V).

% add assignment of V to T if V is interesting (not NULL).
% TODO: filtering out breakage on partial struct initializers,
% e.g. request_firmware_nowait, should fix this upstream
predicate add_gassign(P:pp,T:t_trace,V:scalar,G:g_guard).
add_gassign(P,T,s_trace{V},VG), ~trace_root(V,local{_}),
    ~inline_assignment(T,V),
    guard(P,PG), #and(PG,VG,G), +gassign(P,T,V,G).
add_gassign(_,T,s_trace{V},_), ~trace_root(V,local{_}),
    inline_assignment(T,V),
    +add_inline_assignment(T,V).
iset(P,_,_), assign(P,T,V,G), +add_gassign(P,T,V,G).
assign(p_abort,T,V,G), +add_gassign(p_abort,T,V,G).  % for static initializers

% add assignments which we are inlining from callees
icall(P,_,I), anycall(I,CFN,_), spoly_func(CFN)->wpoly_write_func(CT,CV),
    inst_trace(s_call{I},P,CT,s_trace{T},TG),
    inst_trace(s_call{I},P,CV,V,VG),
    #and(TG,VG,G), +add_gassign(P,T,V,G).

% propagation assignments/failure

% ignore failures and any receive's on C/FT/T/AT
predicate ignore_failure(in C:string,in FT:t_trace,in T:t_trace,in AT:t_trace).

% filter out all assignments except those for C/FT/T
predicate debug_filter(C:string,FT:string,T:string).
%+debug_filter("snd_kcontrol",".info*",".private_data*").
%+debug_filter("file",".f_op*.release*",".private_data*").
%+debug_filter("file",".f_op*.write*",".private_data*").
%+debug_filter("snd_pcm_substream",".ops*.prepare*",".runtime*.private_data*").
%+debug_filter("snd_pcm_substream",".ops*.close*",".runtime*.private_data*").
%+debug_filter("urb",".complete*",".context*").
%+debug_filter("proc_dir_entry",".read_proc*",".data*").
%+debug_filter("vm_area_struct",".vm_ops*.nopage*",".vm_private_data*").
%+debug_filter("request_queue",".make_request_fn*",".queuedata*").
%+debug_filter("usb_interface",".dev.driver*^driver.disconnect*",".dev.driver_data*").
%+debug_filter("dvb_frontend",".ops*.set_tone*",".dvb*.priv*").
%+debug_filter("usb_serial_port",".serial*.type*.close*",".dev.driver_data*").
%+debug_filter("fb_info",".fbops*.fb_setcmap*",".par*").
%+debug_filter("irq_req_t",".Handler*",".Instance*").
%+debug_filter("input_dev",".close*",".private*").
%+debug_filter("atm_dev",".ops*.phy_get*",".dev_data*").
%+debug_filter("snd_ac97",".bus*.ops*.write*",".private_data*").
%+debug_filter("bio",".bi_end_io*",".bi_private*").
%+debug_filter("hotplug_slot",".ops*.enable_slot*",".private*").
%+debug_filter("ac97_codec",".codec_write*",".private_data*").
%+debug_filter("tty_struct",".ldisc.close*",".disc_data*").
%+debug_filter("snd_hwdep",".ops.ioctl*",".private_data*").
%+debug_filter("drm_device",".driver*.unload*",".dev_private*").

% an update to XSTR relative via RT to some base structure C may affect
% the associations between FT and T relative to C
predicate consider_update(XSTR:t_trace,RT:t_trace,
                          C:string,FT:t_trace,T:t_trace,AT:t_trace,
                          P:pp,WT:t_trace,WV:t_trace,WG:g_guard).
gassign(P,fld{XT,XF,XC},V,G),
    spoly_comp(XC,XF)->wpoly_search_comp(C,RT,FT,T,AT), guard_sat(G),
    ~ignore_failure(C,FT,T,AT),
    (~debug_filter(_,_,_);
     debug_filter(C,FTS,TS), trace_string(FT,FTS), trace_string(T,TS)),
    +consider_update(XT,RT,C,FT,T,AT,P,fld{XT,XF,XC},V,G).

predicate add_failed(C:string,FT:t_trace,T:t_trace,AT:t_trace).
add_failed(C,FT,T,AT),
    trace_string(FT,FTS), trace_string(T,TS), trace_string(AT,ATS),
    +add_str_failed(C,FTS,TS,ATS),
    +spoly_receive_comp(C,FT,T,AT)->wpoly_fail().

predicate base_structure(in XT:t_trace,in RT:t_trace,STR:t_trace).
predicate base_structure_fail(in P:pp,in XT:t_trace,in RT:t_trace).

base_structure(XT,RT,STR) :- trace_sub(XT,STR,RT).
base_structure_fail(P,XT,RT) :- ~trace_sub(XT,_,RT),
    cursum(FN,SUM), trace_alias(FN,SUM,XT,RT,_),
    ~(allocate_root(XT,_,_,_), ~allocate_escape(XT,P,empty)).
base_structure_fail(_,XT,RT) :- trace_sub(XT,_,RT), ~trace_derefs(RT,0).

predicate omit_base_structure_fail(in C:string,in FT:t_trace,in T:t_trace,
                                   in AT:t_trace,in UT:t_trace).

~cil_glob(_), consider_update(XT,RT,C,FT,T,AT,P,UT,_,_),
    base_structure_fail(P,XT,RT), ~omit_base_structure_fail(C,FT,T,AT,UT),
    +add_failed(C,FT,T,AT).

% the assignment from XV to XT is one that must be considered for
% the poly correlation RFT/RT/AT relative to STR
predicate poly_assign(STR:t_trace,RFT:t_trace,RT:t_trace,AT:t_trace,
                      P:pp,XT:t_trace,XV:t_trace,XG:g_guard).
consider_update(XT,RT,_,FT,T,AT,P,WT,WV,WG),
    base_structure(XT,RT,STR),
    +poly_assign(STR,FT,T,AT,P,WT,WV,WG).

% make assignments to fake fields for handling calls
predicate func_fake_comp(CFN:string,CFT:t_trace,CT:t_trace,
                         STR:t_trace,RFT:t_trace,RT:t_trace).
dircall(I,CFN), icall(P,_,I), ~debug_filter(_,_,_),
    spoly_func(CFN)->wpoly_search_func(FT,T,AT),
    tostring(I,IS), trace_string(FT,FTS), trace_string(T,TS),
    GSTR = CFN^"$"^IS^"$"^FTS^"$"^TS, STR = root{glob{GSTR}},
    RFT=fld{empty,"$fnptr","$"^GSTR}, trace_compose(STR,RFT,NFT),
    RT=fld{empty,"$fnval","$"^GSTR}, trace_compose(STR,RT,NT),
    +func_fake_comp(CFN,FT,T,STR,drf{RFT},drf{RT}),
    inst_trace(s_call{I},P,FT,FTV,IG0), +add_gassign(P,NFT,FTV,IG0),
    inst_trace(s_call{I},P,T,TV,IG1), +add_gassign(P,NT,TV,IG1),
    (XT=NFT; XT=NT), gassign(P,XT,XV,XG),
    +poly_assign(STR,drf{RFT},drf{RT},AT,P,XT,XV,XG).

predicate ignore_func_failure(in CFN:string,in CFT:t_trace,
                              in CT:t_trace,in AT:t_trace).
ignore_failure(_,FT,T,AT) :- func_fake_comp(CFN,CFT,CT,_,FT,T),
    ignore_func_failure(CFN,CFT,CT,AT).

% assignment effects

% get the value of RT relative to STR after any assignments at P.
% this is similar to convert_point_trace, except that the value of STR
% is preserved (it is the canonical name), and gassign assignments at P
% are taken into account
predicate post_assign_value(in P:pp,in STR:t_trace,in RT:t_trace,
                            NT:t_trace,G:g_guard).

post_assign_value(_,STR,empty,STR,G) :- #bool_g(true,G).
post_assign_value(P,STR,drf{RT},NT,G) :-
    post_assign_value(P,STR,RT,NRT,G0),
    add_gassign(P,NRT,s_trace{NT},G1), #and(G0,G1,G).
post_assign_value(P,STR,drf{RT},NT,G) :-
    post_assign_value(P,STR,RT,NRT,G0),
    \/add_gassign(P,NRT,_,XG):#or_all(XG,MG), #not(MG,G1), ~#bool_g(false,G1),
    val(P,NRT,s_trace{NT},G2), #and3(G0,G1,G2,G).
post_assign_value(P,STR,fld{RT,F,C},NT,G) :-
    post_assign_value(P,STR,RT,NRT,G), trace_simplify(fld{NRT,F,C},NT).
post_assign_value(P,STR,rfld{RT,F,C},NT,G) :-
    post_assign_value(P,STR,RT,NRT,G), trace_simplify(rfld{NRT,F,C},NT).
post_assign_value(P,STR,index{RT,Y,I},NT,G) :-
    post_assign_value(P,STR,RT,NRT,G), trace_simplify(index{NRT,Y,I},NT).

?post_assign_value(_,STR,index{RT,_,I},_,_), I\=s_const{_},
    +warning("post_assign_value used with non-constant index",STR,RT,I).

% resolving assignments

% there is a callee along some path from P to a point in XPSET which might
% observe the value of a relationship between two traces; either it has
% an external side effect, or it makes an indirect call
predicate maybe_observe(in P:pp,in XPSET:set[pp]).
maybe_observe(P,XPSET) :- pp_reach(P,WP), isum(WP,_,I),
    set_member(XPSET,MXP), pp_reach(WP,MXP), isum_target(I,CFN,CSUM,_),
    (salias_write_func(CFN,CSUM)->wfunc_impure();
     salias_write_func(CFN,CSUM)->wfunc_indirect();
     salias_write_func(CFN,CSUM)->wfunc_write(fld{_,_,_})).

% write to one fnptr/data field is definitely followed by a write
% to another fnptr/data field for the same property.
% we can ignore the first one and just look at the second one.
predicate ignore_poly_assign(STR:t_trace,FT:t_trace,T:t_trace,AT:t_trace,
                             P:pp,XT:t_trace,XV:t_trace,XG:g_guard).
poly_assign(STR,FT,T,AT,P0,XT0,XV0,XG0),
    poly_assign(STR,FT,T,AT,_,XT1,_,_),
    \/(poly_assign(STR,FT,T,AT,XP,XT1,_,XG), pp_reach(P0,XP)):#or_all(XG,MXG),
    guard_implies(XG0,MXG),
    \/(poly_assign(STR,FT,T,AT,YP,XT1,_,_), pp_reach(P0,YP)):set_all(YP,YPSET),
    ~maybe_observe(P0,YPSET),
    +ignore_poly_assign(STR,FT,T,AT,P0,XT0,XV0,XG0).
poly_assign(STR,FT,T,AT,P0,XT0,XV0,XG0),
    poly_assign(STR,FT,T,AT,P0,XT1,_,XG1), lessthan(XT0,XT1),
    guard_implies(XG0,XG1),
    +ignore_poly_assign(STR,FT,T,AT,P0,XT0,XV0,XG0).

% for baked-in behaviors on particular function/value traces
predicate ignore_assign_value(in FT:t_trace,in T:t_trace).
predicate transform_assign_value(in FT:t_trace,in T:t_trace,
                                 NFT:t_trace,NT:t_trace).

% NFT/NT is a relationship originating from a poly_assign
% for which we need to consider the possible FNPTR/VALUE correlations.
predicate poly_assign_value(STR:t_trace,FT:t_trace,T:t_trace,AT:t_trace,
                            NFT:t_trace,NT:t_trace,P:pp).
poly_assign(STR,FT,T,AT,P,XT,XV,XG),
    ~ignore_poly_assign(STR,FT,T,AT,P,XT,XV,XG),
    post_assign_value(P,STR,FT,MFT,FG),
    post_assign_value(P,STR,T,MT,TG),
    #and3(XG,FG,TG,G), guard_sat(G),
    ~zero_initialized(P,MFT), ~zero_initialized(P,MT),
    ~ignore_assign_value(MFT,MT),
    (transform_assign_value(MFT,MT,NFT,NT);
     ~transform_assign_value(MFT,MT,_,_), NFT=MFT, NT=MT),
    +poly_assign_value(STR,FT,T,AT,NFT,NT,P).

predicate resolve_value(FT:t_trace,T:t_trace,AT:t_trace,P:pp).
predicate resolve_value_success(FT:t_trace,T:t_trace,AT:t_trace).
predicate resolve_value_receive(FT:t_trace,T:t_trace,AT:t_trace,
                                FNPTR:string,
                                WFN:string,WSUM:sum,WP:pp,WT:t_trace).

poly_assign_value(_,_,_,AT,NFT,NT,P), +resolve_value(NFT,NT,AT,P).

poly_assign_value(STR,FT,T,AT,NFT,NT,_), ~resolve_value_success(NFT,NT,AT),
    trace_sub(FT,fld{empty,_,C},_), +add_failed(C,FT,T,AT),
    func_fake_comp(CFN,CFT,CT,STR,FT,T),
    ~ignore_func_failure(CFN,CFT,CT,AT),
    +spoly_receive_func(CFN,CFT,CT,AT)->wpoly_fail().

poly_assign_value(STR,FT,T,AT,NFT,NT,_), resolve_value_success(NFT,NT,AT),
    resolve_value_receive(NFT,NT,AT,FNPTR,WFN,WSUM,WP,WT),
    trace_string(WT,WTS), trace_string(AT,ATS),
    +add_str_received(FNPTR,ATS,WFN,WSUM,WP,WTS),
    trace_sub(FT,fld{empty,_,C},_),
    +spoly_receive_comp(C,FT,T,AT)->wpoly_receive(FNPTR,WFN,WSUM,WP,WT),
    func_fake_comp(CFN,CFT,CT,STR,FT,T),
    +spoly_receive_func(CFN,CFT,CT,AT)->wpoly_receive(FNPTR,WFN,WSUM,WP,WT).

predicate str_resolve_value(FT:string,T:string,AT:string).
predicate str_resolve_value_fail(FT:string,T:string,AT:string).
%?- str_resolve_value(FT,T,AT).
%?- str_resolve_value_fail(FT,T,AT).
resolve_value(FT,T,AT,_),
    trace_string(FT,FTS), trace_string(T,TS), trace_string(AT,ATS),
    +str_resolve_value(FTS,TS,ATS),
    ~resolve_value_success(FT,T,AT),
    +str_resolve_value_fail(FTS,TS,ATS).

predicate conflicting_type(in FNPTR:string,in AT:t_trace,
                           in XFN:string,in XT:t_trace).
conflicting_type(FNPTR,AT,XFN,XT) :-
    spoly_call_try(FNPTR)->wpoly_call_try(AT,y_comp{EC}),
    trace_type(XFN,XT,y_comp{XC}), XC\=EC.

poly_assign_value(_,FT,T,AT,NFT,NT,_), resolve_value_success(NFT,NT,AT),
    resolve_value_receive(NFT,NT,AT,FNPTR,WFN,_,_,WT),
    conflicting_type(FNPTR,AT,WFN,WT),
    trace_sub(FT,fld{empty,_,C},_),
    trace_string(FT,FTS), trace_string(T,TS), trace_string(AT,ATS),
    +add_str_warning(C,FTS,TS,ATS,FNPTR).

% resolving patterns - transitive correlations

% found a transitive correlation for FT/T, do not directly look for targets
predicate correlate_write(in FT:t_trace,in T:t_trace,in AT:t_trace).

predicate omit_propagate_comp(in C:string,in FT:t_trace,in T:t_trace,
                              in AT:t_trace).

% correlations between FT/T should be tracked as transitive correlations
% relative to type STR. for now this is straight fields of a type STR,
% except in cases where we have added manual annotations
predicate try_propagate_comp(in FT:t_trace,in T:t_trace,in AT:t_trace,
                             C:string,RFT:t_trace,RT:t_trace).
try_propagate_comp(BFT,BT,AT,C,RFT,RT) :-
    trace_remove_uc_sum(BFT,FT), trace_remove_uc_sum(BT,T),
    trace_sub(FT,STR,RFT), trace_sub(T,STR,RT),
    ~(trace_sub(FT,XSTR,_), trace_sub(T,XSTR,_),
      XSTR\=STR, trace_sub(XSTR,STR,_)),
    (trace_sub(RFT,fld{empty,_,C},_), trace_sub(RT,fld{empty,_,C},_),
         NRFT=RFT, NRT=RT;
     trace_sub(RFT,rfld{empty,WF,C},NRFT),
         trace_compose(fld{empty,WF,C},RT,NRT)),
    trace_sub(NRFT,fld{empty,_,C},_), trace_sub(NRT,fld{empty,_,C},_),
    ~poly_candidate_ignore(NRFT), ~poly_candidate_ignore(NRT),
    ~omit_propagate_comp(C,NRFT,NRT,AT),
    ~spoly_receive_comp(C,NRFT,NRT,AT)->wpoly_fail().
correlate_write(FT,T,AT) :- try_propagate_comp(FT,T,AT,_,_,_).

% the funptr and data are both written with data passed in from another struct.
% tell the type about the association and pick up any new associations it finds
resolve_value(FT,T,AT,_), try_propagate_comp(FT,T,AT,C,RFT,RT),
    (XT=RFT; XT=RT), trace_sub(XT,drf{fld{STR,XF,XC}},_),
    +spoly_comp(XC,XF)->wpoly_search_comp(C,STR,RFT,RT,AT),
    +resolve_value_success(FT,T,AT),
    spoly_receive_comp(C,RFT,RT,AT)->wpoly_receive(FNPTR,WFN,WSUM,WP,WT),
    +resolve_value_receive(FT,T,AT,FNPTR,WFN,WSUM,WP,WT).

% T at entry to FN/SUM is the same value as T at entry to FN/s_func
predicate equals_function_entry(in FN:string,in SUM:sum,in T:t_trace).
+equals_function_entry(_,s_func,_).
equals_function_entry(FN,SUM,T) :- SUM=s_loop{_},
    cil_sum_caller(FN,SUM)->caller(FN,PSUM,PP,PI),
    ~(inst_trace(PI,p_remote{FN,PSUM,PP},T,s_trace{XT},_), XT\=T),
    equals_function_entry(FN,PSUM,T).

% correlations between FT/T should be pushed up to the caller.
% for now this is just straight arguments, except in cases where we have
% added manual annotations
predicate try_propagate_caller(in FT:t_trace,in T:t_trace,in AT:t_trace).
try_propagate_caller(FT,T,AT) :- cursum(FN,SUM),
    equals_function_entry(FN,SUM,FT), equals_function_entry(FN,SUM,T),
    ~(function_caller(FN,PFN,PI), sum_funptr(PFN)->sindirect(PI,FN)),
    trace_root(FT,arg{FA}), trace_root(T,arg{TA}), FA\=TA,
    ~poly_candidate_ignore(FT), ~poly_candidate_ignore(T),
    ~spoly_receive_func(FN,FT,T,AT)->wpoly_fail().
correlate_write(FT,T,AT) :- try_propagate_caller(FT,T,AT).

% the funptr and data are both written with data passed in from a caller.
% tell the caller about the association and pick up any associations it finds
cursum(FN,_), resolve_value(FT,T,AT,_), try_propagate_caller(FT,T,AT),
    +spoly_func(FN)->wpoly_search_func(FT,T,AT),
    +resolve_value_success(FT,T,AT),
    spoly_receive_func(FN,FT,T,AT)->wpoly_receive(FNPTR,WFN,WSUM,WP,WT),
    +resolve_value_receive(FT,T,AT,FNPTR,WFN,WSUM,WP,WT).

% resolving patterns - particular functions or sets of functions

predicate extra_reduce_trace_loc(in FN:string,in T:t_trace,NL:t_trace_loc).
extra_reduce_trace_loc(FN,T,NL) :-
    (trace_root(T,glob{_}), NL=tl_glob{T};
     ~trace_root(T,glob{_}), NL=tl_func{FN,T,cxt_any}).

predicate funptr_escape(FT:t_trace,FNPTRS:set[string]).
predicate funptr_escape_threshold(TR:int).

predicate try_funptr_escape(FT:t_trace).
(funptr_escape_threshold(TR); ~funptr_escape_threshold(_), TR=100),
    try_funptr_escape(FT), cil_curfn(FN), extra_reduce_trace_loc(FN,FT,L),
    (LVL=ep_field{1}; LVL=ep_field{2}; LVL=ep_field{4}),
    +try_escape(TR,es_backward_precise{LVL},L),
    ~escape_fail(TR,es_backward_precise{LVL},L),
    \/escape(TR,es_backward_precise{LVL},L,
             tl_glob{root{glob{FNPTR}}}):set_all(FNPTR,FNPTRS),
    +funptr_escape(FT,FNPTRS).

cursum(FN,SUM), resolve_value(FT,T,AT,P), ~correlate_write(FT,T,AT),
    reduce_trace_skolem(FT,MFT),
    +try_funptr_escape(MFT), funptr_escape(MFT,FNPTRS),
    ~(funptr_escape(MFT,XFNPTRS), FNPTRS\=XFNPTRS, set_subset(XFNPTRS,FNPTRS)),
    +resolve_value_success(FT,T,AT),
    set_member(FNPTRS,FNPTR), +resolve_value_receive(FT,T,AT,FNPTR,FN,SUM,P,T).

% resolving patterns - dominated by a particular correlated function pointer

% convert a caller function pointer trace PFT to a potential callee trace FT.
% FT might not equal PFT; use callee_funptr_trace
predicate try_callee_funptr_trace(in FN:string,in PFN:string,in PI:c_instr,
                                  in PFT:t_trace,FT:t_trace).
try_callee_funptr_trace(FN,PFN,PI,PFT,FT) :-
    sum_caller(FN,s_func,PFN,PSUM,PP,s_call{PI}),
    argument(p_remote{PFN,PSUM,PP},PI,AT,s_trace{PT},_),
    trace_sub(PFT,PT,RT), trace_compose(drf{AT},RT,FT).
try_callee_funptr_trace(FN,PFN,PI,PFT,FT) :-
    sum_caller(FN,s_func,PFN,PSUM,PP,s_call{PI}),
    argument(p_remote{PFN,PSUM,PP},PI,AT,s_trace{rfld{PT,F,C}},_),
    trace_sub(PFT,PT,RT), trace_compose(fld{drf{AT},F,C},RT,FT).
try_callee_funptr_trace(FN,PFN,PI,PFT,FT) :-
    spoly_call(PFN,PI)->wpoly_call_comp(PFT,_,RFT,RT,AT),
    trace_sub(AT,STR,RT), trace_compose(STR,RFT,FT).

% filter out from try_callee_funptr_trace only those traces FT which are
% definitely equal to PFT within the caller.
% FNPTR indicates whether PFT is used to invoke the indirect call at PI;
% otherwise, PFT is for an indirect function used to call PFN or one of
% its parents
predicate callee_funptr_trace(in FN:string,in PFN:string,in PI:c_instr,
                              in PFT:t_trace,in FNPTR:bool,FT:t_trace).
callee_funptr_trace(FN,PFN,PI,PFT,FNPTR,FT) :-
    try_callee_funptr_trace(FN,PFN,PI,PFT,FT),
    sum_caller(FN,s_func,PFN,PSUM,PP,s_call{PI}),
    inst_trace(s_call{PI},p_remote{PFN,PSUM,PP},FT,s_trace{PFT},G),
    guard(p_remote{PFN,PSUM,PP},PG),
    (FNPTR=true, callfnptr(p_remote{PFN,PSUM,PP},PI,PFT,FG);
     FNPTR=false, #bool_g(true,FG)),
    #and(PG,FG,PFG), guard_implies(PFG,G),
    (FNPTR=true; FNPTR=false, equals_function_entry(PFN,PSUM,PFT)).

% function FN is always called via function pointer FT, which might be equal
% to FNPTR (there may be multiple values of FNPTR for each FT). LIM is the
% limit for scanning callers of FN
predicate caller_dominate(in LIM:int,in FN:string,FT:t_trace,FNPTR:string).

% function FN is always called via function pointer FT when invoked through
% caller PFN/PI
predicate callsite_dominate(in LIM:int,in FN:string,
                            in PFN:string,in PI:c_instr,
                            FT:t_trace,FNPTR:string).

predicate callsite_dominate_list(in LIM:int,in FN:string,
                                 in CALLERS:list[t_pair[string,c_instr]],
                                 FT:t_trace,FNPTR:string).
callsite_dominate_list(LIM,FN,[pair{PFN,PI}],FT,FNPTR) :-
    callsite_dominate(LIM,FN,PFN,PI,FT,FNPTR).
callsite_dominate_list(LIM,FN,[pair{PFN,PI}|TAIL],FT,FNPTR) :- TAIL\=[],
    callsite_dominate(LIM,FN,PFN,PI,FT,FNPTR0),
    callsite_dominate_list(LIM,FN,TAIL,FT,FNPTR1),
    (FNPTR=FNPTR0; FNPTR=FNPTR1).

caller_dominate(LIM,FN,FT,FNPTR) :- int_ge(LIM,0),
    \/function_caller(FN,PFN,PI):list_all(pair{PFN,PI},CALLERS),
    callsite_dominate_list(LIM,FN,CALLERS,FT,FNPTR).

% domination at a particular call site
callsite_dominate(LIM,FN,PFN,PI,FT,FN) :-
    feasible_caller(PFN,PI,FN,PFT),
    callee_funptr_trace(FN,PFN,PI,PFT,true,FT),
    ~(feasible_caller(PFN,PI,FN,XFT),
      ~callee_funptr_trace(FN,PFN,PI,XFT,true,FT)).

% transitive domination at a particular call site
callsite_dominate(LIM,FN,PFN,PI,FT,FNPTR) :-
    int_sub(LIM,1,NLIM), caller_dominate(NLIM,PFN,PFT,FNPTR),
    callee_funptr_trace(FN,PFN,PI,PFT,false,FT).

%?- caller_dominate(_,FN,FT,FNPTR).
%?caller_dominate(LIM,FN,_,_), int_ge(LIM,0), ~caller_dominate(LIM,FN,_,_),
%    +warning("failed caller_dominate",FN).
%?callsite_dominate(LIM,FN,PFN,PI,_,_), ~callsite_dominate(LIM,FN,PFN,PI,_,_),
%    +warning("failed callsite_dominate",FN,PFN,PI).
%?callee_funptr_trace(FN,PFN,PI,PFT,FNPTR,_),
%    ~callee_funptr_trace(FN,PFN,PI,PFT,FNPTR,_), trace_string(PFT,PFTS),
%    +warning("failed callee_funptr_trace",FN,PFN,PI,PFTS,FNPTR).
%?callsite_dominate_list(LIM,FN,[pair{PFN,PI}|TAIL],_,_), TAIL\=[],
%    ~(callsite_dominate(LIM,FN,PFN,PI,FT,_),
%      callsite_dominate_list(LIM,FN,TAIL,FT,_)),
%    +warning("failed callsite_dominate_list",FN,PFN,PI,TAIL).

predicate receive_funptr_threshold(N:int).

predicate receive_funptr_fail(in C:string,in XFT:t_trace,
                              in FT:t_trace,in AT:t_trace,
                              in XFNSET:set[string]).
predicate receive_funptr(in C:string,in XFT:t_trace,
                         in FT:t_trace,in AT:t_trace,
                         in XFNSET:set[string],FNPTR:string).
receive_funptr_fail(C,XFT,FT,AT,XFNSET) :- set_member(XFNSET,XFN),
    spoly_receive_comp(C,XFT,FT,AT)->wpoly_receive(XFN,WFN,_,_,WT),
    reduce_trace_skolem(WT,NWT), extra_reduce_trace_loc(WFN,NWT,WL),
    (receive_funptr_threshold(TR); ~receive_funptr_threshold(_), TR=40),
    ~escape_set(TR,es_backward_precise{ep_field{4}},WL,_),
    ~escape_set(TR,es_backward_precise{ep_field{2}},WL,_).
receive_funptr(C,XFT,FT,AT,XFNSET,FNPTR) :- set_member(XFNSET,XFN),
    spoly_receive_comp(C,XFT,FT,AT)->wpoly_receive(XFN,WFN,_,_,WT),
    reduce_trace_skolem(WT,NWT), extra_reduce_trace_loc(WFN,NWT,WL),
    (receive_funptr_threshold(TR); ~receive_funptr_threshold(_), TR=40),
    (escape_set(TR,es_backward_precise{ep_field{4}},WL,LSET);
     ~escape_set(TR,es_backward_precise{ep_field{4}},WL,_),
         escape_set(TR,es_backward_precise{ep_field{2}},WL,LSET)),
    set_member(LSET,tl_glob{root{glob{FNPTR}}}).

try_funptr_escape(FT), cursum(FN,SUM), caller_dominate(5,FN,XFT,_),
    equals_function_entry(FN,SUM,XFT),
    trace_sub(FT,_,FAT), FAT=drf{fld{empty,_,_}},
    try_propagate_comp(XFT,FT,FAT,C,RXFT,RFT),
    (XT=RXFT; XT=RFT), trace_sub(XT,drf{fld{STR,XF,XC}},_),
    +spoly_comp(XC,XF)->wpoly_search_comp(C,STR,RXFT,RFT,FAT),
    \/caller_dominate(5,FN,XFT,XFN):set_all(XFN,XFNSET),
    ~receive_funptr_fail(C,RXFT,RFT,FAT,XFNSET),
    \/receive_funptr(C,RXFT,RFT,FAT,XFNSET,FNPTR):set_all(FNPTR,FNPTRS),
    +funptr_escape(FT,FNPTRS).

try_funptr_escape(FT), cursum(FN,SUM), caller_dominate(5,FN,FT,XFN),
    equals_function_entry(FN,SUM,FT),
    set_singleton(XFN,FNPTRS), +funptr_escape(FT,FNPTRS).

% baked in behavior

import "poly_data_extra.clp".
import "poly_data_baked.clp".
