
% file

% writes of file->f_op where the private_data must be NULL but we can't prove it.
cil_curfn("chrdev_open"), +ignore_failure("file",_,_,_).
cil_curfn("drivers/char/misc.c:misc_open"), +ignore_failure("file",_,_,_).
cil_curfn("drivers/usb/core/file.c:usb_open"), +ignore_failure("file",_,_,_).
cil_curfn("sound/core/sound.c:snd_open"), +ignore_failure("file",_,_,_).
cil_curfn("fs/open.c:__dentry_open"), +ignore_failure("file",_,_,_).

% spurious modset information on do_mmap calls
cil_curfn("drivers/char/drm/i810_dma.c:i810_map_buffer"),
    +ignore_failure("file",_,drf{fld{empty,"private_data","file"}},_).
cil_curfn("drivers/char/drm/i830_dma.c:i830_map_buffer"),
    +ignore_failure("file",_,drf{fld{empty,"private_data","file"}},_).

% do_readv_writev splits up function pointer assigns/calls across a loop
% boundary in a way we can't handle. there are other calls to read/write
% so just ignore this one
+ignore_callgraph_site("fs/read_write.c:do_readv_writev",_,_).

% write function called by read function, wierd
+ignore_callgraph_site("fs/nfsd/nfsctl.c:nfsctl_transaction_read",_,"fs/nfsd/nfsctl.c:nfsctl_transaction_write").

% calls lookup_instantiate_filp and sets private_data afterwards.
% this seems to be the only possible file_operations for the file
cil_curfn("fs/9p/vfs_inode.c:v9fs_vfs_create"),
    +transform_assign_value(drf{fld{_,F,"file_operations"}},T,
                            drf{fld{root{glob{"v9fs_file_operations"}},F,"file_operations"}},T).

% TODO: what are the possible file_operations here?
cil_curfn("fuse_finish_open"), +ignore_failure("file",_,_,_).

% explicit test comparing file->f_op to these operations
cil_curfn("fs/fuse/inode.c:fuse_fill_super"),
    +transform_assign_value(drf{fld{_,F,"file_operations"}},T,
                            drf{fld{root{glob{"fuse_dev_operations"}},F,"file_operations"}},T).

% might reset file->f_op to tty_fops, which we already know is a target here
cil_curfn("drivers/char/tty_io.c:tty_open"), +ignore_failure("file",_,_,_).

% initialization code of custom file_operations. this should occur before
% any files have been assigned these operations.
cil_curfn("snd_tea575x_init"), +ignore_failure("file",_,_,_).
cil_curfn("udp_proc_register"), +ignore_failure("file",_,_,_).
cil_curfn("tcp_proc_register"), +ignore_failure("file",_,_,_).
cil_curfn("drivers/scsi/sg.c:sg_proc_init"), +ignore_failure("file",_,_,_).
cil_curfn("drivers/acpi/processor_perflib.c:acpi_cpufreq_add_file"), +ignore_failure("file",_,_,_).
cil_curfn("drivers/acpi/video.c:acpi_video_bus_add_fs"), +ignore_failure("file",_,_,_).
cil_curfn("drivers/acpi/video.c:acpi_video_device_add_fs"), +ignore_failure("file",_,_,_).

% replaces file operations with dummy ops
cil_curfn("snd_card_disconnect"), +ignore_failure("file",_,_,_).

% sets both private_data and f_op, but then calls open method on new f_op which
% will set private_data to a new value. the discovered correlations are bogus
cil_curfn("drivers/media/dvb/dvb-core/dvbdev.c:dvb_device_open"), +ignore_failure("file",_,_,_).

% I don't think there is any value for read_proc at this point
% in the procdir hierarchy
cil_curfn("reiserfs_proc_info_init"), +ignore_failure("proc_dir_entry",_,_,_).

% snd_kcontrol

% weird one, private_data and fnptrs are written to two different structures
% which are copied between by external call snd_ctl_new
transform_assign_value(T,XT,root{glob{FN}},XT) :-
    cil_curfn("sound/core/control.c:snd_ctl_elem_add"), trace_remove_uc_sum(T,NT),
    (NT=drf{fld{_,"get","snd_kcontrol"}}, FN="sound/core/control.c:snd_ctl_elem_user_get";
     NT=drf{fld{_,"put","snd_kcontrol"}}, FN="sound/core/control.c:snd_ctl_elem_user_put";
     NT=drf{fld{_,"info","snd_kcontrol"}}, FN="sound/core/control.c:snd_ctl_elem_user_info").

cil_curfn("snd_hda_add_new_ctls"), +funptr_escape_threshold(500).

% urb

% overwriting the urb context from one port to another. this is startup code for
% the usb_serial so we known the existing urb's context is a port and not
% some other structure
cil_curfn("drivers/usb/serial/mct_u232.c:mct_u232_startup"),
    +ignore_failure("urb",_,_,_).

% fills in NULL fnptrs with generic operations
cil_curfn("drivers/usb/serial/usb-serial.c:fixup_generic"),
    +ignore_failure("usb_serial_port",_,_,_).

% snd_pcm

% each snd_pcm has a different runtime (see snd_pcm_attach_substream),
% so modifying the fields of one runtime will not affect any others
+omit_base_structure_fail("snd_pcm_substream",_,_,_,
                          fld{drf{fld{_,"runtime","snd_pcm_substream"}},_,"snd_pcm_runtime"}).

% snd_pcm_new does not fill in its output parameter under out-of-memory conds
omit_zero_initialized_fail("snd_pcm_new",s_func,_,T) :-
    T = drf{fld{drf{drf{root{arg{5}}}},"private_free","snd_pcm"}}.

% weird data structure here, snd_pcm->private_data is copied to each
% individual substream when the substream is attached
% (snd_pcm_attach_substream) but the ops are copied across by snd_pcm_set_ops
% when the snd_pcm is created. writes of snd_pcm->private_data should
% be correlated with calls to snd_pcm_set_ops. TODO: should make sure each
% write of private_data has one of these
dircall(I,"snd_pcm_set_ops"), icall(P,_,I),
    spoly_comp("snd_pcm_substream","ops")->
        wpoly_search_comp("snd_pcm_substream",empty,RFT,RT,AT),
    ~ignore_failure("snd_pcm_substream",RFT,RT,AT),
    RT = drf{fld{empty,"private_data","snd_pcm_substream"}},
    trace_sub(RFT,drf{fld{empty,"ops","snd_pcm_substream"}},RFTF),
    trace_compose(drf{root{arg{2}}},RFTF,CFT),
    inst_trace(s_call{I},P,CFT,s_trace{NFT},FTG),
    inst_trace(s_call{I},P,fld{drf{root{arg{0}}},"private_data","snd_pcm"},s_trace{PT},TG),
    (gassign(_,PT,NT,ATG); ~gassign(_,PT,_,_), val(P,PT,s_trace{NT},ATG)),
    guard(P,PG), #and4(TG,ATG,FTG,PG,G), guard_sat(G),
    +poly_assign_value(empty,RFT,RT,AT,NFT,NT,P).
cil_curfn("snd_pcm_attach_substream"), +ignore_failure("snd_pcm_substream",_,_,_).
cil_curfn("snd_pcm_set_ops"), +ignore_failure("snd_pcm_substream",_,_,_).

% really weird use of snd_pcm_substream here. snd_cmi8330_pcm sets these two special open
% callbacks which revert the substream->private_data to another value.
cil_curfn("sound/isa/cmi8330.c:snd_cmi8330_pcm"),
    +ignore_failure("snd_pcm_substream",_,_,_).
cil_curfn("sound/isa/cmi8330.c:snd_cmi8330_playback_open"),
    +ignore_failure("snd_pcm_substream",_,_,_).
cil_curfn("sound/isa/cmi8330.c:snd_cmi8330_capture_open"),
    +ignore_failure("snd_pcm_substream",_,_,_).
+ignore_callgraph_site("sound/isa/cmi8330.c:snd_cmi8330_playback_open",_,_).
+ignore_callgraph_site("sound/isa/cmi8330.c:snd_cmi8330_capture_open",_,_).

% stop some funky propagation when the pcm's private_data is copied to it's runtime's private_data
% e.g. pdacf_pcm_capture_open
+omit_propagate_comp("snd_pcm_substream",_,drf{fld{empty,"private_data","snd_pcm_substream"}},
                     drf{fld{drf{fld{_,"runtime","snd_pcm_substream"}},"private_data","snd_pcm_runtime"}}).

% snd_rawmidi

% rawmidi streams work similary to pcm streams, a call to snd_rawmidi_set_ops
% is correlated with a set of the rmidi->private_data. except here the private
% data is reached via a parent pointer rather than copying during attach()
predicate assign_rawmidi_ops(P:pp,STR:t_trace,OPS:t_trace,PRIV:t_trace).
dircall(I,"snd_rawmidi_set_ops"), icall(P,_,I),
    inst_trace(s_call{I},P,index{drf{root{arg{0}}},y_comp{"snd_rawmidi_substream"},
                                 s_trace{drf{root{arg{1}}}}},s_trace{STR},_),
    inst_trace(s_call{I},P,drf{root{arg{2}}},s_trace{OPS},_),
    gassign(_,fld{_,"private_data","snd_rawmidi"},PRIV,_), #bool_g(true,G),
    +add_gassign(P,fld{drf{fld{STR,"rmidi","snd_rawmidi_substream"}},"private_data","snd_rawmidi"},s_trace{PRIV},G),
    +add_gassign(P,fld{STR,"ops","snd_rawmidi_substream"},s_trace{OPS},G).
cil_curfn("snd_rawmidi_set_ops"), +ignore_failure("snd_rawmidi_substream",_,_,_).
+omit_base_structure_fail("snd_rawmidi_substream",_,_,_,fld{_,"private_data","snd_rawmidi"}).

% function which loops through and sets substream->ops by hand instead of using snd_rawmidi_set_ops
cil_curfn("sound/drivers/mtpav.c:snd_mtpav_get_RAWMIDI"),
    gassign(P,fld{STR,"ops","snd_rawmidi_substream"},_,G),
    +gassign(P,fld{drf{fld{STR,"rmidi","snd_rawmidi_substream"}},"private_data","snd_rawmidi"},drf{root{arg{0}}},G).    

% device

% functions called to fill in driver data for various devices.
% inline these in the caller
cil_curfn("include/linux/device.h:dev_set_drvdata"), +inline_assignment(_,_).
cil_curfn("include/linux/pci.h:pci_set_drvdata"), +inline_assignment(_,_).
cil_curfn("include/linux/i2c.h:i2c_set_clientdata"), +inline_assignment(_,_).
cil_curfn("drivers/usb/serial/usb-serial.h:usb_set_serial_port_data"), +inline_assignment(_,_).
cil_curfn("include/linux/usb.h:usb_set_intfdata"), +inline_assignment(_,_).

cil_curfn("usb_register_driver"), +ignore_escape_function("usb_register_driver").

% pci_dev

% creates custom operations
cil_curfn("saa7146_register_extension"), +ignore_failure("pci_dev",_,_,_).

% these callers set a device's driver *after* successful return from the probe
% function, rather than before. treat these call sites as dominating with
% a call to probe
callee_funptr_trace(_,"drivers/ide/setup-pci.c:ide_scan_pcidev",_,PDRV,_,DRV) :-
    PDRV = drf{fld{rfld{drf{root{local{"l"}}},"node","pci_driver"},"probe","pci_driver"}},
    DRV = drf{fld{drf{fld{drf{root{arg{0}}},"driver","pci_dev"}},"probe","pci_driver"}}.
callee_funptr_trace(_,"drivers/pci/pci-driver.c:pci_call_probe",_,PDRV,_,DRV) :-
    PDRV = drf{fld{drf{root{arg{0}}},"probe","pci_driver"}},
    DRV = drf{fld{drf{fld{drf{root{arg{0}}},"driver","pci_dev"}},"probe","pci_driver"}}.
cil_curfn("drivers/ide/setup-pci.c:ide_scan_pcidev"), +ignore_failure("pci_dev",_,_,_).
cil_curfn("drivers/pci/pci-driver.c:__pci_device_probe"), +ignore_failure("pci_dev",_,_,_).

% TODO: this is overly broad and ignores possible aliasing between
% devices when the driver_data is written
+omit_base_structure_fail("pci_dev",_,_,_,fld{_,"driver_data","device"}).

% these are cases where do_ide_setup_pci_device is called with a particular
% ide_pci_device_t from the probe function of a driver. since we smush together
% all calls to do_ide_setup_pci_device, we lose the correlation.
predicate receive_driver(FN:string,DRIVER:string).
cil_curfn(FN), receive_driver(FN,DRIVER),
    +transform_assign_value(drf{fld{_,F,"pci_driver"}},T,drf{fld{root{glob{DRIVER}},F,"pci_driver"}},T).
+receive_driver("drivers/ide/pci/aec62xx.c:init_chipset_aec62xx",
                "drivers/ide/pci/aec62xx.c:driver").
+receive_driver("drivers/ide/pci/siimage.c:setup_mmio_siimage",
                "drivers/ide/pci/siimage.c:driver").

% similarly, pciehp_probe is only called when the incoming pci_dev has a particular driver
+receive_driver("drivers/pci/hotplug/pciehp_core.c:pciehp_probe",
                "drivers/pci/pcie/portdrv_pci.c:pcie_portdrv").

% called with a child data structure of the pci_dev with a parent pointer back up
callee_funptr_trace("cx8802_init_common",_,_,PFT,_,FT) :-
    trace_sub(PFT,drf{root{arg{0}}},RT), trace_compose(drf{fld{drf{root{arg{0}}},"pci","cx8802_dev"}},RT,FT).
callee_funptr_trace("drivers/media/dvb/b2c2/flexcop-pci.c:flexcop_pci_init",_,_,PFT,_,FT) :-
    trace_sub(PFT,drf{root{arg{0}}},RT), trace_compose(drf{fld{drf{root{arg{0}}},"pdev","flexcop_pci"}},RT,FT).
callee_funptr_trace("drivers/message/i2o/pci.c:i2o_pci_alloc",_,_,PFT,_,FT) :-
    trace_sub(PFT,drf{root{arg{0}}},RT), trace_compose(drf{fld{drf{root{arg{0}}},"pdev","i2o_controller"}},RT,FT).
callee_funptr_trace("drivers/scsi/sym53c8xx_2/sym_glue.c:sym_attach",_,_,PFT,_,FT) :-
    trace_sub(PFT,drf{root{arg{0}}},RT), trace_compose(drf{fld{drf{root{arg{2}}},"pdev","sym_device"}},RT,FT).
+omit_comp_impure("cx8802_dev","pci",_).
+omit_comp_impure("flexcop_pci","pdev",_).
+omit_comp_impure("i2o_controller","pdev",_).
+omit_comp_impure("sym_device","pdev",_).

% ignore call site where agp_amd64_probe is called directly.
% TODO: need to sort out device capability probing
+ignore_callgraph_site("agp_amd64_init",_,"drivers/char/agp/amd64-agp.c:agp_amd64_probe").

% the pci_dev we care about is NULL at these call sites
+ignore_callgraph_site("drivers/net/defxx.c:dfx_eisa_init",_,"drivers/net/defxx.c:dfx_init_one_pci_or_eisa").
+ignore_callgraph_site("drivers/net/tlan.c:TLan_EisaProbe",_,"drivers/net/tlan.c:TLan_probe1").
+ignore_callgraph_site("drivers/net/pcnet32.c:pcnet32_probe_vlbus",_,"drivers/net/pcnet32.c:pcnet32_probe1").

% punt on these. backward tracking not yet good enough
cil_curfn("drivers/pci/hotplug/pciehp_core.c:pciehp_probe"), +ignore_failure("pci_dev",_,_,_).
cil_curfn("drivers/isdn/hardware/avm/b1isa.c:b1isa_probe"), +ignore_failure("pci_dev",_,_,_).
cil_curfn("drivers/isdn/hardware/avm/t1isa.c:t1isa_probe"), +ignore_failure("pci_dev",_,_,_).
cil_curfn("drivers/net/3c59x.c:vortex_probe1"), +ignore_failure("pci_dev",_,_,_).

% functions which fill in driver fnptr fields during driver registration
cil_curfn("drivers/char/epca.c:init_PCI"), +omit_base_structure_fail("pci_dev",_,_,_,_).
cil_curfn("pcie_port_service_register"), +omit_base_structure_fail("pci_dev",_,_,_,_).
cil_curfn("platform_driver_register"), +omit_base_structure_fail("pci_dev",_,_,_,_).

% i2c_client

% TODO: ditto wrt pci_dev
+omit_base_structure_fail("i2c_client",_,_,_,fld{_,"driver_data","device"}).

% net_device

% assigns old value of hard_start_xmit
cil_curfn("net/atm/mpc.c:stop_mpc"), +ignore_failure("net_device",_,_,_).

% assigns old value of change_mtu
cil_curfn("drivers/net/wan/hdlc_ppp.c:ppp_close"), +ignore_failure("net_device",_,_,_).

% correlations going on here which we aren't capturing currently. TODO: fixme
cil_curfn("drivers/net/wan/hdlc_ppp.c:ppp_open"), +ignore_failure("net_device",_,_,_).
cil_curfn("hdlc_ppp_ioctl"), +ignore_failure("net_device",_,_,_).

% cpc_open is called both through dev->open as well as directly at this call site.
% at this point the dev->open is in fact cpc_open (see cpc_init_card)
+ignore_callgraph_site("cpc_tty_init",_,"cpc_open").

% the polling_netdev here does not have any fnptrs set up other than poll
cil_curfn("drivers/net/e1000/e1000_main.c:e1000_sw_init"),
    +ignore_assign_value(drf{fld{_,_,"net_device"}},_).

% assigns dev->priv the value it just read out. whatever
cil_curfn("drivers/net/wan/sbni.c:sbni_probe1"),
    +ignore_failure("net_device",_,drf{fld{empty,"priv","net_device"}},_).

% fb_info

% global variable used instead of kmalloc, weird
cil_curfn("drivers/video/tridentfb.c:trident_pci_probe"), +ignore_failure("fb_info",_,_,_).

% copies ops structures from heap to local data
cil_curfn("drivers/video/i810/i810_main.c:i810_init_defaults"), +omit_base_structure_fail("fb_info",_,_,_,_).
cil_curfn("drivers/video/cyblafb.c:cybla_pci_probe"), +omit_base_structure_fail("fb_info",_,_,_,_).
cil_curfn("drivers/video/matrox/matroxfb_base.c:initMatrox2"), +omit_base_structure_fail("fb_info",_,_,_,_).

% overwrites fb_info->ops fields directly. these are not shared between fb_info
cil_curfn("aty_init_cursor"), +omit_base_structure_fail("fb_info",_,_,_,_).
cil_curfn("drivers/video/nvidia/nvidia.c:nvidiafb_set_par"), +omit_base_structure_fail("fb_info",_,_,_,_).

% overwrites matrox_fb_info.fbops rather than going through fb_info
cil_curfn("matrox_cfbX_init"), +omit_base_structure_fail("fb_info",_,_,_,_).

% tty_struct

% filling in driver_data after initialization; there should already be one of the same type there
cil_curfn("drivers/char/keyboard.c:kbd_keycode"), +ignore_failure("tty_struct",_,_,_).

% initialization code for tty_driver called before any tty_struct has been created
cil_curfn("tty_set_operations"), +ignore_failure("tty_struct",_,_,_).
cil_curfn("tty_register_driver"), +ignore_failure("tty_struct",_,_,_).
cil_curfn("cpc_tty_init"), +ignore_failure("tty_struct",_,_,_).
cil_curfn("drivers/char/pty.c:legacy_pty_init"), +omit_base_structure_fail("tty_struct",_,_,_,_).
cil_curfn("drivers/char/pty.c:unix98_pty_init"), +omit_base_structure_fail("tty_struct",_,_,_,_).

% per-tty_struct initialization, the ldisc data is not filled in yet.
cil_curfn("drivers/char/tty_io.c:tty_ldisc_assign"), +ignore_failure("tty_struct",_,_,_).

% calls ptm_driver->open directly instead of tty->driver->open. the two must alias here
+ignore_callgraph_site("drivers/char/tty_io.c:ptmx_open",_,_).

% trouble figuring out which caller tty_struct's are invoking at the call site here
callsite_dominate(_,FN,PFN,PI,FT,FN) :-
    (PFN="drivers/char/tty_io.c:init_dev", tostring(PI,"\"#927\"");
     PFN="drivers/char/tty_io.c:init_dev", tostring(PI,"\"#963\"");
     PFN="drivers/char/tty_io.c:tty_set_ldisc", tostring(PI,"\"#744\"");
     PFN="drivers/char/tty_io.c:tty_set_ldisc", tostring(PI,"\"#821\"");
     PFN="drivers/char/tty_io.c:tty_set_ldisc", tostring(PI,"\"#890\"")),
    FT=drf{fld{fld{drf{root{arg{0}}},"ldisc","tty_struct"},
               "open","tty_ldisc"}}.

% snd_info_entry

% snd_card_proc_new does not fill in its output parameter under out-of-memory conditions
% which is not currently handled by the zerofield analysis
omit_zero_initialized_fail("snd_card_proc_new",_,_,T) :- trace_root(T,arg{2}).

% currently fail on this one because we don't keep track that fields of this haven't
% yet escaped into the heap-at-large
+omit_zero_initialized_fail("snd_info_create_card_entry",_,_,_).

% for some reason this function doesn't write the c.text.write field, so to get
% correlations with callers accurately just inline these assignments
cil_curfn("include/sound/info.h:snd_info_set_text_ops"), +inline_assignment(_,_).

% given the above inlining, we can ignore escaping involving these variables.
% these are reused for creating several different snd_info_entry's
ignore_escape_edge(_,tl_func{"snd_emu10k1_proc_init",T,_},_) :- trace_root(T,local{"entry"}).
ignore_escape_edge(_,tl_func{"sound/pci/cs4281.c:snd_cs4281_proc_init",T,_},_) :- trace_root(T,local{"entry"}).
ignore_escape_edge(_,tl_func{"sound/pci/mixart/mixart.c:snd_mixart_proc_init",T,_},_) :- trace_root(T,local{"entry"}).

cil_curfn("snd_ca0106_proc_init"), +omit_trace_clobber(_,_,_).
cil_curfn("sound/core/pcm.c:snd_pcm_stream_proc_init"), +omit_trace_clobber(_,_,_).
cil_curfn("sound/core/pcm.c:snd_pcm_substream_proc_init"), +omit_trace_clobber(_,_,_).

% seq_file

% functions which fill in seq_ops dynamically
cil_curfn("net/ipv4/tcp_ipv4.c:tcp_seq_open"), +omit_base_structure_fail("seq_file",_,_,_,_).
cil_curfn("net/ipv4/udp.c:udp_seq_open"), +omit_base_structure_fail("seq_file",_,_,_,_).
cil_curfn("single_open"), +omit_base_structure_fail("seq_file",_,_,_,_).

% separately handling calls to seq_open, see aliasing_full.clp
cil_curfn("seq_open"), +ignore_failure("seq_file",_,_,_).

% weird caller for single_open...
cil_curfn("net/sunrpc/rpc_pipe.c:rpc_info_open"),
    +transform_assign_value(drf{fld{_,"show","seq_operations"}},FT,
                            root{glob{"net/sunrpc/rpc_pipe.c:rpc_show_info"}},FT),
    cil_sum_body("single_open",s_func)->iset(XP,_,_),
    assign(p_remote{"single_open",s_func,XP},fld{_,FLD,"seq_operations"},s_trace{root{glob{XFN}}},_), FLD\="show",
    +transform_assign_value(drf{fld{_,FLD,"seq_operations"}},XFT,
                            root{glob{XFN}},XFT).

% changing private to point to other object of same type
cil_curfn("arch/i386/kernel/cpu/mtrr/if.c:mtrr_file_add"), +ignore_failure("seq_file",_,_,_).

% seq_open only clobbers file->private_data if it was originally non-NULL.
% we need to know file->private data still points to p->m after the call to correlate the writes here
cil_curfn("fs/proc/base.c:mounts_open"), dircall(I,"seq_open"), +omit_trace_clobber(s_call{I},_,_).

% more calls our modset info is overly imprecise for
cil_curfn("fs/proc/base.c:mountstats_open"), dircall(I,CFN),
    (CFN="include/linux/sched.h:task_lock"; CFN="include/linux/sched.h:task_unlock"),
    +omit_trace_clobber(s_call{I},_,_).

% function called in both the ->start and ->next paths, confusing the dominator analysis
+ignore_callgraph_site("net/ipv4/ipvs/ip_vs_conn.c:ip_vs_conn_seq_next",_,
                       "net/ipv4/ipvs/ip_vs_conn.c:ip_vs_conn_array").

% usb_serial_port

% ditto the overly broad non-aliasing as for pci_dev
+omit_base_structure_fail("usb_serial_port",_,_,_,fld{_,"driver_data","device"}).

% the serial's type->attach is being used at this call site but we can't prove it
callee_funptr_trace(_,"usb_serial_probe",_,PATCH,_,ATCH) :-
    PATCH=drf{fld{_,"attach","usb_serial_driver"}},
    ATCH=drf{fld{drf{fld{drf{root{arg{0}}},"type","usb_serial"}},"attach","usb_serial_driver"}}.

predicate parent_pointer_trace(in T:t_trace).
transform_assign_value(FT,T,NFT,T) :-
    uc_trace_sub(FT,STR,RT), trace_sub(RT,PARENT,XRT), parent_pointer_trace(PARENT),
    trace_compose(STR,XRT,NFT).

% capture parent pointer relationship between usb_serial and usb_serial_port:
% usb_serial->port[*]->serial == usb_serial
+parent_pointer_trace(drf{fld{drf{index{fld{empty,"port","usb_serial"},_,_}},"serial","usb_serial_port"}}).

% usb_interface

+omit_base_structure_fail("usb_interface",_,_,_,fld{_,"driver_data","device"}).

% functions filling in device->driver before/during probe
cil_curfn("driver_probe_device"), +ignore_failure("usb_interface",_,_,_), +ignore_failure("pci_dev",_,_,_).
cil_curfn("phy_attach"), +ignore_failure("usb_interface",_,_,_), +ignore_failure("pci_dev",_,_,_).
cil_curfn("usbvideo_register"), +ignore_failure("usb_interface",_,_,_), +ignore_failure("pci_dev",_,_,_).

callee_funptr_trace("drivers/media/dvb/b2c2/flexcop-usb.c:flexcop_usb_init",_,_,PFT,_,FT) :-
    trace_sub(PFT,drf{root{arg{0}}},RT), trace_compose(drf{fld{drf{root{arg{0}}},"uintf","flexcop_usb"}},RT,FT).
+omit_comp_impure("flexcop_usb","uintf",_).

% mtd_info

% TODO: being lazy on these functions
cil_curfn("drivers/mtd/devices/mtdram.c:init_mtdram"), +ignore_failure("mtd_info",_,_,_).
cil_curfn("drivers/mtd/devices/slram.c:register_device"), +ignore_failure("mtd_info",_,_,_).
cil_curfn("drivers/mtd/nand/nandsim.c:ns_init_module"), +ignore_failure("mtd_info",_,_,_).

% pnp_card_link

% TODO: fix silly breakage in zero_initialized
cil_curfn("drivers/pnp/card.c:card_probe"), +ignore_failure("pnp_card_link",_,_,_).

% Scsi_Host

% fills in missing fields of scsi_host_template
cil_curfn("NCR_700_detect"), +omit_base_structure_fail("Scsi_Host",_,_,_,_).

% inline the assignment of the host template in scsi_host_alloc so we can
% distinguish the different call sites where hosts are allocated. since the
% structure didn't exist before the call, we have to do this by hand
dircall(I,"scsi_host_alloc"), icall(P,_,I), callret(P,I,RT,_),
    inst_trace(s_call{I},P,drf{root{arg{0}}},TEMPLATE,G),
    +add_gassign(P,fld{uc_sum{s_call{I},P,RT},"hostt","Scsi_Host"},TEMPLATE,G).
cil_curfn("scsi_host_alloc"), +ignore_failure("Scsi_Host",_,_,_).

% hwgroup

cil_curfn("drivers/ide/ide-iops.c:__ide_set_handler"), +inline_assignment(fld{_,"handler","hwgroup_s"},_).
cil_curfn("ide_set_handler"), +inline_assignment(fld{_,"handler","hwgroup_s"},_).

% saa7146_dev

% need to prove ->ext_priv is NULL here. need better zerofield / side effects
cil_curfn("drivers/media/common/saa7146_core.c:saa7146_init_one"),
    +ignore_failure("saa7146_dev",_,_,_).

% elevators

% the assignment of the elevator_data occurs in elevator_attach, being passed the
% return value of the init method for the elevator_type. ignore the attach assign
% and treat the return statement in the init function like the true assign
cil_curfn("block/elevator.c:elevator_attach"), +ignore_failure("elevator_queue",_,_,_).
cil_curfn(FN), function_caller(FN,"block/elevator.c:elevator_init_queue",_),
    add_gassign(P,root{return},V,G),
    +add_gassign(P,fld{drf{root{arg{1}}},"elevator_data","elevator_queue"},V,G).

% request queues

predicate transform_assign_field_function(FN:string,F:string,GFN:string).
cil_curfn(FN), transform_assign_field_function(FN,F,GFN),
    +transform_assign_value(drf{fld{_,F,_}},T0,root{glob{GFN}},T0),
    +transform_assign_value(uc_sum{_,_,fld{_,F,_}},T1,root{glob{GFN}},T1).

% the make_request_fn field is set here through function calls we aren't
% tracking precisely enough, so just specify the function set here
+transform_assign_field_function(
    "drivers/block/pktcdvd.c:pkt_init_queue","make_request_fn",
    "drivers/block/pktcdvd.c:pkt_make_request").
+transform_assign_field_function(
    "drivers/scsi/scsi_scan.c:scsi_alloc_sdev","make_request_fn",
    "block/ll_rw_blk.c:__make_request").
+transform_assign_field_function(
    "mmc_init_queue","make_request_fn",
    "block/ll_rw_blk.c:__make_request").

% dvb_frontend

% each dvb_frontend has a different ops created during attach(),
% so modifying the fields of one ops will not affect any others
+omit_base_structure_fail("dvb_frontend",_,_,_,
                          fld{drf{fld{_,"ops","dvb_frontend"}},_,_}).

% replaces some of the callbacks with new callbacks that don't look at demodulator_priv.
% lnbp21_exit restores and invokes the original release callback (to avoid resource leaks)
cil_curfn("drivers/media/dvb/frontends/lnbp21.h:lnbp21_init"),
    +ignore_failure("dvb_frontend",_,drf{fld{empty,"demodulator_priv","dvb_frontend"}},_).
cil_curfn("drivers/media/dvb/frontends/lnbp21.h:lnbp21_exit"),
    +ignore_failure("dvb_frontend",_,drf{fld{empty,"demodulator_priv","dvb_frontend"}},_).

% Qdisc

% creates Qdisc_ops on the heap during startup
cil_curfn("net/sched/sch_teql.c:teql_master_setup"),
    +omit_base_structure_fail("Qdisc",_,_,_,_).

% fills in empty Qdisc_ops fields with nop operations
cil_curfn("register_qdisc"),
    +ignore_failure("Qdisc",_,_,_).

% virtual mappings

% might set vma->vm_ops to a new value. I don't think that this branch
% is actually reachable
cil_curfn("drivers/char/agp/frontend.c:agp_mmap"),
    +ignore_failure("vm_area_struct",_,_,_).

% pcmcia

% weird function that sets an IRQ's instance but not handler.
% the set of functions we get off the embedded structure is too general
cil_curfn("drivers/net/pcmcia/com20020_cs.c:com20020_probe"),
    +ignore_failure("irq_req_t",_,_,_).

% atm_dev

% similar to scsi_host_alloc, atm_dev_register both creates a device and sets
% its operations field.
dircall(I,"atm_dev_register"), icall(P,_,I), callret(P,I,RT,_),
    inst_trace(s_call{I},P,drf{root{arg{1}}},OPS,G),
    T = fld{uc_sum{s_call{I},P,RT},"ops","atm_dev"},
    +assign(P,T,OPS,G), +add_gassign(P,T,OPS,G).

% calls atm_dev_register with these ops in its only caller, firestream_init_one.
cil_curfn("drivers/atm/firestream.c:fs_init"),
    +transform_assign_value(drf{fld{_,F,"atmdev_ops"}},T,
                            drf{fld{root{glob{"drivers/atm/firestream.c:ops"}},F,"atmdev_ops"}},T).

% snd_ac97

% similar to scsi_host_alloc, snd_ac97_bus creates a bus and sets its ops.
%dircall(I,"snd_ac97_bus"), icall(P,_,I),
%    inst_trace(s_call{I},P,drf{root{arg{4}}},s_trace{RT},G0),
%    inst_trace(s_call{I},P,drf{root{arg{2}}},OPS,G1),
%    T = fld{uc_sum{s_call{I},P,RT},"ops","snd_ac97_bus"},
%    #and(G0,G1,G), +assign(P,T,OPS,G), +add_gassign(P,T,OPS,G).
%+omit_base_structure_fail("snd_ac97",_,_,_,_).

% input_dev

% the device is empty here and there are no funptrs initialized.
cil_curfn("drivers/input/mouse/psmouse-base.c:psmouse_switch_protocol"),
    +zero_initialized(_,drf{fld{drf{fld{drf{root{arg{0}}},"dev","psmouse"}},_,_}}).
cil_curfn("drivers/input/keyboard/atkbd.c:atkbd_set_device_attrs"),
    +zero_initialized(_,drf{fld{drf{fld{drf{root{arg{0}}},"dev","atkbd"}},_,_}}).

% snd_hwdep

% we don't mark snd_hwdep_new as returning allocated data
dircall(I,"snd_hwdep_new"), icall(P,_,I),
    cursum(FN,SUM), +allocate_filter(FN,SUM,fld{fld{uc_sum{s_call{I},P,_},"ops","snd_hwdep"},_,_}).

% drm_device

% undetected initialization code
cil_curfn(FN),
    (FN="drivers/char/drm/r128_cce.c:r128_do_init_cce";
     FN="drivers/char/drm/i810_dma.c:i810_dma_initialize";
     FN="drivers/char/drm/i830_dma.c:i830_dma_initialize"),
    +zero_initialized(_,drf{fld{drf{fld{drf{root{arg{0}}},"driver","drm_device"}},_,_}}).
