import "aliasing_sum.clp".

import "../smemory/memory.clp".
import "../smemory/scalar_sat.clp".
import "../smemory/aliasing_full.clp".

analyze analysis_name("poly_data_init").
analyze session_name("cil_sum_body").
analyze priority(0).

% output

predicate str_poly_target(FN:string,LINE:int,FNE:string,FTS:string).
?- str_poly_target(FN,LINE,FNE,FT).

predicate str_poly_comp(FN:string,LINE:int,FNE:string,FTS:string,
                        C:string,RFTS:string,RTS:string,ATS:string).
?- str_poly_comp(FN,LINE,FNE,FT,C,RFT,RT,AT).

predicate str_poly_func(FN:string,LINE:int,FNE:string,FTS:string,
                        TS:string,ATS:string).
?- str_poly_func(FN,LINE,FNE,FT,T,AT).

% seed information

predicate add_poly_target(I:c_instr,FT:t_trace).
add_poly_target(I,FT), cil_curfn(FN),
    +spoly_call(FN,I)->clear_old_preds(),
    +spoly_call(FN,I)->wpoly_call_target(FT),
    icall(P,_,I), point_location(P,_,LINE), cil_instr_call(I,E),
    exp_string(E,ES), trace_string(FT,FTS),
    +str_poly_target(FN,LINE,ES,FTS).

predicate add_poly_comp(I:c_instr,FT:t_trace,
                        C:string,RFT:t_trace,RT:t_trace,AT:t_trace).
add_poly_comp(_,_,C,RFT,RT,AT),
    (XT=RFT; XT=RT), trace_sub(XT,drf{fld{STR,XF,XC}},_),
    +spoly_comp(XC,XF)->wpoly_search_comp(C,STR,RFT,RT,AT).
add_poly_comp(I,FT,C,RFT,RT,AT), cil_curfn(FN),
    +spoly_call(FN,I)->wpoly_call_comp(FT,C,RFT,RT,AT),
    icall(P,_,I), point_location(P,_,LINE), cil_instr_call(I,E),
    exp_string(E,ES), trace_string(FT,FTS),
    trace_string(RFT,RFTS), trace_string(RT,RTS), trace_string(AT,ATS),
    +str_poly_comp(FN,LINE,ES,FTS,C,RFTS,RTS,ATS).

predicate add_poly_func(I:c_instr,FT:t_trace,T:t_trace,AT:t_trace).
add_poly_func(_,FT,T,AT),
    cil_curfn(FN), +spoly_func(FN)->wpoly_search_func(FT,T,AT).
add_poly_func(I,FT,T,AT), cil_curfn(FN),
    +spoly_call(FN,I)->wpoly_call_func(FT,T,AT),
    icall(P,_,I), point_location(P,_,LINE), cil_instr_call(I,E),
    exp_string(E,ES), trace_string(FT,FTS),
    trace_string(T,TS), trace_string(AT,ATS),
    +str_poly_func(FN,LINE,ES,FTS,TS,ATS).

% RT is a trace relative to AT we want to track data through.
% T is an actual argument which might correspond to AT
predicate candidate_trace(in T:t_trace,in I:c_instr,in AT:t_trace,RT:t_trace).

% pull in spoly_call_try seed information supplied by other analyses
candidate_trace(_,I,AT,RT) :-
    anycall(I,CFN,_), spoly_call_try(CFN)->wpoly_call_try(XAT,_),
    trace_sub(XAT,AT,RT).

% the argument trace is a void* pointer or an integer
% (which are sometimes used to masquerade pointers)
candidate_trace(drf{T},_,_,empty) :-
    cil_curfn(FN), trace_type(FN,T,Y),
    (Y=y_ptr{y_void}; Y=y_int{K}, ikind_bytes(K,_,BS), pointer_bytes(BS)).

% RT is a trace relative to values of type C we want to track data through
predicate candidate_field(in C:string,RT:t_trace).
candidate_field(C,drf{fld{empty,F,C}}) :-
    comp_field(C,_,_,F,y_ptr{y_void}).
candidate_field(C,RT) :-
    comp_field(C,_,_,F,y_comp{XC}), candidate_field(XC,XRT),
    trace_compose(fld{empty,F,C},XRT,RT).
candidate_trace(T,_,_,RT) :-
    cil_curfn(FN), trace_type(FN,T,y_comp{C}),
    candidate_field(C,RT).

predicate indir_target(P:pp,I:c_instr,FT:t_trace,FNG:g_guard).
icall(P,_,I), ~dircall(I,_), callfnptr(P,I,FT,FNG),
    guard_sat(FNG), ~trace_sub(FT,uc_exp{_,_,_},_),
    guard(P,PG), #and(FNG,PG,NFNG), guard_sat(NFNG),
    +indir_target(P,I,FT,NFNG), +add_poly_target(I,FT).

% funptr trace FT is passed T through argument AT, where AT is interesting
predicate indir_arg(I:c_instr,XFT:t_trace,
                    FT:t_trace,T:t_trace,AT:t_trace,G:g_guard).
indir_target(P,I,FT,FNG),
    cil_instr_call_arg(I,AN,AE), eval(P,AE,s_trace{T},TG),
    candidate_trace(T,I,drf{root{arg{AN}}},RT),
    trace_compose(drf{root{arg{AN}}},RT,AT),
    trace_compose_point(P,T,RT,s_trace{NT},RTG),
    guard_implies(FNG,TG), guard_implies(FNG,RTG),
    +indir_arg(I,FT,FT,NT,AT,FNG).
indir_target(P,I,FT,FNG),
    cil_instr_call_arg(I,AN,AE), eval(P,AE,s_trace{T},TG),
    candidate_trace(T,I,drf{root{arg{AN}}},RT),
    trace_compose(drf{root{arg{AN}}},RT,AT),
    trace_compose(T,RT,NT),
    guard_implies(FNG,TG),
    +indir_arg(I,FT,FT,NT,AT,FNG).

% remove the effects of unconstrained writes within calls, in cases where
% the funptr/data might have already been written within a callee
indir_arg(I,XFT,FT,T,AT,G), trace_sub(FT,uc_sum{_,_,ST},FTR),
    trace_compose(drf{ST},FTR,NFT), +indir_arg(I,XFT,NFT,T,AT,G).
indir_arg(I,XFT,FT,T,AT,G), trace_sub(T,uc_sum{_,_,ST},TR),
    trace_compose(drf{ST},TR,NT), +indir_arg(I,XFT,FT,NT,AT,G).

% account for intermediate writes which have filled in structure contents.
% TODO: this is kind of messy and conflates the program state at entry
% with the program state at the call site
~cil_curfn("drivers/usb/misc/auerswald.c:auerchain_unlink_all"),  % exception
    indir_arg(I,XFT,FT,T,AT,G), icall(P,_,I),
    trace_sub(FT,SFT,RFT), iset(XP,_,_), assign(XP,PT,s_trace{SFT},AG),
    pp_reach(XP,P), guard(XP,XG), #and(XG,AG,MG), guard_implies(G,MG),
    trace_compose(drf{PT},RFT,NFT), +indir_arg(I,XFT,NFT,T,AT,G).

% at call instruction I, it has been determined that T == NT and the two
% can be substituted for one another
predicate equality_substitute(I:c_instr,T:t_trace,NT:t_trace).
indir_arg(I,XFT,FT,T,AT,G), trace_sub(FT,MT,RT),
    equality_substitute(I,MT,NMT), ~trace_sub(NMT,MT,_),
    trace_compose(NMT,RT,NFT), +indir_arg(I,XFT,NFT,T,AT,G).

% find equalities we can use to substitute at call sites.
% this is particularly handy for functions which scan device lists looking
% for particular drivers, e.g. i2c_del_driver
icall(P,_,I), ~dircall(I,_), guard(P,G), #g_id(G,ID),
    ID=sc_cmp{_,s_trace{T0},s_trace{T1},false,LEN}, pointer_bits(LEN),
    cil_curfn(FN), trace_type(FN,T0,_), trace_type(FN,T1,_),
    #id_g(sc_cmp{b_eq,s_trace{T0},s_trace{T1},false,LEN},NG),
    guard_implies(G,NG),
    +equality_substitute(I,T0,T1), +equality_substitute(I,T1,T0).

predicate type_candidate(I:c_instr,FT:t_trace,T:t_trace,AT:t_trace,
                         C:string,STR:t_trace,FTR:t_trace,TR:t_trace).
indir_arg(I,XFT,FT,T,AT,_), trace_sub(FT,STR,FTR), trace_sub(T,STR,TR),
    trace_sub(FT,fld{STR,_,C},_), trace_sub(T,fld{STR,_,C},_),
    ~poly_candidate_ignore(FTR), ~poly_candidate_ignore(TR),
    +type_candidate(I,XFT,T,AT,C,STR,FTR,TR).
indir_arg(I,XFT,FT,T,AT,_), trace_sub(FT,WSTR,_), trace_sub(T,WSTR,WTR),
    trace_sub(FT,rfld{WSTR,WF,C},FTR), trace_sub(T,fld{WSTR,_,_},_),
    trace_simplify(rfld{WSTR,WF,C},STR),
    trace_compose(fld{empty,WF,C},WTR,TR),
    ~poly_candidate_ignore(FTR), ~poly_candidate_ignore(TR),
    +type_candidate(I,XFT,T,AT,C,STR,FTR,TR).
indir_arg(I,XFT,FT,T,AT,_), trace_sub(FT,WSTR,WFTR), trace_sub(T,WSTR,_),
    trace_sub(FT,fld{WSTR,_,_},_), trace_sub(T,rfld{WSTR,WF,C},TR),
    trace_simplify(rfld{WSTR,WF,C},STR),
    trace_compose(fld{empty,WF,C},WFTR,FTR),
    ~poly_candidate_ignore(FTR), ~poly_candidate_ignore(TR),
    +type_candidate(I,XFT,T,AT,C,STR,FTR,TR).
indir_arg(I,XFT,FT,T,AT,_), trace_sub(FT,WSTR,FTR), trace_sub(T,WSTR,TR),
    WSTR=drf{_}, trace_sub(AT,fld{_,F,C},TR),
    C="usb_interface",  % restricted to usb for now
    trace_compose(fld{empty,F,C},FTR,NFTR),
    trace_compose(fld{empty,F,C},TR,NTR),
    trace_simplify(rfld{WSTR,F,C},STR),
    ~poly_candidate_ignore(NFTR), ~poly_candidate_ignore(NTR),
    +type_candidate(I,XFT,T,AT,C,STR,NFTR,NTR).

predicate omit_field_indirect_write(in C:string,in F:string).

% T contains a subtrace drf{fld{_,F,C}} where field F might sometimes be
% indirectly written in the program.
predicate field_indirect_write(in T:t_trace).
field_indirect_write(T) :- trace_sub(T,drf{fld{_,F,C}},_),
    ~omit_field_indirect_write(C,F),
    escape_set(20,es_forward,tl_comp{C,fld{empty,F,C}},XSET),
    set_member(XSET,XL), XL\=tl_comp{C,fld{empty,F,C}},
    salias_used(XL)->swrite(_,_).
field_indirect_write(T) :- trace_sub(T,drf{fld{_,F,C}},_),
    ~omit_field_indirect_write(C,F),
    ~escape_set(20,es_forward,tl_comp{C,fld{empty,F,C}},_).

% type polymorphism: the function pointer and argument are in the same struct
type_candidate(I,FT,T,AT,C,STR,FTR,TR),
    ~(type_candidate(I,FT,T,AT,_,XSTR,_,_), XSTR\=STR, trace_sub(XSTR,STR,_)),
    ~field_indirect_write(FTR), ~field_indirect_write(TR),
    +add_poly_comp(I,FT,C,FTR,TR,AT).

% function polymorphism: the function pointer and argument are separate pieces
% of data passed into the current function
indir_arg(I,FT,_,T,AT,_),
    trace_root(FT,arg{A0}), trace_root(T,arg{A1}), A0\=A1,
    ~poly_candidate_ignore(FT), ~poly_candidate_ignore(T),
    ~type_candidate(I,FT,T,AT,_,_,_,_),
    ~field_indirect_write(FT), ~field_indirect_write(T),
    +add_poly_func(I,FT,T,AT).

% BAKED

% add netdev_priv(net_device) when passing in net devices to indirect calls
candidate_field(C,fld{empty,F,C}) :-
    C="net_device", netdev_priv_field(F).

% for notify structure breakage, e.g. irttp_disconnect_indication
+omit_comp_impure("__anonstruct_notify_t_71","instance",_).

% &private_data is taken in isdn_open, where it is an integer rather than
% pointer data
%+omit_field_indirect_write("file","private_data").

% pcmcia_request_irq passes &dev.driver in as the data with an IRQ
% as a mechanism to identify the driver. weird! the pointer
% won't be used for writing
+omit_field_indirect_write("device","driver").
