import "aliasing_sum.clp".
import "../smemory/memory.clp".
import "../smemory/scalar_sat.clp".
import "../smemory/aliasing_full.clp".
import "../base/sumcallers.clp".

using set.
using boundops.

type t_zerofield_boundset.

predicate zero_step(in t_zerofield_boundset,
                    in FN:string,in SUM:sum,in ENTRY:bool,in T:t_trace,
                    XFN:string,XSUM:sum,XP:pp,XT:t_trace).
predicate zero_step_threshold(in t_zerofield_boundset,
                    in FN:string,in SUM:sum,in ENTRY:bool,in T:t_trace).

predicate zero_step_failed(in FN:string,in SUM:sum,in P:pp,in T:t_trace).
predicate zero_step_next(in FN:string,in SUM:sum,in P:pp,in T:t_trace,
                         XFN:string,XSUM:sum,XENTRY:bool,XT:t_trace).

predicate omit_zero_initialized_fail(in FN:string,in SUM:sum,
                                     in P:pp,in T:t_trace).

type t_zero_fail_why ::=
    zf_prop_failed{FN:string,SUM:sum,P:pp,T:t_trace}
  | zf_transitive{FN:string,SUM:sum,P:pp,T:t_trace}
  | zf_threshold.

type t_zerofield_node ::= zf_node{FN:string,SUM:sum,P:pp,T:t_trace}.
type t_zerofield_boundset = boundset[t_zerofield_node,t_zerofield_node].

predicate zero_initialized_fail(in t_zerofield_boundset,
                                in FN:string,in SUM:sum,in P:pp,in T:t_trace,
                                Y:t_zero_fail_why).
zero_initialized_fail(X,FN,SUM,P,T,zf_prop_failed{FN,SUM,P,T}) :-
    ~omit_zero_initialized_fail(FN,SUM,P,T),
    zero_step_failed(FN,SUM,P,T).
zero_initialized_fail(X,FN,SUM,P,T,zf_threshold) :-
    ~omit_zero_initialized_fail(FN,SUM,P,T),
    zero_step_next(FN,SUM,P,T,MFN,MSUM,MENTRY,MT),
    zero_step_threshold(X,MFN,MSUM,MENTRY,MT).
zero_initialized_fail(X,FN,SUM,P,T,zf_transitive{YFN,YSUM,YP,YT}) :-
    ~omit_zero_initialized_fail(FN,SUM,P,T),
    zero_step_next(FN,SUM,P,T,MFN,MSUM,MENTRY,MT),
    ~zero_step_threshold(X,MFN,MSUM,MENTRY,MT),
    zero_step(X,MFN,MSUM,MENTRY,MT,YFN,YSUM,YP,YT),
    zero_initialized_fail(X,YFN,YSUM,YP,YT,_).

predicate zero_initialized(in P:pp,in T:t_trace).
zero_initialized(P,drf{T}) :- cursum(FN,SUM),
    boundset_create(zf_node{FN,SUM,P,T},10,X),
    ~zero_initialized_fail(X,FN,SUM,P,T,_).

% PROPAGATION RULES

zero_step(X,FN,SUM,true,T,XFN,XSUM,XP,XT) :-
    cil_sum_caller(FN,SUM)->caller(XFN,XSUM,XP,XI),
    boundset_insert(X,zf_node{XFN,XSUM,XP,empty}),
    inst_trace(XI,p_remote{XFN,XSUM,XP},T,s_trace{XT},_).
zero_step_threshold(X,FN,SUM,true,T) :-
    cil_sum_caller(FN,SUM)->caller(XFN,XSUM,XP,_),
    ~boundset_insert(X,zf_node{XFN,XSUM,XP,empty}).

zero_step(X,FN,SUM,false,T,FN,SUM,P,XT) :-
    cil_sum_body(FN,SUM)->sum_bound(_,_,P),
    boundset_insert(X,zf_node{FN,SUM,P,empty}),
    convert_point_trace(p_remote{FN,SUM,P},T,s_trace{XT},G), guard_sat(G).
zero_step_threshold(X,FN,SUM,false,T) :-
    cil_sum_body(FN,SUM)->sum_bound(_,_,P),
    ~boundset_insert(X,zf_node{FN,SUM,P,empty}).

predicate entry_filter(in FN:string,in SUM:sum,in T:t_trace).
entry_filter(_,SUM,T) :- trace_alloc(SUM,T,ak_external), ~trace_root(T,glob{_}),
    ~trace_sub(T,rfld{_,_,_},_), ~trace_sub(T,index{_,_,_},_),
    trace_derefs(T,N), int_le(N,3).

predicate exit_filter(in FN:string,in SUM:sum,in T:t_trace,
                      CFN:string,CSUM:sum,CT:t_trace).
exit_filter(FN,SUM,T,CFN,CSUM,CT) :-
    ~trace_sub(T,rfld{_,_,_},_), ~trace_sub(T,index{_,_,_},_),
    match_call_exit(FN,SUM,T,_,_,CFN,CSUM,CT).

predicate has_original_value(in FN:string,in SUM:sum,in P:pp,in T:t_trace).
has_original_value(FN,SUM,P,T) :-
    ~(val(p_remote{FN,SUM,P},T,V,_), V\=s_trace{drf{T}}, V\=s_const{0}).

zero_step_failed(FN,SUM,P,T) :-
    ~has_original_value(FN,SUM,P,T),
    ~(trace_sub(T,drf{XT},_), allocate_filter(FN,SUM,XT)).

zero_step_failed(FN,SUM,P,T) :- has_original_value(FN,SUM,P,T),
    ~allocate_filter(FN,SUM,T),
    ~entry_filter(FN,SUM,T), ~exit_filter(FN,SUM,T,_,_,_).

zero_step_next(FN,SUM,P,T,FN,SUM,true,T) :- has_original_value(FN,SUM,P,T),
    ~allocate_filter(FN,SUM,T), entry_filter(FN,SUM,T).

zero_step_next(FN,SUM,P,T,CFN,CSUM,false,CT) :- has_original_value(FN,SUM,P,T),
    ~allocate_filter(FN,SUM,T), exit_filter(FN,SUM,T,CFN,CSUM,CT).
