% check per-type safety

import "typecheck.clp".

analyze session_name("cil_comp").
analyze analysis_name("casting_comp").

analyze priority(1).

% PREDICATES

type h_check ::=
     % check always holds
    hc_true{maybe[pp],TEXT:string}
     % check never holds
  | hc_false{maybe[pp],TEXT:string}
     % TP is preserved within XHKL relative to STR
  | hc_preserve{TP:t_property,XHKL:h_kind_loc,STR:t_trace,P:pp}.

% RULES

% fill in current location info for typecheck.clp

cil_comp(COMP), +curloc(hl_comp{COMP}),
    cil_comp_location(_,_,LINE), +curloc_line(LINE),
    comp_location(COMP,FILE,MINLINE,MAXLINE), STITLE="type "^COMP,
    +scasting_post(hl_comp{COMP})->
        sloc_info(COMP,STITLE,FILE,MINLINE,MAXLINE),
    casting_display_name(hl_comp{COMP},td_casting_overall,DNAME),
    +search_add(COMP,DNAME).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECKING FAILURES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ------------------------------------ hc_true

+check_info(hc_true{_,TEXT},no,TEXT).

% ------------------------------------ hc_false

+check_fail(hc_false{_,_}).

+check_info(hc_false{_,TEXT},no,TEXT).

% ------------------------------------ hc_preserve

check_require(hc_preserve{TP,XHKL,STR,P},XHKL,XHK) :-
    XHK = hk_preserve{P,TP,lw_comp{STR}}.

check_info(hc_preserve{TP,_,_,_},no,TEXT) :- property_dstring(TP,TEXT).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECKING METHODS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ------------------------------------ hk_invariant

predicate relevant_trace(in RT:t_trace,XHKL:h_kind_loc,STR:t_trace,P:pp).
relevant_trace(RT,XHKL,STR,P) :-
    cil_comp(C), trace_relative(RT), ~trace_derefs(RT,0), trace_sub(RT,XRT,WRT),
    (salias_escape(tl_comp{C,XRT},false)->sescape(W,SL,_);
     ~trace_derefs(WRT,0), salias_escape(tl_comp{C,XRT},true)->sescape(W,SL,_)),
    (SL=tl_func{FN,FT,_}, W=te_func{FN,SUM,P}, trace_sub(FT,STR,XRT), XHKL=hl_func{FN,SUM};
     SL=tl_glob{GT}, W=te_glob{G}, P=p_abort, trace_sub(GT,STR,XRT), XHKL=hl_glob{G}).

predicate relevant_property(in RTP:t_property,XHKL:h_kind_loc,STR:t_trace,P:pp).
relevant_property(RTP,XHKL,STR,P) :-
    property_trace_used(RTP,RT,_), relevant_trace(RT,XHKL,STR,P).

method_check(hk_invariant{RTP},hm_local,HC) :-
    ~trivial_property(RTP,p_abort,_,_),
    (relevant_property(RTP,XHKL,STR,P), reptrace_property(empty,STR,RTP,TP),
         HC = hc_preserve{TP,XHKL,STR,P};
     ~relevant_property(RTP,_,_,_), HC = hc_true{no,"Fields are never written"}).

% baked in behavior

import "typebaked_check.clp".
