% post-mortem analysis of derivations performed and proved/failed facts

import "typehdr.clp".

using display.

% ANALYSIS

analyze session_name("scasting_post").
analyze analysis_name("xcasting_derive_post").

% PREDICATES

predicate casting_display_add(DNAME:string,CAT:string,
                              TITLE:string,ITEMS:list[displayitem]).
sloc_info(SNAME,_,FILE,MINLINE,MAXLINE), int_add(MINLINE,5,NLINE),
    casting_display_add(DNAME,CAT,TITLE,ITEMS),
    +display_add(DNAME,true,CAT,TITLE,SNAME,
                 FILE,MINLINE,MAXLINE,NLINE,ITEMS).

% RULES

scasting_post(HKL), sprint_fail(HK,STR),
    ~scasting_proved(HKL)->scheckh(HK), +print_string(STR).

% display styles/status

predicate check_status(in HC:h_check_hash,STYLE:string,LINKS:list[displaylink])
    succeeds [once].
check_status(HC,"check_prove",[]) :-
    ~scheck_fail(HC), ~scheck_require(HC,_,_,_).
check_status(HC,"check_prove",[d_link{DNAME,LINK}]) :-
    scheck_require(HC,XHKL,XHK,LINK),
    scasting_proved(XHKL)->scheckh(XHK),
    casting_display_name(XHKL,td_casting_kind{XHK},DNAME).
check_status(HC,"check_fail",[]) :-
    scheck_fail(HC).
check_status(HC,"check_fail",[d_link{DNAME,LINK}]) :-
    scheck_require(HC,XHKL,XHK,LINK),
    ~scasting_proved(XHKL)->scheckh(XHK),
    (scasting_post(XHKL); scasting_tried(XHKL)->scheckh(XHK)),
    casting_display_name(XHKL,td_casting_kind{XHK},DNAME).
check_status(HC,"check_ignore",[d_link{DNAME,LINK}]) :-
    scheck_require(HC,XHKL,XHK,LINK),
    ~scasting_proved(XHKL)->scheckh(XHK),
    ~scasting_post(XHKL), ~scasting_tried(XHKL)->scheckh(XHK),
    casting_display_name(XHKL,td_casting_overall,DNAME).

predicate method_fail(in HK:h_kind_hash,in HM:h_method).
method_fail(HK,HM) :- scheck_info(HK,HM,HC,_,_), scheck_fail(HC).
method_fail(HK,HM) :- scheck_info(HK,HM,HC,_,_), scheck_require(HC,XHKL,XHK,_),
    ~scasting_proved(XHKL)->scheckh(XHK).

predicate method_style(in HK:h_kind_hash,in HM:h_method,
                       TITLE:string,STYLE:string) succeeds [once].
method_style(HK,HM,TITLE,"method_fail") :-
    method_fail(HK,HM), scheck_name(HK,_,_,HKS), tostring(HM,HMS),
    TITLE = "Failure to prove " ^ HKS ^ " via " ^ HMS.
method_style(HK,HM,TITLE,"method_prove") :-
    ~method_fail(HK,HM), scheck_name(HK,_,_,HKS), tostring(HM,HMS),
    TITLE = "Proved " ^ HKS ^ " via " ^ HMS.

predicate kind_fail(in HK:h_kind_hash).
kind_fail(HK) :- ~(scheck_info(HK,HM,_,_,_), ~method_fail(HK,HM)).

predicate kind_style(in HK:h_kind_hash,STYLE:string).
kind_style(HK,"exp_kind_fail")  :- kind_fail(HK), sprint_fail(HK,_).
kind_style(HK,"kind_fail")      :- kind_fail(HK), ~sprint_fail(HK,_).
kind_style(HK,"exp_kind_prove") :- ~kind_fail(HK), sprint_fail(HK,_).
kind_style(HK,"kind_prove")     :- ~kind_fail(HK), ~sprint_fail(HK,_).

% display item generation

predicate check_item(HK:h_kind_hash,HM:h_method,ITEM:displayitem).
scheck_info(HK,HM,HC,LINE,TEXT), check_status(HC,STYLE,LINKS),
    +check_item(HK,HM,d_line_text{LINE,STYLE,d_bottom,TEXT,LINKS}).

predicate method_item(HK:h_kind_hash,ITEM:displayitem).
scasting_post(HKL), scheck_name(HK,_,LINE,_), scheck_info(HK,HM,_,_,_),
    method_style(HK,HM,TITLE,STYLE),
    \/check_item(HK,HM,CITEM):list_all(CITEM,CITEMS),
    casting_display_name(HKL,td_casting_kind{HK},BDNAME),
    NITEMS=[d_line_text{LINE,"back_kind",d_right,"All proof attempts",
                        [d_link{BDNAME,"link"}]}|CITEMS],
    casting_display_name(HKL,td_casting_method{HK,HM},DNAME),
    +casting_display_add(DNAME,"cast_method",TITLE,NITEMS),
    +method_item(HK,d_line_text{LINE,STYLE,d_bottom,TITLE,
                                [d_link{DNAME,"link"}]}).

predicate check_why(in HK:h_kind_hash,
                    XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method).
check_why(HK,XHKL,XHK,XHM) :- scasting_post(HKL),
    (scasting_derive(HKL)->sfirststep(HK,XHKL,XHK,XHM);
     scasting_derive(HKL)->snextstep(HK,XHKL,XHK,XHM,_,_)).

predicate kind_item(ITEM:displayitem).
scasting_post(HKL),
    scheck_name(HK,_,LINE,HKS), kind_style(HK,STYLE),
    TITLE = "Attempts to prove " ^ HKS,
    (scasting_tried(HKL)->scheckh(HK),
         tostring(HK,HSH), tostring(HKL,HKLS),
         NTITLE = TITLE^" ['"^HKLS^"' "^HSH^"]";
     ~scasting_tried(HKL)->scheckh(HK), NTITLE = TITLE),
    \/method_item(HK,MITEM):list_all(MITEM,MITEMS),
    \/(check_why(HK,XHKL,XHK,XHM),
       casting_display_name(XHKL,td_casting_method{XHK,XHM},XDNAME),
       YLINK=d_link{XDNAME,"why"}):list_all(YLINK,YLIST),
    (YLIST=[], NITEMS=MITEMS;
     YLIST\=[], NITEMS=[d_line_text{LINE,"back_why",d_right,
                                    "Reasons for adding",YLIST}|MITEMS]),
    casting_display_name(HKL,td_casting_kind{HK},DNAME),
    +casting_display_add(DNAME,"cast_kind",NTITLE,NITEMS),
    +kind_item(d_line_text{LINE,STYLE,d_bottom,
                           TITLE,[d_link{DNAME,"link"}]}).

scasting_post(HKL), sloc_info(_,STITLE,_,_,_),
    TITLE = "All proof attempts within " ^ STITLE,
    \/kind_item(ITEM):list_all(ITEM,ITEMS),
    casting_display_name(HKL,td_casting_overall,DNAME),
    +casting_display_add(DNAME,"cast_overall",TITLE,ITEMS).
