% post-mortem analysis that scans to see all derivations that were
% needed to prove a particular cast

import "typehdr.clp".

% ANALYSIS

analyze session_name("scasting_post").
analyze analysis_name("xcasting_derive_scan").

% PREDICATES

predicate method_any(in HKL:h_kind_loc,in HK:h_kind_hash,HM:h_method).
method_any(HKL,HK,HM) :- scasting_post(HKL)->scheck_info(HK,HM,_,_,_).

predicate method_fail(in HKL:h_kind_loc,in HK:h_kind_hash,in HM:h_method).
method_fail(HKL,HK,HM) :-
    scasting_post(HKL)->scheck_info(HK,HM,HC,_,_),
    scasting_post(HKL)->scheck_fail(HC).
method_fail(HKL,HK,HM) :-
    scasting_post(HKL)->scheck_info(HK,HM,HC,_,_),
    scasting_post(HKL)->scheck_require(HC,XHKL,XHK,_),
    ~scasting_proved(XHKL)->scheckh(XHK).

% for a method which did not fail, what are the other proofs it depends on
predicate method_depends(in HKL:h_kind_loc,in HK:h_kind_hash,in HM:h_method,
                         XHKL:h_kind_loc,XHK:h_kind_hash).
method_depends(HKL,HK,HM,XHKL,XHK) :-
    scasting_post(HKL)->scheck_info(HK,HM,HC,_,_),
    scasting_post(HKL)->scheck_require(HC,XHKL,XHK,_).

% we have proved HK on the current location, and all the XHKL/XHK
% were needed to do so
predicate proof_dependency(HK:h_kind_hash,XHKL:h_kind_loc,XHK:h_kind_hash).

% we have proved HK on the current location, and extra info HX
% was needed to do so
predicate proof_extra(HK:h_kind_hash,HX:h_method_extra).

% XHKL/XHK was proved using XHM and is used to prove HK on this loc.
% add any extra information associated with XHM
predicate proof_method(HK:h_kind_hash,XHKL:h_kind_loc,XHK:h_kind_hash,
                       XHM:h_method).
proof_method(HK,XHKL,XHK,XHM),
    scasting_post(XHKL)->scheck_extra(XHK,XHM,HX),
    +proof_extra(HK,HX).

predicate str_dependency(HKL:h_kind_loc,HK:string,XHKL:h_kind_loc,XHK:string).
?- str_dependency(HKL,HK,XHKL,XHK).

predicate str_extra(HKL:h_kind_loc,HK:string,HX:h_method_extra).
?- str_extra(HKL,HK,HX).

scasting_post(HKL), proof_dependency(HK,XHKL,XHK),
    scheck_name(HK,HKS,_,_), scasting_post(XHKL)->scheck_name(XHK,XHKS,_,_),
    +str_dependency(HKL,HKS,XHKL,XHKS).

scasting_post(HKL), proof_extra(HK,HX),
    scheck_name(HK,HKS,_,_), +str_extra(HKL,HKS,HX).

% RULES

% HKL/HK was proved trivially
predicate follow_trivial(in HKL:h_kind_loc,in HK:h_kind_hash,HM:h_method).
follow_trivial(HKL,HK,HM) :-
    method_any(HKL,HK,HM), ~method_fail(HKL,HK,HM),
    ~method_depends(HKL,HK,HM,_,_).

% proving XKL/HK required that XHKL/XHK be proved.
% will not generate any tuples of HKL/HK was proved trivially.
predicate follow_depends(in HKL:h_kind_loc,in HK:h_kind_hash,HM:h_method,
                         XHKL:h_kind_loc,XHK:h_kind_hash).
follow_depends(HKL,HK,HM,XHKL,XHK) :-
    ~follow_trivial(HKL,HK,_),
    \/(method_any(HKL,HK,XHM), ~method_fail(HKL,HK,XHM))
      :list_all(XHM,[HM|_]),
    method_depends(HKL,HK,HM,XHKL,XHK).

% kinds we are interested in getting the proof hierarchy for.
% HKL is the current location, HK is an explicit cast proved safe in it
predicate initial_kind(HKL:h_kind_loc,HK:h_kind_hash).
scasting_post(HKL), sprint_fail(HK,_),
    scasting_proved(HKL)->scheckh(HK), +initial_kind(HKL,HK).

initial_kind(HKL,HK), follow_trivial(HKL,HK,HM),
    +proof_method(HK,HKL,HK,HM).
initial_kind(HKL,HK), follow_depends(HKL,HK,HM,XHKL,XHK),
    +proof_method(HK,HKL,HK,HM), +proof_dependency(HK,XHKL,XHK).
proof_dependency(HK,XHKL,XHK), follow_trivial(XHKL,XHK,XHM),
    +proof_method(HK,XHKL,XHK,XHM).
proof_dependency(HK,XHKL,XHK), follow_depends(XHKL,XHK,XHM,YHKL,YHK),
    +proof_method(HK,XHKL,XHK,XHM), +proof_dependency(HK,YHKL,YHK).

% fill in scasting_restrict summaries

import "typeuse.clp".

%scasting_post(HKL), proof_extra(HK,hx_future_preserves{C,F}),
%    scheck_name(HK,HKS,_,_),
%    +scasting_restrict_field(C,F)->wrestrict_field(HKL,HKS).

%scasting_post(HKL), proof_extra(HK,hx_future_assign{FN,SUM,P,PT,V}),
%    scheck_name(HK,HKS,_,_),
%    +scasting_restrict_assign(FN,SUM)->wrestrict_assign(P,PT,V,HKL,HKS).
