% initial population of sites where casts occur and other information.
% this analysis is (mostly) intraprocedural and runs as only a single pass

import "typeuse.clp".
import "../daliasing/zerofield.clp".
import "../base/sumcallers.clp".

analyze session_name("cil_sum_body").
analyze analysis_name("init_casting_func").

analyze priority(1).

% OUTPUT 0PREDICATES

predicate str_typed_use(FN:string,SUM:sum,P:pp,T:string,V:t_typed_val).
?- str_typed_use(FN,SUM,P,T,V).

predicate str_typed_write(FN:string,SUM:sum,P:pp,T:string,V:t_typed_val).
?- str_typed_write(FN,SUM,P,T,V).

predicate str_typed_use_field(C:string,F:string,
                              FN:string,SUM:sum,P:pp,T:string,V:t_typed_val).
?- str_typed_use_field(C,F,FN,SUM,P,T,V).

predicate str_typed_explicit(FN:string,SUM:sum,T:string,V:t_typed_val).
?- str_typed_explicit(FN,SUM,T,V).

predicate str_allocate_untyped(FN:string,T:string).
?- str_allocate_untyped(FN,T).

% STATS OUTPUT PREDICATES

predicate str_ignored_use(FN:string,SUM:sum,P:pp,T:string,V:t_typed_val).
?- str_ignored_use(FN,SUM,P,T,V).

predicate str_subtyped_use(FN:string,SUM:sum,P:pp,T:string,V:t_typed_val,PV:t_typed_val).
?- str_subtyped_use(FN,SUM,P,T,V,PV).

% RULES

predicate add_explicit_fixed(FN:string,SUM:sum,T:t_trace,V:t_typed_val).
add_explicit_fixed(FN,SUM,T,V),
    +scasting_explicit(FN,SUM)->sexplicit_fixed(T,V),
    trace_string(T,TS), +str_typed_explicit(FN,SUM,TS,V).

import "typebaked_init.clp".

predicate typed_use(P:pp,T:t_trace,V:t_typed_val,G:g_guard,WRITE:bool).
predicate ignore_typed_use(in P:pp,in T:t_trace,in V:t_typed_val,in G:g_guard).

% ignore all casts to unions and to types that do not contain pointers
% that will ever be dereferenced
ignore_typed_use(_,_,C,_) :- scasting_type_pointer(C)->wpointer_field(false,_).
ignore_typed_use(_,_,C,_) :- scasting_type_pointer(C)->wpointer_field(_,false).
ignore_typed_use(_,_,C,_) :- cil_comp(C)->cil_comp_name(_,_,false).

% compute locations of explicit casts

% V is a supertype of the explicit type of T
predicate explicit_super_type(in T:t_trace,in V:t_typed_val).
explicit_super_type(T,V) :-
    explicit_type(T,XV), check_subtype(y_comp{XV},y_comp{V}).

cursum(FN,SUM), typed_use(P,T,V,G,WRITE), ~explicit_super_type(T,V),
    ~(WRITE=false, ignore_typed_use(P,T,V,G)),
    +scasting_explicit(FN,SUM)->sexplicit(P,T,V,G,WRITE),
    HK=hk_prior{l_value{T,V,G},P,true},
    tostring(FN,FNS), tostring(SUM,SUMS),
    tostring(P,PS), trace_string(T,TS), tostring(TS,TSS), tostring(V,VS),
    (WRITE=true, +str_typed_write(FN,SUM,P,TS,V), PRED="str_unsafe_write";
     WRITE=false, +str_typed_use(FN,SUM,P,TS,V), PRED="str_unsafe_use"),
    STR = PRED^"("^FNS^","^SUMS^","^PS^","^TSS^","^VS^").",
    hash_kind(HK,HKH),
    +scasting_post(hl_func{FN,SUM})->sprint_fail(HKH,STR),
    +scasting_try(hl_func{FN,SUM})->scheck(HK),
    scasting_proved(hl_func{FN,SUM})->scheckh(HKH),
    +scasting_explicit_safe(FN,SUM)->sexplicit(P,T,V,G,WRITE).

cursum(FN,SUM), typed_use(P,T,V,G,false),
    ~explicit_super_type(T,V), ~ignore_typed_use(P,T,V,G),
    (T=drf{fld{_,F,C}}; T=uc_sum{_,_,fld{_,F,C}}), trace_string(T,TS),
    +str_typed_use_field(C,F,FN,SUM,P,TS,V).

% any casts of argument traces we see in this function we should tell the
% polydata.clp analysis about for looking at funptr correlations
cursum(FN,s_func), typed_use(P,T,V,G,false),
    ~explicit_super_type(T,V), ~ignore_typed_use(P,T,V,G),
    trace_root(T,arg{_}), ~poly_candidate_ignore(T),
    +spoly_call_try(FN)->wpoly_call_try(T,y_comp{V}).

% fill in scasting_explicit_field with any explicitly used fields.
% TODO: this misses cases where the field has its address taken via a void*
%   which is cast to something later in the program
cursum(FN,SUM), typed_use(P,T,V,G,false),
    ~explicit_super_type(T,V), ~ignore_typed_use(P,T,V,G),
    T=fld{_,F,C}, +scasting_explicit_field(C,F)->sexplicit_field(V,FN,SUM).

cil_exp_cast(E,_,TYP), convert_type(TYP,y_ptr{y_comp{C}}),
    exp_point(E,P), eval(P,E,s_trace{T},EG),
    guard(P,PG), #and(PG,EG,WG), guard_sat(WG), #bool_g(true,G),
    +typed_use(P,T,C,G,false).
cursum(FN,_), icall(P,_,I),
    callret(P,I,RT,_), trace_type(FN,RT,y_ptr{y_comp{C}}),
    anycall(I,CFN,_), trace_type(CFN,root{return},Y), Y\=y_ptr{y_comp{C}},
    assign(P,RT,s_trace{T},VG),
    guard(P,PG), #and(PG,VG,WG), guard_sat(WG), #bool_g(true,G),
    +typed_use(P,T,C,G,false).

% ignored explicit casts and subtype casts

cursum(FN,SUM), typed_use(P,T,V,G,false),
    ~explicit_super_type(T,V), ignore_typed_use(P,T,V,G),
    trace_string(T,TS),
    +str_ignored_use(FN,SUM,P,TS,V).

cursum(FN,SUM), typed_use(P,T,V,_,false),
    ~explicit_super_type(T,V), % ~ignore_typed_use(P,T,V,G),
    explicit_type(T,XV), check_subtype(y_comp{V},y_comp{XV}),
    +scasting_subtype(V)->wsubtype(XV),
    trace_string(T,TS),
    +str_subtyped_use(FN,SUM,P,TS,V,XV).

% explicit casts which directly depend on polymorphic data constraints

predicate str_typed_poly(FN:string,SUM:sum,P:pp,T:string,V:t_typed_val,
                         C:string,FT:string,DT:string,AT:string).
?- str_typed_poly(FN,SUM,P,T,V,C,FT,DT,AT).

predicate str_typed_poly_fail(FN:string,SUM:sum,P:pp,T:string,V:t_typed_val,
                              C:string,FT:string,DT:string,AT:string).
?- str_typed_poly_fail(FN,SUM,P,T,V,C,FT,DT,AT).

cursum(FN,SUM), typed_use(P,T,V,G,false),
    ~explicit_super_type(T,V), ~ignore_typed_use(P,T,V,G),
    follow_impure_values(P,s_trace{T},_,s_trace{NT}),
    SUM=s_func, cil_sum_caller(FN,SUM)->caller(PFN,_,_,s_call{PI}),
    sum_funptr(PFN)->sindirect(PI,_),
    trace_sub(NT,ST,_), spoly_call(PFN,PI)->wpoly_call_comp(_,C,FT,DT,ST),
    trace_string(T,TS), trace_string(FT,FTS),
    trace_string(DT,DTS), trace_string(ST,STS),
    +str_typed_poly(FN,SUM,P,TS,V,C,FTS,DTS,STS),
    spoly_receive_comp(C,FT,DT,ST)->wpoly_fail(),
    +str_typed_poly_fail(FN,SUM,P,TS,V,C,FTS,DTS,STS).

% compute set of allocators which do not generate typed data

% maybe T was assigned some typed data in this function. this is just a guess.
predicate maybe_typed(in T:t_trace).
maybe_typed(T) :- exit(EXITP), val(EXITP,T,s_trace{XT},_),
    access(_,WT,_), trace_sub(WT,fld{XT,_,_},_).
maybe_typed(T) :- exit(EXITP), val(EXITP,T,s_trace{XT},_),
    trace_sub(XT,BT,_), allocate_root(BT,I,CT,_), icall(P,_,I),
    allocate_call(I,P,FN,CT,true,_),
    ~scasting_allocate_untyped(FN)->wfunc_allocate(CT,_).

% add untyped allocations where we don't think the trace was cast into
% anything. if we guessed wrong about the cast we are still sound.
% - if there was no cast in the function and we guessed there was, we're fine.
% - if there was a cast in the function and we guessed not, we will add
%   wfunc_untyped and won't be able to verify the function's cast.
cursum(FN,s_func), salias_write_func(FN,s_func)->wfunc_allocate(T,ESCAPE),
    ~maybe_typed(T), +scasting_allocate_untyped(FN)->wfunc_allocate(T,ESCAPE),
    trace_string(T,TS), +str_allocate_untyped(FN,TS),
    exit(EXITP), val(EXITP,T,s_trace{XT},_),
    +scasting_allocate_untyped(FN)->wfunc_untyped(XT).

% identify types for freshly allocated locations

icall(P,_,I), allocate_call(I,P,AFN,CT,WRAP,_), allocate_root(DT,I,CT,_),
    (WRAP=false;
     WRAP=true, scasting_allocate_untyped(AFN)->wfunc_allocate(CT,_)),
    typed_use(_,DT,V,_,false), ~(typed_use(_,DT,XV,_,false), XV\=V),
    cursum(FN,SUM), +add_explicit_fixed(FN,SUM,DT,V).

% identify impure writes of fields assumed to be type preserving

predicate use_impure_type(in T:t_trace,V:t_typed_val).
use_impure_type(T,V) :- explicit_type(T,V).
use_impure_type(T,V) :- cursum(FN,SUM), add_explicit_fixed(FN,SUM,T,V).

iset(P,_,_), assign(P,T,s_trace{DT},TG), T=fld{_,F,C},
    \/scasting_restrict_field(C,F)->wrestrict_field(HKL,_):count_all(HKL,N),
    int_ge(N,5),
    guard(P,PG), #and(TG,PG,G), guard_sat(G),
    (use_impure_type(DT,V); ~use_impure_type(DT,_), V="<unknown>"),
    #bool_g(true,WG), +typed_use(P,DT,V,WG,true).
