% additional baked-in information about types. only runs during init_casting_func pass

predicate sizeof_exp(in E:c_exp,C:string).
sizeof_exp(E,C) :- cil_exp_sizeof(E,TYP), cil_type_comp(TYP,C).
sizeof_exp(E,C) :- cil_exp_sizeof(E,TYP), cil_type_named(TYP,_,NTYP), cil_type_comp(NTYP,C).
sizeof_exp(E,C) :- cil_exp_sizeofe(E,ETYP), exp_type(ETYP,y_comp{C}).
sizeof_exp(E,C) :- cil_exp_cast(E,CE,_), sizeof_exp(CE,C).

% linux - per_cpu structures

% ignore casts of values fetched via the per_cpu macros, this is an internal mechanism
cil_exp_binop(E,b_plusa,LE,RE,_), exp_point(E,P),
    eval(P,RE,s_trace{drf{index{root{glob{"__per_cpu_offset"}},_,_}}},_),
    eval(P,LE,s_trace{T},_),
    +ignore_typed_use(_,T,_,_).

% linux - structures with physical memory pointers

% these contain pointers, but I think they refer to addresses in physical
% memory that the driver uses to communicate with the device. ignore casts on them
+ignore_typed_use(_,_,"__anonstruct_PORT_12",_).
+ignore_typed_use(_,_,"__anonstruct_I596_SCB_378",_).
+ignore_typed_use(_,_,"_I596_RFD",_).
+ignore_typed_use(_,_,"_I596_RBD",_).
+ignore_typed_use(_,_,"_I596_TBD",_).

% linux - net_device

predicate add_netdev_setup(FN:string,COMP:string).
add_netdev_setup(FN,COMP), netdev_priv(drf{root{arg{0}}},APRIV),
    +add_explicit_fixed(FN,s_func,APRIV,COMP).

predicate add_netdev_return(P:pp,I:c_instr,COMP:string).
add_netdev_return(P,I,COMP), callret(P,I,RT,_),
    netdev_priv(uc_sum{s_call{I},P,RT},RPRIV),
    cursum(FN,SUM), +add_explicit_fixed(FN,SUM,RPRIV,COMP).

predicate unknown_alloc_netdev(FN:string,LINE:int).
?- unknown_alloc_netdev(FN,LINE).

predicate handled_netdev(FN:string).

dircall(I,"alloc_netdev"), icall(P,_,I),
    cil_instr_call_arg(I,0,ESIZE), cil_instr_call_arg(I,2,ESETUP),
    sizeof_exp(ESIZE,COMP), eval(P,ESETUP,s_trace{root{glob{SETUPFN}}},_),
    +add_netdev_setup(SETUPFN,COMP), +add_netdev_return(P,I,COMP).

dircall(I,"alloc_netdev"), icall(P,_,I),
    cil_instr_call_arg(I,0,ESIZE), cil_instr_call_arg(I,2,ESETUP),
    (~sizeof_exp(ESIZE,_), ~cil_exp_x_intval(ESIZE,0); ~eval(P,ESETUP,s_trace{root{glob{_}}},_)),
    cil_curfn(XFN), ~handled_netdev(XFN),
    point_location(P,_,LINE), +unknown_alloc_netdev(XFN,LINE).

predicate alloc_netdev_wrapper(FN:string,A:int).
alloc_netdev_wrapper(FN,_), +handled_netdev(FN).

% functions which wrap alloc_netdev, using a generic setup function
% and take a private data size argument
+alloc_netdev_wrapper("alloc_etherdev",0).
+alloc_netdev_wrapper("alloc_irdadev",0).
+alloc_netdev_wrapper("alloc_ltalkdev",0).
+alloc_netdev_wrapper("alloc_fcdev",0).
+alloc_netdev_wrapper("alloc_hippi_dev",0).
+alloc_netdev_wrapper("alloc_fddidev",0).
+alloc_netdev_wrapper("alloc_trdev",0).

dircall(I,FN), alloc_netdev_wrapper(FN,A), icall(P,_,I),
    cil_instr_call_arg(I,A,ESIZE), sizeof_exp(ESIZE,COMP),
    callret(P,I,RT,_), netdev_priv(uc_sum{s_call{I},P,RT},RPRIV),
    cursum(XFN,SUM), +add_explicit_fixed(XFN,SUM,RPRIV,COMP).

dircall(I,FN), alloc_netdev_wrapper(FN,A), icall(P,_,I),
    cil_instr_call_arg(I,A,ESIZE), ~sizeof_exp(ESIZE,_), ~cil_exp_x_intval(ESIZE,0),
    cil_curfn(XFN), ~handled_netdev(XFN),
    point_location(P,_,LINE), +unknown_alloc_netdev(XFN,LINE).

predicate alloc_netdev_other(FN:string,C:string).
alloc_netdev_other(FN,_), +handled_netdev(FN).

% functions which wrap alloc_netdev or another wrapper, storing regular
% private data in a strange way, possibly including extended storage
+alloc_netdev_other("__alloc_ei_netdev","ei_device").
+alloc_netdev_other("drivers/net/pcmcia/axnet_cs.c:axnet_probe","ei_device").
+alloc_netdev_other("alloc_orinocodev","orinoco_private").
+alloc_netdev_other("alloc_ieee80211","ieee80211_device").
+alloc_netdev_other("drivers/net/chelsio/cxgb2.c:init_one","adapter").
+alloc_netdev_other("drivers/net/wireless/hostap/hostap_hw.c:prism2_init_local_data","hostap_interface").
+alloc_netdev_other("net/atm/lec.c:lecd_attach","lec_priv").
+alloc_netdev_other("arlan_probe","arlan_private").

cil_curfn(XFN), alloc_netdev_other(XFN,COMP),
    dircall(I,"alloc_netdev"), icall(P,_,I),
    cil_instr_call_arg(I,2,ESETUP), eval(P,ESETUP,s_trace{root{glob{SETUPFN}}},_),
    +add_netdev_setup(SETUPFN,COMP), +add_netdev_return(P,I,COMP).
cil_curfn(XFN), alloc_netdev_other(XFN,COMP),
    dircall(I,FN), icall(P,_,I), alloc_netdev_wrapper(FN,_),
    +add_netdev_return(P,I,COMP).

% hdlcdrv_register allocates a structural subtype of hdlcdrv_state
+alloc_netdev_wrapper("hdlcdrv_register",1).
cil_curfn("hdlcdrv_register"),
    +add_netdev_setup("drivers/net/hamradio/hdlcdrv.c:hdlcdrv_setup","hdlcdrv_state").

% alloc_ieee80211 pads the private data with yet another structure
dircall(I,"alloc_ieee80211"), icall(P,_,I),
    cil_instr_call_arg(I,0,ESIZE), sizeof_exp(ESIZE,COMP),
    callret(P,I,RT,_), cursum(FN,SUM), netdev_priv(uc_sum{s_call{I},P,RT},PRIVT),
    +add_explicit_fixed(FN,SUM,fld{PRIVT,"priv","ieee80211_device"},COMP).

dircall(I,"alloc_hdlcdev"), icall(P,_,I), cursum(FN,SUM),
    callret(P,I,RT,_), inst_trace(s_call{I},P,drf{root{arg{0}}},s_trace{T},_),
    netdev_priv(uc_sum{s_call{I},P,RT},PRIVT), explicit_type(T,COMP),
    +add_explicit_fixed(FN,SUM,drf{fld{PRIVT,"priv","hdlc_device_struct"}},COMP).

% linux mmc_hosts

dircall(I,"mmc_alloc_host"), icall(P,_,I),
    cil_instr_call_arg(I,0,ESIZE), sizeof_exp(ESIZE,COMP),
    callret(P,I,RT,_), cursum(FN,SUM),
    +add_explicit_fixed(FN,SUM,index{fld{uc_sum{s_call{I},P,RT},"private","mmc_host"},y_int{iulong},s_const{0}},COMP).

% linux fb_info

dircall(I,"framebuffer_alloc"), icall(P,_,I),
    cil_instr_call_arg(I,0,ESIZE), sizeof_exp(ESIZE,COMP),
    callret(P,I,RT,_), cursum(FN,SUM),
    +add_explicit_fixed(FN,SUM,drf{fld{uc_sum{s_call{I},P,RT},"par","fb_info"}},COMP).

% linux snd_card

dircall(I,"snd_card_new"), icall(P,_,_I),
    cil_instr_call_arg(I,3,ESIZE), sizeof_exp(ESIZE,COMP),
    callret(P,I,RT,_), cursum(FN,SUM),
    +add_explicit_fixed(FN,SUM,drf{fld{uc_sum{s_call{I},P,RT},"private_data","snd_card"}},COMP).

% linux Scsi_Host

dircall(I,"scsi_host_alloc"), icall(P,_,I),
    cil_instr_call_arg(I,1,ESIZE), sizeof_exp(ESIZE,COMP),
    callret(P,I,RT,_), cursum(FN,SUM),
    +add_explicit_fixed(FN,SUM,fld{uc_sum{s_call{I},P,RT},"hostdata","Scsi_Host"},COMP).

% linux caches

dircall(I,"kmem_cache_create"), icall(P,_,I),
    cil_instr_call_arg(I,1,ESIZE), sizeof_exp(ESIZE,COMP),
    inst_trace(s_call{I},P,drf{root{arg{4}}},s_trace{root{glob{CTOR}}},_),
    +add_explicit_fixed(CTOR,s_func,drf{root{arg{0}}},COMP),
    inst_trace(s_call{I},P,drf{root{arg{5}}},s_trace{root{glob{DTOR}}},_),
    +add_explicit_fixed(DTOR,s_func,drf{root{arg{0}}},COMP).

% file

% file->private_data is zero at entry to the open methods.
% we can't prove this because some of the opens are mutually recursive,
% and there's still a lot of code from __dentry_open back to file creation
omit_zero_initialized_fail(FN,s_func,_,fld{drf{root{arg{1}}},"private_data","file"}) :-
    cil_sum_caller(FN,s_func)->caller("fs/open.c:__dentry_open",s_func,_,s_call{PI}),
    spoly_call("fs/open.c:__dentry_open",PI)->wpoly_call_target(_).
