% main sessions for use in deriving facts about a program

% locations we can prove facts for along with the type of the fact itself.
% every particular fact is an (h_kind_loc,h_kind) pair
type h_kind_loc.
type h_kind.

% methods for proving an h_kind holds for a given h_kind_loc.
% each method is separate and if any one succeeds then the h_kind is proved
type h_method.

% individual check to perform to see whether some HM can prove some HK.
% can be many corresponding to each HM, each of which must be proved
type h_check.

% extra data that can be associated with each HM, but does not influence
% how that HM is actually proved
type h_method_extra.

% hashed string representations of h_kind and h_check, for places where we do
% not need to dissect the actual representation. hashes are reasonably unique
type h_kind_hash = int.
type h_check_hash = int.

% those facts we are attempting to prove
session scasting_try(HKL:h_kind_loc) containing [scheck].

% those facts which we attempted to prove and generated dependencies for
session scasting_tried(HKL:h_kind_loc) containing [scheckh].

% those facts we have proved
session scasting_proved(HKL:h_kind_loc) containing [scheckh].

predicate scheck(HK:h_kind).
predicate scheckh(HK:h_kind_hash).

% information about derivations which can be made if facts HK/HKL are proved
session scasting_derive(HKL:h_kind_loc)
    containing [sprint,sfirststep,snextstep].

% string to print if HK is successfully proved
predicate sprint(HK:h_kind_hash,STR:string).

predicate sfirststep(HK:h_kind_hash,
                     XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method).
predicate snextstep(HK:h_kind_hash,
                    XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method,
                    NHKL:h_kind_loc,NHK:h_kind_hash).

% information for doing post-mortem analysis of derivation state
session scasting_post(HKL:h_kind_loc)
    containing [sloc_info,scheck_name,scheck_info,
                scheck_fail,scheck_require,scheck_extra,sprint_fail].

% overall info on the location currently being described
predicate sloc_info(SNAME:string,STITLE:string,
                    FILE:string,MINLINE:int,MAXLINE:int).

% name of each check we want to generate UI data for
predicate scheck_name(HK:h_kind_hash,NAME:string,LINE:int,TEXT:string).

% individual methods and checks which apply for each scheck_name
predicate scheck_info(HK:h_kind_hash,HM:h_method,
                      HC:h_check_hash,LINE:int,TEXT:string).

% individual checks which failed outright or require another check to hold
predicate scheck_fail(HC:h_check_hash).
predicate scheck_require(HC:h_check_hash,
                         XHKL:h_kind_loc,XHK:h_kind_hash,LINK:string).

% extra information associated with a particular method
predicate scheck_extra(HK:h_kind_hash,HM:h_method,HX:h_method_extra).

% string to print if HK was never proved
predicate sprint_fail(HK:h_kind_hash,STR:string).

using strops.
using intops.

% PREDICATES

% get the hash representation for an h_kind or h_check
predicate hash_kind(in h_kind,h_kind_hash).
predicate hash_check(in h_check,h_check_hash).
hash_kind(HK,HASH) :- tostring(HK,HKS), str_hash(HKS,HASH).
hash_check(HC,HASH) :- tostring(HC,HCS), str_hash(HCS,HASH).

% names of displays that can be generated for the UI, per h_kind_loc.
type t_casting_display ::=
    td_casting_overall
  | td_casting_kind{h_kind_hash}
  | td_casting_method{h_kind_hash,h_method}.

predicate casting_display_name(in HKL:h_kind_loc,in D:t_casting_display,
                               DNAME:string) succeeds [once].
casting_display_name(HKL,td_casting_overall,STR) :-
    tostring(HKL,HKLS), STR = "casting_overall:"^HKLS.
casting_display_name(HKL,td_casting_kind{HKH},STR) :-
    tostring(HKL,HKLS), tostring(HKH,HS),
    STR = "casting_kind:"^HKLS^":"^HS.
casting_display_name(HKL,td_casting_method{HKH,HM},STR) :-
    tostring(HKL,HKLS), tostring(HKH,HS), tostring(HM,HMS),
    STR = "casting_method:"^HKLS^":"^HS^":"^HMS.
