import "../daliasing/memory.clp".

using display.

% SESSIONS

type t_loc ::=
    yl_func_arg{FN:string,A:int}
  | yl_func_ret{FN:string}
  | yl_glob{G:string}
  | yl_comp{C:string,F:string}
  | yl_target{CFN:string}
      % intermediate non-summary locations:
  | yl_call{I:c_instr}
  | yl_assign{T:t_trace}.

type t_xloc ::=
    yx_func{FN:string}
  | yx_glob{G:string}.
predicate curloc(t_xloc) succeeds [once].
predicate curloc_info(SNAME:string,FILE:string,
                      MINLINE:int,MAXLINE:int,LINE:int) succeeds [once].

type t_ref ::=
    yr_call{FN:string,I:c_instr,IS:string}.

session sum_funptr_prop(Y:t_loc) containing [sref].
predicate sref(t_ref).

predicate funptr_assign(T:t_trace,V:t_trace).
session sum_init_assign(GLOB:string) containing [funptr_assign].
session sum_init_assign_field(COMP:string,FLD:string) containing [funptr_assign].

% CUSTOMIZATION

% T is handled by an annotation and should not be marked as failed
predicate handle(in T:t_trace).

% references on T should be propagated to XY
predicate handle_ref(in T:t_trace,XY:t_loc).
handle(T) :- handle_ref(T,_).

% the type of T is ok and propagation should proceed as normal
predicate handle_type(in T:t_trace).

import "funptr_baked.clp".

% PREDICATES

predicate tref(T:t_trace,I:t_ref,P:pp,Y:t_loc).
predicate tfail(T:t_trace).

predicate str_indirect(FN:string,I:string,CFN:string).
?- str_indirect(FN,I,CFN).

predicate str_failed(YX:t_xloc,T:string).
curloc(YX), tfail(T), ~handle(T), trace_string(T,TS), +str_failed(YX,TS).
?- str_failed(YX,T).

predicate str_handled(YX:t_xloc,T:string).
curloc(YX), tfail(T), handle(T), trace_string(T,TS), +str_handled(YX,TS).
?- str_handled(YX,TS).

predicate str_ref(T:string,Y:t_loc).
%tref(T,_,_,Y), trace_string(T,TS), +str_ref(TS,Y).
%?- str_ref(T,Y).

% PROPAGATION RULES

predicate xeval(in E:c_exp,T:t_trace).
predicate xassign(P:pp,LT:t_trace,RT:t_trace).
predicate xargument(P:pp,I:c_instr,A:int,T:t_trace).

% remove any array index{} subtraces from the input trace
predicate trace_strip_index(in T:t_trace,NT:t_trace) succeeds [once].

% we don't need to handle uc_val/uc_sum traces because disable_val() is on and they
% won't be generated
+trace_strip_index(root{R},root{R}).
+trace_strip_index(uc_exp{P,E,Y},uc_exp{P,E,Y}).
+trace_strip_index(uc_val{P,T},uc_val{P,T}).

trace_strip_index(drf{T},drf{NT}) :- trace_strip_index(T,NT).
trace_strip_index(fld{T,F,C},fld{NT,F,C}) :- trace_strip_index(T,NT).
trace_strip_index(rfld{T,F,C},rfld{NT,F,C}) :- trace_strip_index(T,NT).
trace_strip_index(index{T,_,_},NT) :- trace_strip_index(T,NT).

xeval(E,T) :- exp_point(E,P), eval(P,E,s_trace{BT},_), trace_strip_index(BT,T).
fill_assign(P,BLT,s_trace{BRT},_,_), trace_strip_index(BLT,LT), trace_strip_index(BRT,RT), +xassign(P,LT,RT).
fill_argument(P,I,root{arg{A}},s_trace{BT},_,_), trace_strip_index(BT,T), +xargument(P,I,A,T).

icall(P,_,I), cil_instr_call(I,FNE), ~dircall(I,_), xeval(FNE,drf{T}),
    cil_curfn(FN), exp_string(FNE,IS),
    +tref(T,yr_call{FN,I,IS},P,yl_call{I}).

predicate any_access(P:pp,T:t_trace).
xassign(P,LT,RT), +any_access(P,LT), +any_access(P,RT).
xargument(P,_,_,T), +any_access(P,T).

any_access(P,T), trace_root(T,glob{G}),
    sum_funptr_prop(yl_glob{G})->sref(I),
    +tref(drf{root{glob{G}}},I,P,yl_glob{G}).
any_access(P,T), trace_sub(T,fld{STR,F,C},_),
    sum_funptr_prop(yl_comp{C,F})->sref(I),
    +tref(drf{fld{STR,F,C}},I,P,yl_comp{C,F}).
icall(P,_,XI), anycall(XI,CFN,_), xargument(_,XI,A,T),
    sum_funptr_prop(yl_func_arg{CFN,A})->sref(I),
    +tref(T,I,P,yl_func_arg{CFN,A}).
cil_curfn(FN), fill_assign(P,root{return},_,_,_),
    sum_funptr_prop(yl_func_ret{FN})->sref(I),
    +tref(drf{root{return}},I,P,yl_func_ret{FN}).

% cross-assignment propagation
tref(drf{PT},I,_,_), xassign(P,PT,DT), +tref(DT,I,P,yl_assign{PT}).

% unrecognized forms of propagation
tref(drf{drf{PPT}},_,_,_), +tfail(drf{drf{PPT}}).
tref(drf{PT},_,_,_), xassign(_,_,PT), +tfail(drf{PT}).
tref(drf{PT},_,_,_), xargument(_,_,_,PT), +tfail(drf{PT}).
tref(T,_,_,_), (cil_curfn(FN); ~cil_curfn(_), FN=""),
    ~trace_type(FN,T,y_func{_,_,_}),
    ~handle_type(T), +tfail(T).
tref(T,_,_,_), T=drf{root{R}},
    icall(P,_,I), callret(P,I,root{R},_), dircall(I,CFN),
    ~trace_type(CFN,drf{root{return}},y_func{_,_,_}),
    ~handle_type(T), +tfail(T).

predicate add_ref(Y:t_loc,I:t_ref,T:t_trace).
add_ref(Y,I,T), ~tfail(T), +sum_funptr_prop(Y)->sref(I).

tref(T,I,_,Y), T=drf{root{glob{G}}}, Y\=yl_glob{G}, +add_ref(yl_glob{G},I,T).
tref(T,I,_,Y), T=drf{fld{_,F,C}}, Y\=yl_comp{C,F}, +add_ref(yl_comp{C,F},I,T).
tref(T,I,_,_), T=drf{root{arg{A}}}, cil_curfn(FN),
    +add_ref(yl_func_arg{FN,A},I,T).
tref(T,I,_,_), T=drf{root{R}},
    icall(P,_,XI), callret(P,XI,root{R},_), anycall(XI,CFN,_),
    +add_ref(yl_func_ret{CFN},I,T).
tref(T,I,_,_), tfail(T), handle_ref(T,XY),
    +add_ref(XY,I,T), +sum_funptr_prop(XY)->sref(I).
tref(T,I,_,_), T=root{glob{CFN}}, +add_ref(yl_target{CFN},I,T).    

add_ref(yl_target{CFN},yr_call{FN,XI,XIS},_),
    +str_indirect(FN,XIS,CFN), +sum_funptr(FN)->sindirect(XI,CFN).

predicate intermediate_ref(in Y:t_loc,NT:maybe[t_trace]).
+intermediate_ref(yl_call{_},no).
+intermediate_ref(yl_assign{PT},yes{drf{PT}}).

% UI INTEGRATION

predicate ref_display(in Y:t_loc,in I:t_ref,DNAME:string).
ref_display(Y,I,DNAME) :-
    tostring(Y,YS), tostring(I,IS),
    DNAME = "funptr_ref:"^YS^":"^IS.

predicate ref_item(in T:t_trace,in I:t_ref,ITEM:displayitem).
ref_item(T,I,ITEM) :-
    tref(T,I,P,Y), point_location(P,_,LINE),
    trace_string(T,TS), tostring(Y,YS),
    TEXT = TS^" holds reference from "^YS,
    (intermediate_ref(Y,_), LINKS=[];
     ~intermediate_ref(Y,_), ref_display(Y,I,DNAME),
         LINKS=[d_link{DNAME,"link"}]),
    ITEM = d_line_text{LINE,"funptr_ref",d_bottom,TEXT,LINKS}.
ref_item(T,I,ITEM) :-
    tref(T,I,_,Y), intermediate_ref(Y,yes{XT}), ref_item(XT,I,ITEM).
ref_item(T,I,ITEM) :-
    handle_ref(T,Y), curloc_info(_,_,_,_,LINE),
    trace_string(T,TS), tostring(Y,YS),
    TEXT = TS^" has annotated propagation to "^YS,
    ITEM = d_line_text{LINE,"funptr_annot",d_bottom,TEXT,[]}.

predicate total_item(ITEM:displayitem).
curloc_info(SNAME,FILE,MINLINE,MAXLINE,NLINE),
    add_ref(Y,I,T), ref_display(Y,I,DNAME),
    \/ref_item(T,I,ITEM):list_all(ITEM,ITEMS),
    tostring(Y,YS), I=yr_call{FN,_,IS},
    TITLE = YS^" may flow to "^IS^" within "^FN,
    +display_add(DNAME,false,"funptr_ref",TITLE,SNAME,
                 FILE,MINLINE,MAXLINE,NLINE,ITEMS),
    +total_item(d_line_text{NLINE,"funptr_any",d_bottom,
                            TITLE,[d_link{DNAME,"link"}]}).

curloc(YX), curloc_info(SNAME,FILE,MINLINE,MAXLINE,NLINE),
    \/total_item(ITEM):list_all(ITEM,ITEMS),
    tostring(YX,YXS), DNAME = "funptr_any:"^YXS,
    TITLE = "All references within "^SNAME,
    +display_add(DNAME,true,"funptr_any",TITLE,SNAME,
                 FILE,MINLINE,MAXLINE,NLINE,ITEMS),
    +search_add(SNAME,DNAME).
