
import "memory.clp".
import "../daliasing/aliasing_sum.clp".
import "../daliasing/escape.clp".
import "../module/callgraph_sum.clp".

% omit regular trace_alias/trace_clobber for specified traces
predicate omit_trace_alias(in FN:string,in SUM:sum,
                           in T0:t_trace,in T1:t_trace).
predicate omit_trace_clobber(in I:sum,in P:pp,in T:t_trace).

% pessimistic mod set information for call/loop summaries

trace_clobber(I,P,XT) :- ~omit_trace_clobber(I,P,XT),
    isum_target_p(P,I,CFN,CSUM,direct),
    salias_write_func(CFN,CSUM)->wfunc_write(CT),
    inst_trace(I,P,CT,s_trace{T},_),
    break_point(P,FN,SUM,_), trace_alias(FN,SUM,T,XT,_).
trace_clobber(s_call{I},P,T) :- callret(P,I,T,_).

% turn inline writes in CFN into direct writes in this function.
% (these assign edges will be captured and persisted for remote queries)
% TODO: fix stopgap for broken write_inline due to not modelling
% copy_to/from_user assembly
predicate added_write_inline(I:c_instr,CT:t_trace).
icall(P,_,I), dircall(I,CFN),
    CFN\="copy_from_user", CFN\="copy_to_user",
    salias_write_func(CFN,s_func)->wfunc_write_inline(CT,CV),
    +added_write_inline(I,CT),
    inst_trace(s_call{I},P,CT,s_trace{T},TG),
    inst_scalar(s_call{I},P,CV,V,VG), #and(TG,VG,G), +assign(P,T,V,G).

% add an assign for the function's return value assignment,
% even if there is no write_inline in the callee
icall(P,_,I), callret(P,I,T,G),
    ~added_write_inline(I,root{return}),
    base_val(P,T,V,_), % this will be uc_sum so guard is true
    +assign(P,T,V,G).

% functions called via I, following any loops
predicate loop_callee(in P:pp,in I:sum,CFN:string).
loop_callee(P,s_call{I},CFN) :- isum_target_p(P,s_call{I},CFN,_,_).
loop_callee(P,s_loop{PH},CFN) :- isum_target_p(P,s_loop{PH},FN,SUM,_),
    cil_sum_body(FN,SUM)->isum(XP,_,XI),
    loop_callee(p_remote{FN,SUM,XP},XI,CFN).

predicate address_escaped(in I:sum,in P:pp,in FN:string,in T:t_trace).
address_escaped(_,_,_,T) :- trace_sub(T,uc_sum{_,_,_},_).
address_escaped(I,P,FN,T) :-
    reduce_trace_skolem(T,NT),
    ~escape_set(10,es_forward_precise{ep_nocontext},tl_func{FN,NT,cxt_any},_).
address_escaped(I,P,FN,T) :-
    reduce_trace_skolem(T,NT),
    escape_set(10,es_forward_precise{ep_nocontext},tl_func{FN,NT,cxt_any},SET),
    set_member(SET,L),
    (L\=tl_func{_,_,_};
     L=tl_func{CFN,AT,_}, loop_callee(P,I,CFN), trace_root(AT,arg{_});
     I=s_loop{_}, L=tl_func{FN,XT,_}, XT\=T).

% T is something other than the address of a local variable
% declared by a parent function
predicate parent_address_escaped(in I:sum,in P:pp,in FN:string,in T:t_trace).
parent_address_escaped(_,_,FN,T) :-
    ~escape_set(10,es_backward,tl_func{FN,T,cxt_any},_).
parent_address_escaped(_,_,FN,T) :-
    escape_set(10,es_backward,tl_func{FN,T,cxt_any},LSET),
    set_member(LSET,L), L\=tl_func{_,_,_},
    L\=tl_func{_,drf{root{arg{_}}},_},
    L\=tl_func{_,root{local{_}},_}.
parent_address_escaped(I,P,FN,T) :-
    escape_set(10,es_backward,tl_func{FN,T,cxt_any},LSET),
    set_member(LSET,tl_func{XFN,XT,_}),
    trace_alloc(s_func,XT,ak_stack_local),
    address_escaped(I,P,XFN,XT).

predicate local_data(in I:sum,in P:pp,in T:t_trace).
local_data(I,P,T) :- trace_alloc(s_func,T,ak_stack_local),
    break_point(P,FN,_,_), ~address_escaped(I,P,FN,T).
local_data(I,P,T) :- T=drf{root{arg{_}}},
    break_point(P,FN,_,_), ~parent_address_escaped(I,P,FN,T).
local_data(I,P,T) :- trace_root(T,glob{G}), trace_derefs(T,0),
    break_point(P,FN,_,_), cil_body(FN)->cil_var_name(_,G,_),
    ~address_escaped(I,P,FN,T).
local_data(s_call{I},P,T) :-
    trace_sub(T,uc_sum{s_call{I},P,BT},_), callret(P,I,BT,_),
    icall(P,_,I), allocate_call(I,P,_,root{return},_,_).
local_data(_,P,T) :- break_point(P,FN,_,_), trace_type(FN,T,y_func{_,_,_}).

predicate isum_callee(in FN:string,in SUM:sum,in I:sum,CFN:string).
isum_callee(FN,SUM,s_call{I},CFN) :-
    cil_sum_callee(FN,SUM)->callee(_,s_call{I},CFN,s_func).
isum_callee(FN,SUM,s_loop{PH},CFN) :-
    cil_sum_callee(FN,SUM)->callee(_,s_loop{PH},FN,CSUM), CSUM\=s_loop{PH},
    isum_callee(FN,s_loop{PH},CSUM,CFN).

% ignore impure writes in FN for C/F
predicate omit_comp_impure(in C:string,in F:string,in FN:string).

trace_clobber(I,P,T) :- ~omit_trace_clobber(I,P,T),
    ~local_data(I,P,T), T=fld{_,F,C},
    break_point(P,FN,SUM,_), isum_callee(FN,SUM,I,CFN),
    salias_write_comp(C,F)->wcomp_impure(XFN), ~omit_comp_impure(C,F,XFN),
    transitive_callee(CFN,XFN).
trace_clobber(I,P,T) :- ~omit_trace_clobber(I,P,T),
    ~local_data(I,P,T), T=fld{_,F,C},
    break_point(P,FN,SUM,_), isum_callee(FN,SUM,I,CFN),
    salias_write_comp(C,F)->wcomp_init(CFN,_).
trace_clobber(I,P,T) :- ~omit_trace_clobber(I,P,T),
    ~local_data(I,P,T), T\=fld{_,_,_},
    isum_target_p(P,I,CFN,CSUM,direct),
    salias_write_func(CFN,CSUM)->wfunc_impure().

% pessimistic aliasing information. this is the same as aliasing_mid.clp,
% except that we also consider the possibility of extra possible aliasing:
% *x may alias y->f
% *x may alias z
% these aliases may only happen if the two sides are the same type and
% using escape() to follow &y->f/&z forward or x backward, both might
% reach the other

% old rules from aliasing_mid.clp
trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    T0=fld{STR0,F,C}, T1=fld{STR1,F,C},
    trace_alias(FN,SUM,STR0,STR1,G).
trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    T0\=fld{_,_,_}, T1\=fld{_,_,_},
    (T0\=root{_}; T0=root{glob{_}}),
    (T1\=root{_}; T1=root{glob{_}}),
    trace_type(FN,T0,Y), trace_type(FN,T1,Y),
    #id_g(sc_bit{sb_alias{T0,T1}},G).

predicate escape_non_alias(in FN:string,in AT:t_trace,in XT:t_trace).
escape_non_alias(FN,AT,XT) :-
    reduce_trace_skolem(AT,NAT), reduce_trace_skolem(XT,NXT),
    reduce_trace_loc(tl_func{FN,NAT,cxt_any},AL),
    reduce_trace_loc(tl_func{FN,NXT,cxt_any},XL),
    (escape_set(10,es_forward,AL,ALSET), ~set_member(ALSET,XL);
     escape_set(10,es_backward,XL,XLSET), ~set_member(XLSET,AL)).

trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    (T0=fld{_,_,_}; T0=root{_}), T1\=fld{_,_,_}, T1\=root{_},
    trace_type(FN,T0,Y), trace_type(FN,T1,Y),
    ~escape_non_alias(FN,T0,T1),
    #id_g(sc_bit{sb_alias{T0,T1}},G).
trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    T0\=fld{_,_,_}, T0\=root{_}, (T1=fld{_,_,_}; T1=root{_}),
    trace_type(FN,T0,Y), trace_type(FN,T1,Y),
    ~escape_non_alias(FN,T1,T0),
    #id_g(sc_bit{sb_alias{T0,T1}},G).

omit_trace_alias(_,_,T,_) :- trace_root(T,local{_}), ~trace_derefs(T,0).
omit_trace_alias(_,_,_,T) :- trace_root(T,local{_}), ~trace_derefs(T,0).

% BAKED

% fields we're pretty sure are pure but can't prove it yet

% probe-related fields
+omit_comp_impure("pci_dev","driver",_).  % is driver probe really reentrant???
+omit_comp_impure("device","driver",_).
+omit_comp_impure("cdrom_device_info","ops",_).
+omit_comp_impure("i2c_client","driver",_).
omit_comp_impure("net_device",F,_) :-
    comp_field("net_device",_,_,F,y_ptr{y_func{_,_,_}}).
omit_comp_impure("input_dev",F,_) :-
    comp_field("input_dev",_,_,F,y_ptr{y_func{_,_,_}}).
+omit_comp_impure("snd_pcm_ops",_,_).
+omit_comp_impure("snd_pcm_substream",_,_).  % runtime, ops
+omit_comp_impure("snd_info_entry_text",_,_).
+omit_comp_impure("snd_timer_hardware",_,_).
+omit_comp_impure("snd_ac97_bus","private_free",_).
+omit_comp_impure("fb_info","fbops",_), +omit_comp_impure("fb_info","class_device",_).
+omit_comp_impure("dvb_frontend","dvb",_).
+omit_comp_impure("irq_req_t","Handler",_).
+omit_comp_impure("input_dev","event",_).
+omit_comp_impure("tty_ldisc",_,_).

% filesystem-related fields
+omit_comp_impure("file","f_op",_).
+omit_comp_impure("inode","i_fop",_).
+omit_comp_impure("file_operations","write",_).  % see acpi_video_device_add_fs and several others

% other fields
+omit_comp_impure("dvb_demux_feed","demux",_).
+omit_comp_impure("proc_dir_entry","read_proc",_).
+omit_comp_impure("aoedev","gd",_).  % initialization deferred to another thread
+omit_comp_impure("hotplug_slot","ops",_).
+omit_comp_impure("ac97_codec","codec_write",_), +omit_comp_impure("ac97_codec","codec_read",_).
+omit_comp_impure("snd_hwdep_ops",_,_).

% side effects we want but don't get from salias_write_func

% get the value of T after I/P
predicate post_sum_val(in I:sum,in P:pp,in T:t_trace,V:scalar,G:g_guard).
post_sum_val(I,P,T,V,G) :- trace_clobber(I,P,T), V=s_trace{uc_sum{I,P,T}}, #bool_g(true,G).
post_sum_val(I,P,T,V,G) :- ~trace_clobber(I,P,T), val(P,T,V,G).

% alloc_hdlcdev sets its retval.$net_device$.priv to its first argument
dircall(I,"alloc_hdlcdev"), icall(P,_,I),
    argument(P,I,root{arg{0}},V,G),
    callret(P,I,RT,_), netdev_priv(uc_sum{s_call{I},P,RT},PRIVT),
    +assign(P,fld{PRIVT,"priv","hdlc_device"},V,G).

% seq_open casts file->private_data to a seq_file and updates op to the input argument
dircall(I,"seq_open"), icall(P,_,I),
    inst_trace(s_call{I},P,drf{root{arg{1}}},OPS,G0),
    inst_trace(s_call{I},P,fld{drf{root{arg{0}}},"private_data","file"},s_trace{PRIV},G1),
    post_sum_val(s_call{I},P,PRIV,s_trace{DPRIV},G2), OPFLD=fld{DPRIV,"op","seq_file"},
    guard(P,G3), #and4(G0,G1,G2,G3,G), +assign(P,OPFLD,OPS,G).
