import "memory.clp".
import "aliasing_full.clp".
import "scalar_sat.clp".

analyze session_name("cil_sum_body").

predicate str_comp_pure(C:string,F:string,PTR:bool).
predicate str_comp_impure(C:string,F:string,PTR:bool).
predicate str_comp_access_pure(FN:string,SUM:sum,T:string,C:string,F:string,PTR:bool).
predicate str_comp_access_impure(FN:string,SUM:sum,T:string,C:string,F:string,PTR:bool).
?- str_comp_pure(C,F,PTR).
?- str_comp_impure(C,F,PTR).
?- str_comp_access_pure(FN,SUM,T,C,F,PTR).
?- str_comp_access_impure(FN,SUM,T,C,F,PTR).

access(_,T,_), T=fld{STR,F,C}, trace_string(STR,STRS),
    cursum(FN,SUM), trace_type(FN,T,Y),
    (Y=y_ptr{_}, PTR=true; Y\=y_ptr{_}, PTR=false),
    ~salias_write_comp(C,F)->wcomp_impure(_),
    +str_comp_pure(C,F,PTR),
    +str_comp_access_pure(FN,SUM,STRS,C,F,PTR).

access(_,T,_), T=fld{STR,F,C}, trace_string(STR,STRS),
    cursum(FN,SUM), trace_type(FN,T,Y),
    (Y=y_ptr{_}, PTR=true; Y\=y_ptr{_}, PTR=false),
    salias_write_comp(C,F)->wcomp_impure(_),
    +str_comp_impure(C,F,PTR),
    +str_comp_access_impure(FN,SUM,STRS,C,F,PTR).

predicate str_func_pure(FN:string,ZERO:bool).
predicate str_func_impure(FN:string).
predicate str_func_call_pure(FN:string,SUM:sum,LINE:int,I:c_instr,ZERO:bool).
predicate str_func_call_impure(FN:string,SUM:sum,LINE:int,I:c_instr).
?- str_func_pure(FN,ZERO).
?- str_func_impure(FN).
?- str_func_call_pure(FN,SUM,LINE,I,ZERO).
?- str_func_call_impure(FN,SUM,LINE,I).

predicate non_return_write(in FN:string).
non_return_write(FN) :-
    salias_write_func(FN,s_func)->wfunc_write(T), T\=root{return}.

cursum(FN,s_func), ~salias_write_func(FN,s_func)->wfunc_impure(),
    (non_return_write(FN), ZERO=false;
     ~non_return_write(FN), ZERO=true),
    +str_func_pure(FN,ZERO).
cursum(FN,s_func), salias_write_func(FN,s_func)->wfunc_impure(),
    +str_func_impure(FN).

predicate legitimate_call(in I:c_instr).
legitimate_call(I) :- anycall(I,CFN,_),
    cil_body(CFN)->cil_fundec_location(_,_,_).

predicate non_return_write_call(in I:c_instr).
non_return_write_call(I) :- anycall(I,CFN,_), non_return_write(CFN).

cursum(FN,SUM), icall(P,_,I), legitimate_call(I), point_location(P,_,LINE),
    ~(anycall(I,CFN,_), salias_write_func(CFN,s_func)->wfunc_impure()),
    (non_return_write_call(I), ZERO=false;
     ~non_return_write_call(I), ZERO=true),
    +str_func_call_pure(FN,SUM,LINE,I,ZERO).
cursum(FN,SUM), icall(P,_,I), legitimate_call(I), point_location(P,_,LINE),
    anycall(I,CFN,_), salias_write_func(CFN,s_func)->wfunc_impure(),
    +str_func_call_impure(FN,SUM,LINE,I).

predicate str_trace_alias_any(FN:string,SUM:sum,T0:string,T1:string).
predicate str_trace_alias_array(FN:string,SUM:sum,T0:string,T1:string,
                                SAME:bool,Y:t_type).
predicate str_trace_alias_field(FN:string,SUM:sum,T0:string,T1:string,
                                SAME:bool,C:string,F:string,Y:t_type).
?- str_trace_alias_any(FN,SUM,T0,T1).
?- str_trace_alias_array(FN,SUM,T0,T1,SAME,Y).
?- str_trace_alias_field(FN,SUM,T0,T1,SAME,C,F,Y).

% check any found alias that is not between composite types.
% composite type alias queries are only due to recursion within trace_alias
predicate chk_trace_alias(FN:string,SUM:sum,T0:t_trace,T1:t_trace).
cursum(FN,SUM), ?trace_alias(FN,SUM,T0,T1,_), T0\=T1,
    ~trace_type(FN,T0,y_comp{_}), ~trace_type(FN,T1,y_comp{_}),
    trace_alias(FN,SUM,T0,T1,_),
    (lessthan(T0,T1), XT0=T0, XT1=T1;
     lessthan(T1,T0), XT0=T1, XT1=T0),
    +chk_trace_alias(FN,SUM,XT0,XT1).

chk_trace_alias(FN,SUM,T0,T1), trace_string(T0,T0S), trace_string(T1,T1S),
    +str_trace_alias_any(FN,SUM,T0S,T1S).
chk_trace_alias(FN,SUM,T0,T1),
    trace_sub(T0,index{BT0,Y,_},RT), trace_sub(T1,index{BT1,Y,_},RT),
    trace_string(T0,T0S), trace_string(T1,T1S),
    (BT0=BT1, SAME=true; BT0\=BT1, SAME=false),
    +str_trace_alias_array(FN,SUM,T0S,T1S,SAME,Y).
chk_trace_alias(FN,SUM,T0,T1), (T0=fld{_,F,C}; T1=fld{_,F,C}),
    trace_string(T0,T0S), trace_string(T1,T1S),
    (T0=fld{_,F,C}, T1=fld{_,F,C}, SAME=true;
     T0\=fld{_,F,C}, SAME=false;
     T1\=fld{_,F,C}, SAME=false),
    comp_field(C,_,_,F,Y),
    +str_trace_alias_field(FN,SUM,T0S,T1S,SAME,C,F,Y).

predicate type_has_pointer(in C:string).
type_has_pointer(C) :- comp_field(C,_,_,_,y_ptr{_}).
type_has_pointer(C) :- comp_field(C,_,_,_,y_comp{XC}), type_has_pointer(XC).

% argument CT to I/P might flow to a point where a field of it is accessed
predicate may_use_field(in I:c_instr,in P:pp,in CT:t_trace,PTR:bool,T:t_trace).
may_use_field(I,P,CT,PTR,T) :- inst_trace(s_call{I},P,CT,s_trace{T},_),
    reduce_trace_skolem(T,MT),
    (MT=uk_index{NT,_}; MT\=uk_index{_,_}, NT=MT),
    cil_curfn(FN), reduce_trace_loc(tl_func{FN,NT,cxt_any},L),
    (~escape_set(1000,es_forward,L,_), PTR=true;
     escape_set(1000,es_forward,L,NLSET), set_member(NLSET,NL),
         salias_used(NL)->sfld(_,_,XC,_),
         (type_has_pointer(XC), PTR=true;
          ~type_has_pointer(XC), PTR=false)).

predicate memcpy_call(FN:string,SUM:sum,LINE:int).
predicate memcpy_miss(FN:string,SUM:sum,LINE:int).
predicate memcpy_fail(FN:string,SUM:sum,LINE:int,PTR:bool).
predicate memcpy_trace_fail(FN:string,SUM:sum,LINE:int,PTR:bool,T:t_trace).
?- memcpy_call(FN,SUM,LINE).
?- memcpy_miss(FN,SUM,LINE).
?- memcpy_fail(FN,SUM,LINE,PTR).
?- memcpy_trace_fail(FN,SUM,LINE,PTR,T).

cursum(FN,SUM), dircall(I,CFN), memcpy_func(CFN,AT,_,_),
    icall(P,_,I), point_location(P,_,LINE),
    +memcpy_call(FN,SUM,LINE),
    ~memcpy_handled(I), +memcpy_miss(FN,SUM,LINE),
    may_use_field(I,P,drf{AT},PTR,T),
    ~(PTR=false, may_use_field(I,P,drf{AT},true,_)),
    +memcpy_fail(FN,SUM,LINE,PTR),
    +memcpy_trace_fail(FN,SUM,LINE,PTR,T).

predicate memset_call(FN:string,SUM:sum,LINE:int).
predicate memset_miss(FN:string,SUM:sum,LINE:int).
predicate memset_fail(FN:string,SUM:sum,LINE:int,PTR:bool).
predicate memset_trace_fail(FN:string,SUM:sum,LINE:int,PTR:bool,T:t_trace).
?- memset_call(FN,SUM,LINE).
?- memset_miss(FN,SUM,LINE).
?- memset_fail(FN,SUM,LINE,PTR).
?- memset_trace_fail(FN,SUM,LINE,PTR,T).

cursum(FN,SUM), dircall(I,CFN), memset_func(CFN,AT,_,_),
    icall(P,_,I), point_location(P,_,LINE),
    +memset_call(FN,SUM,LINE),
    ~memset_handled(I), +memset_miss(FN,SUM,LINE),
    may_use_field(I,P,drf{AT},PTR,T),
    ~(PTR=false, may_use_field(I,P,drf{AT},true,_)),
    +memset_fail(FN,SUM,LINE,PTR),
    +memset_trace_fail(FN,SUM,LINE,PTR,T).
