% middle-of-the-road aliasing/modset information. fairly pessimistic about aliasing,
% very optimistic about modsets. requires modset information generated by
% daliasing/init_usemod_func.clp

import "memory.clp".
import "../daliasing/aliasing_sum.clp".

% omit regular trace_alias/trace_clobber for specified traces
predicate omit_trace_alias(in FN:string,in SUM:sum,in T0:t_trace,in T1:t_trace).
predicate omit_trace_clobber(in I:sum,in P:pp,in T:t_trace).

% optimistic mod set information for call/loop summaries.
% only those traces which we definitely know might be updated are clobbered

trace_clobber(I,P,XT) :- ~omit_trace_clobber(I,P,XT),
    cfg_isum(P,I), isum_target_p(P,I,CFN,CSUM,direct),
    salias_write_func(CFN,CSUM)->wfunc_write(CT),
    inst_trace(I,P,CT,s_trace{T},_),
    break_point(P,FN,SUM,_), trace_alias(FN,SUM,T,XT,_).
trace_clobber(s_call{I},P,T) :- callret(P,I,T,G).

% turn inline writes in CFN into direct writes in this function.
% (these assign edges will be captured and persisted for remote queries)
predicate added_write_inline(I:c_instr,CT:t_trace).
icall(P,_,I), dircall(I,CFN),
    CFN\="copy_from_user", CFN\="copy_to_user",
    salias_write_func(CFN,s_func)->wfunc_write_inline(CT,CV),
    +added_write_inline(I,CT),
    inst_trace(s_call{I},P,CT,s_trace{T},TG),
    inst_scalar(s_call{I},P,CV,V,VG), #and(TG,VG,G), +assign(P,T,V,G).

% add an assign for the function's return value assignment,
% even if there is no write_inline in the callee
icall(P,_,I), callret(P,I,T,G),
    ~added_write_inline(I,root{return}),
    base_val(P,T,V,_), % this will be uc_sum so guard is true
    +assign(P,T,V,G).

% type-based aliasing information.
% x->f cannot alias y->g
% x->f cannot alias *p, unless x->f is itself a structure
% *p cannot alias *q if p and q point to objects of different types

trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    T0=fld{STR0,F,C}, T1=fld{STR1,F,C},
    trace_alias(FN,SUM,STR0,STR1,G).
trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    T0\=fld{_,_,_}, T1\=fld{_,_,_},
    (T0\=root{_}; T0=root{glob{_}}),
    (T1\=root{_}; T1=root{glob{_}}),
    trace_type(FN,T0,Y), trace_type(FN,T1,Y),
    #id_g(sc_bit{sb_alias{T0,T1}},G).
trace_alias(FN,SUM,T0,T1,G) :- T0\=T1, ~omit_trace_alias(FN,SUM,T0,T1),
    (T0\=fld{_,_,_}, T1=fld{_,_,_}; T1\=fld{_,_,_}, T0=fld{_,_,_}),
    trace_type(FN,T0,Y), trace_type(FN,T1,Y), Y=y_comp{_},
    #id_g(sc_bit{sb_alias{T0,T1}},G).

omit_trace_alias(_,_,T,_) :- trace_root(T,local{_}), ~trace_derefs(T,0).
omit_trace_alias(_,_,_,T) :- trace_root(T,local{_}), ~trace_derefs(T,0).

predicate print_trace_aliases().
predicate str_trace_alias(FN:string,SUM:sum,T0:string,T1:string).
?- str_trace_alias(FN,SUM,T0,T1).

print_trace_aliases(),
    ?trace_alias(FN,SUM,T0,T1,_), T0\=T1, trace_alias(FN,SUM,T0,T1,_),
    trace_string(T0,T0S), trace_string(T1,T1S),
    +str_trace_alias(FN,SUM,T0S,T1S).
