%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% baked-in code-base specific information to use during memory analysis.
% imported directly by both smemory/memory.clp and daliasing/memory.clp

% memcpy/memset

predicate memcpy_handled(I:c_instr).

predicate memcpy_func(FN:string,TGT:t_trace,SRC:t_trace,LEN:t_trace).
+memcpy_func("memcpy",root{arg{0}},root{arg{1}},root{arg{2}}).
+memcpy_func("memmove",root{arg{0}},root{arg{1}},root{arg{2}}).
+memcpy_func("include/asm/string.h:__memcpy",
             root{arg{0}},root{arg{1}},root{arg{2}}).
+memcpy_func("include/asm/string.h:__constant_memcpy",
             root{arg{0}},root{arg{1}},root{arg{2}}).

% recognize and model certain calls to memcpy. for now we just handle cases
% which copy an entire structure from one place to another
dircall(I,FN), memcpy_func(FN,TGT,SRC,LEN), icall(P,_,I), guard(P,G),
    inst_trace(s_call{I},P,drf{TGT},s_trace{TGTD},TGTG), guard_implies(G,TGTG),
    inst_trace(s_call{I},P,drf{SRC},s_trace{SRCD},SRCG), guard_implies(G,SRCG),
    cil_curfn(WFN),
    trace_type(WFN,TGTD,y_comp{C}), trace_type(WFN,SRCD,y_comp{C}),
    inst_trace(s_call{I},P,drf{LEN},s_const{N},LNG), guard_implies(G,LNG),
    cil_comp(C)->cil_comp_bytes(_,N),
    #bool_g(true,TG), +fill_assign(P,TGTD,s_trace{drf{SRCD}},y_comp{C},TG),
    +memcpy_handled(I).

predicate memset_handled(I:c_instr).

predicate memset_func(FN:string,TGT:t_trace,VAL:t_trace,LEN:t_trace).
+memset_func("memset",root{arg{0}},root{arg{1}},root{arg{2}}).
+memset_func("include/asm/string.h:__memset_generic",
             root{arg{0}},root{arg{1}},root{arg{2}}).
+memset_func("include/asm/string.h:__constant_c_memset",
             root{arg{0}},root{arg{1}},root{arg{2}}).
+memset_func("include/asm/string.h:__constant_c_and_count_memset",
             root{arg{0}},root{arg{1}},root{arg{2}}).

% similar to memcpy, recognize memset calls that set all the fields of an
% entire structure to zero
dircall(I,FN), memset_func(FN,TGT,VAL,LEN), icall(P,_,I), guard(P,G),
    inst_trace(s_call{I},P,drf{TGT},s_trace{TGTD},TGTG), guard_implies(G,TGTG),
    inst_trace(s_call{I},P,drf{VAL},s_const{0},VALG), guard_implies(G,VALG),
    cil_curfn(WFN), trace_type(WFN,TGTD,y_comp{C}),
    inst_trace(s_call{I},P,drf{LEN},s_const{N},LNG), guard_implies(G,LNG),
    cil_comp(C)->cil_comp_bytes(_,N),
    #bool_g(true,TG), +fill_assign(P,TGTD,s_const{0},y_comp{C},TG),
    +memset_handled(I).

% linux

% special case 'reverse field' pointer arithmetic from container_of macro:
% (ptr - (int)&((type*)0)->field)
cil_exp_binop(E,b_minuspi,LE,RE,_),
    cil_exp_cast(LE,LCE,_),
    cil_exp_cast(RE,RCE,_), cil_exp_addr(RCE,LV), cil_lval_mem(LV,ME,OFF),
    cil_exp_cast(ME,CME,_), cil_exp_x_intval(CME,0),
    +omit_eval(_,E), ?eval(P,E,_,_), eval_trace(P,LCE,T,G0),
    trace_rev_offset(P,T,OFF,FT,G1), #and(G0,G1,G), +eval(P,E,s_trace{FT},G).

% get names for untyped hidden fields $comp$N
predicate str_hidden_field(in COMP:string,in N:int,FLD:string).
?str_hidden_field(COMP,N,_),
    (int_ge(N,0), tostring(N,NS);
     int_lt(N,0), int_neg(N,NN), tostring(NN,NNS), str_cat("neg",NNS,NS)),
    str_cat_list(["$",COMP,"$",NS],FLD),
    +str_hidden_field(COMP,N,FLD).

% special case 'hidden field' pointer arithmetic: adding a fixed offset N
% to a structure pointer where N is negative or is larger than the size of
% the structure yields a hidden field for the comp type and offset
cil_exp_binop(E,b_pluspi,LE,RE,_),
    cil_exp_cast(LE,LCE,NTYP),
    convert_type(NTYP,y_ptr{DNY}), exp_type(LCE,y_ptr{DOY}),
    type_width(DOY,WIDTH), type_width(DNY,1),
    cil_exp_x_intval(RE,N), (int_lt(N,0); int_ge(N,WIDTH)),
    (DOY=y_comp{C}; DOY\=y_comp{_}, C="void"),
    str_hidden_field(C,N,F),
    +omit_eval(_,E), ?eval(P,E,_,_), eval_trace(P,LE,T,G),
    trace_simplify(fld{T,F,C},FT), +eval(P,E,s_trace{FT},G).
cil_exp_binop(E,b_pluspi,LE,RE,_),
    cil_exp_cast(_,E,_), exp_type(LE,y_ptr{y_comp{C}}),
    type_width(y_comp{C},N), cil_exp_x_intval(RE,RN),
    (RN=1; RN=2), int_mul(RN,N,NN), str_hidden_field(C,NN,F),
    +omit_eval(_,E), ?eval(P,E,_,_), eval_trace(P,LE,T,G),
    trace_simplify(fld{T,F,C},FT), +eval(P,E,s_trace{FT},G).

% linux and other systems occasionally cast pointers into integers to
% apply bitmasks and various other twiddling to them, typically to hide data
% in the lower bits or to force the pointer to a particular alignment.
% rather than understand exactly what's going on we just strip away the integer
% operations and evaluate to the core pointer/integer value within the exp
predicate arith_operator_cast(E:c_exp,BE:c_exp) order [E,BE].
arith_op(B), cil_exp_binop(CE,B,LE,RE,_), B\=b_pluspi, B\=b_minuspi,
    (cil_exp_cast(E,CE,TYP), cil_type_ptr(TYP,_); arith_operator_cast(E,CE)),
    (cil_exp_x_intval(LE,_), NE=RE;
     cil_exp_x_intval(RE,_), NE=LE;
     cil_exp_lval(RE,LV), cil_lval_var(LV,X,_),
         cil_var_name(X,"__per_cpu_offset",_), NE=LE),
    ~cil_exp_x_intval(NE,_), +arith_operator_cast(E,NE).
cil_exp_unop(CE,u_bnot,LE,_),
    (cil_exp_cast(E,CE,TYP), cil_type_ptr(TYP,_); arith_operator_cast(E,CE)),
    ~cil_exp_x_intval(LE,_),
    +arith_operator_cast(E,LE).
arith_operator_cast(E,CE), cil_exp_cast(CE,XCE,_), +arith_operator_cast(E,XCE).

predicate omit_nested_cast(in BE:c_exp).
omit_nested_cast(BE) :- cil_exp_cast(BE,_,_).
omit_nested_cast(BE) :- cil_exp_binop(BE,B,_,_,_), B\=b_pluspi, B\=b_minuspi.
omit_nested_cast(BE) :- cil_exp_unop(BE,_,_,_).

arith_operator_cast(E,BE), ~omit_nested_cast(BE),
    +omit_eval(_,E), ?eval(P,E,_,_), eval(P,BE,V,G), +eval(P,E,V,G).

% understand __builtin_expect, which just returns its first argument.
% updating assign() over the isum will force a strong update
% of the return value
dircall(I,"__builtin_expect"), icall(P,_,I),
    cil_instr_call_arg(I,0,E), callret(P,I,T,G0), eval(P,E,V,G1),
    #and(G0,G1,G), +assign(P,T,V,G).
