import "memory.clp".
import "aliasing_full.clp".
import "scalar_sat.clp".

analyze session_name("cil_sum_body").

?val(P,T,_,_), guard(P,PG), val(P,T,V0,G0), val(P,T,V1,G1), lessthan(V0,V1),
    #and3(PG,G0,G1,G), guard_sat(G),
    point_location(P,_,LINE), trace_string(T,TS),
    scalar_string(V0,V0S), scalar_string(V1,V1S),
    guard_string(G0,G0S), guard_string(G1,G1S),
    +warning("Overlapping val()",LINE,TS,V0S,V1S,G0S,G1S).

?val(P,T,_,_), guard(P,PG), \/val(P,T,_,XG):#or_all(XG,MG),
    ~guard_implies(PG,MG),
    point_location(P,_,LINE), trace_string(T,TS),
    +warning("Insufficient val()",LINE,TS).

predicate lval_candidate(P:pp,LV:c_lval).
(iset(P,_,I), cil_instr_set(I,LV,_);
 icall(P,_,I), cil_instr_call_ret(I,LV);
 iasm(P,_,I), cil_instr_asm_out(I,_,_,_,LV)),
    +lval_candidate(P,LV).

lval_candidate(P,LV), guard(P,PG), lval(P,LV,T0,G0), lval(P,LV,T1,G1), lessthan(T0,T1),
    #and3(PG,G0,G1,G), guard_sat(G),
    point_location(P,_,LINE), lval_string(LV,LVS),
    trace_string(T0,T0S), trace_string(T1,T1S),
    guard_string(G0,G0S), guard_string(G1,G1S),
    +warning("Overlapping lval()",LINE,LVS,T0S,T1S,G0S,G1S).

lval_candidate(P,LV), guard(P,PG), \/lval(P,LV,_,XG):#or_all(XG,MG),
    ~guard_implies(PG,MG),
    point_location(P,_,LINE), lval_string(LV,LVS),
    +warning("Insufficient lval()",LINE,LVS).

predicate eval_candidate(P:pp,E:c_exp).
(iset(P,_,I), cil_instr_set(I,_,E);
 icall(P,_,I), cil_instr_call(I,E);
 icall(P,_,I), cil_instr_call_arg(I,_,E);
 iasm(P,_,I), cil_instr_asm_in(I,_,_,_,E);
 branch(P,_,_,E)),
    +eval_candidate(P,E).

eval_candidate(P,E), guard(P,PG), eval(P,E,V0,G0), eval(P,E,V1,G1), lessthan(V0,V1),
    #and3(PG,G0,G1,G), guard_sat(G),
    point_location(P,_,LINE), exp_string(E,ES),
    scalar_string(V0,V0S), scalar_string(V1,V1S),
    guard_string(G0,G0S), guard_string(G1,G1S),
    +warning("Overlapping eval()",LINE,ES,V0S,V1S,G0S,G1S).

eval_candidate(P,E), guard(P,PG), \/eval(P,E,_,XG):#or_all(XG,MG),
    ~guard_implies(PG,MG),
    point_location(P,_,LINE), exp_string(E,ES),
    +warning("Insufficient eval()",LINE,ES).
