
open Spec
open Clpautil

module IO = Streamio
module H = Hashtbl
module V = Vector

let find_handlers = ref []

let reset_fns : (unit -> unit) list ref = ref []
let reset() = List.iter (fun f -> f()) !reset_fns

type setinfo = {
  key : t_val;
  capacity : int;
  values : (t_val,unit) H.t;
}

let tbl_sets : setinfo V.t = V.create 10
let tbl_var2set : (t_val*int,int) H.t = H.create 10
let _ = prepend reset_fns (fun () -> V.clear tbl_sets; H.clear tbl_var2set)

let add_boundset key capacity =
  let si =
    try H.find tbl_var2set (key,capacity)
    with Not_found ->
      let si = V.append tbl_sets
        { key = key; capacity = capacity; values = H.create capacity } in
      H.add tbl_var2set (key,capacity) si; si in
  si

let vt_boundset typ0 typ1 = make_type "boundset" [|typ0;typ1|]
let vk_boundset = register_val_kind {
  vki_tostring = (fun depth v ->
    let si = V.get tbl_sets v in
      "BSET[" ^ (string_of_val depth si.key) ^ ","
              ^ (string_of_int si.capacity) ^ "]");
  vki_compare = compare;
  vki_dump = (fun _ _ -> failwith "Cannot dump boundset");
  vki_load = (fun _ -> failwith "Cannot load boundset");
}

let val2boundset = val2cst vk_boundset
let boundset2val = cst2val vk_boundset

let vt_xvar = vt_var "X"
let vt_yvar = vt_var "Y"
let vt_xboundset = vt_boundset vt_xvar vt_yvar

let _ = prepend find_handlers
  (("boundset_create",[|vt_xvar;vt_int;vt_xboundset|]),
   [[|true;true;false|],[Sk_once]],
   fun (name,args) ->
     let si = add_boundset args.(0) (val2sint args.(1)) in
       [name,[|args.(0);args.(1);boundset2val si|]])

let _ = prepend find_handlers
  (("boundset_create_unbounded",[|vt_xvar;vt_xboundset|]),
   [[|true;false|],[Sk_once]],
   fun (name,args) ->
     let si = add_boundset args.(0) 0 in
       [name,[|args.(0);boundset2val si|]])

let _ = prepend find_handlers
  (("boundset_insert",[|vt_xboundset;vt_yvar|]),
   [[|true;true|],[Sk_zero;Sk_once]],
   fun (name,args) ->
     let sinfo = V.get tbl_sets (val2boundset args.(0)) in
       if H.mem sinfo.values args.(1) then [name,args]
       else if sinfo.capacity > 0 && H.length sinfo.values < sinfo.capacity then
         (H.add sinfo.values args.(1) (); [name,args])
       else [])

let _ = prepend find_handlers
  (("boundset_remaining",[|vt_xboundset;vt_int|]),
   [[|true;true|],[Sk_once]],
   fun (name,args) ->
     let sinfo = V.get tbl_sets (val2boundset args.(0)) in
     let count = val2sint args.(1) in
       if sinfo.capacity = 0 then [name,args]
       else if sinfo.capacity - H.length sinfo.values >= count then [name,args]
       else [])

let _ = register_package { empty_package with
  p_name = "boundops";
  p_reset = reset;
  p_find_handlers = !find_handlers;
  p_abstypes = [vt_xboundset]
}
