(* Unique ID counter abstraction.
 *
 * This module provides functions to hand out consecutively numbered integers.
 * By introducing an abstraction barrier, we remove any chance of confusion
 * as to whether the value stored is the last number or the next number
 * to be handed out. *)

type t = int ref

(* "init n" returns a counter whose first allocation will be the integer n *)
let init (n:int) : t = ref n

(* "allocate" returns the next value of the counter *)
let allocate (c:t) = let v = !c in incr c; v
