(*
 * Authors:
 *   Brian Hackett  <bhackett@stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

type t_resource =
    RLIMIT_CORE
  | RLIMIT_CPU
  | RLIMIT_DATA
  | RLIMIT_FSIZE
  | RLIMIT_NOFILE
  | RLIMIT_STACK
  | RLIMIT_AS

external rlimit_set : t_resource -> (Int64.t * Int64.t) -> unit = "ml_setrlimit"
external rlimit_get : t_resource -> (Int64.t * Int64.t) = "ml_getrlimit"
external is_infinite : Int64.t -> bool = "ml_infiniterlimit"

(* useful helper for getting the true current memory usage, in bytes.
 * this is really inefficient and invokes shell commands so should
 * be called only when debugging memory usage *)
let get_current_memory() =
  let pid = Unix.getpid() in
  let file = Printf.sprintf "memory-%d.txt" pid in
  let cmd = Printf.sprintf "ps -p %d -o size= > %s" pid file in
    ignore (Unix.system cmd);
    let chan = open_in file in
    let sizestr = input_line chan in
    let size = Int64.of_string sizestr in
      close_in chan;
      Int64.mul size (Int64.of_int 1024)
